/* Generated by ./dbus-analyze-signals.py.  Do not edit! */
const char *dbus_signals = 
"    <signal name='AccountConnecting'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountDisabled'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountEnabled'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountSettingInfo'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountSetInfo'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountCreated'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountDestroying'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountAdded'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountRemoved'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountStatusChanging'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountStatusChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountActionsChanged'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountAliasChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountAuthorizationRequested'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountAuthorizationRequestedWithMessage'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountAuthorizationDenied'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountAuthorizationGranted'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountErrorChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountSignedOn'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountSignedOff'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='AccountConnectionError'>\n"
"      <arg type='p'/>\n"
"      <arg type='i'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyStatusChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyPrivacyChanged'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyIdleChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='i'/>\n"
"      <arg type='i'/>\n"
"    </signal>\n"
"    <signal name='BuddySignedOn'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddySignedOff'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyGotLoginTime'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BlistNodeAdded'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BlistNodeRemoved'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyAdded'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyRemoved'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyRemovedFromGroup'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyIconChanged'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='UpdateIdle'>\n"
"    </signal>\n"
"    <signal name='BlistNodeExtendedMenu'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BlistNodeAliased'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyCapsChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='i'/>\n"
"      <arg type='i'/>\n"
"    </signal>\n"
"    <signal name='CertificateStored'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='CertificateDeleted'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='CipherAdded'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='CipherRemoved'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='CmdAdded'>\n"
"      <arg type='p'/>\n"
"      <arg type='i'/>\n"
"      <arg type='i'/>\n"
"    </signal>\n"
"    <signal name='CmdRemoved'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SigningOn'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SignedOn'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SigningOff'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SignedOff'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ConnectionError'>\n"
"      <arg type='p'/>\n"
"      <arg type='i'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='Autojoin'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='WritingImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='WroteImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='SentAttention'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='GotAttention'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='SendingImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SentImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ReceivingImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ReceivedImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='BlockedImMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='WritingChatMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='WroteChatMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='SendingChatMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='SentChatMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='ReceivingChatMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ReceivedChatMsg'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='ConversationCreated'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ConversationUpdated'>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='DeletingConversation'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyTyping'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyTyped'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='BuddyTypingStopped'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatBuddyJoining'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='ChatBuddyJoined'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='ChatBuddyFlags'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='ChatBuddyLeaving'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatBuddyLeft'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='DeletingChatBuddy'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatInvitingUser'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatInvitedUser'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatInvited'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatInviteBlocked'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatJoined'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatJoinFailed'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatLeft'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ChatTopicChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ClearedMessageHistory'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ConversationExtendedMenu'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='UriHandler'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='Quitting'>\n"
"    </signal>\n"
"    <signal name='FileRecvAccept'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileSendAccept'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileRecvStart'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileSendStart'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileSendCancel'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileRecvCancel'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileSendComplete'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileRecvComplete'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='FileRecvRequest'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='ImageDeleting'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='LogTimestamp'>\n"
"      <arg type='p'/>\n"
"      <arg type='x'/>\n"
"      <arg type='b'/>\n"
"    </signal>\n"
"    <signal name='NetworkConfigurationChanged'>\n"
"    </signal>\n"
"    <signal name='DisplayingEmailNotification'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='DisplayingEmailsNotification'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='u'/>\n"
"    </signal>\n"
"    <signal name='DisplayingUserinfo'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='PluginLoad'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='PluginUnload'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SavedstatusChanged'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SavedstatusAdded'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SavedstatusDeleted'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='SavedstatusModified'>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='PlayingSoundEvent'>\n"
"      <arg type='i'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='IrcSendingText'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
"    <signal name='IrcReceivingText'>\n"
"      <arg type='p'/>\n"
"      <arg type='p'/>\n"
"    </signal>\n"
;
