/* Generated by ./dbus-analyze-functions.py.  Do not edit! */
static DBusMessage*
purple_accounts_find_any_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *protocol;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &protocol, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	protocol = (protocol && protocol[0]) ? protocol : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_accounts_find_any(name, protocol), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_find_connected_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *protocol;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &protocol, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	protocol = (protocol && protocol[0]) ? protocol : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_accounts_find_connected(name, protocol), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BLIST_NODE_IS_CHAT_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = PURPLE_BLIST_NODE_IS_CHAT(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BLIST_NODE_IS_BUDDY_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = PURPLE_BLIST_NODE_IS_BUDDY(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BLIST_NODE_IS_CONTACT_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = PURPLE_BLIST_NODE_IS_CONTACT(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BLIST_NODE_IS_GROUP_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = PURPLE_BLIST_NODE_IS_GROUP(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BUDDY_IS_ONLINE_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	RESULT = PURPLE_BUDDY_IS_ONLINE(buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BLIST_NODE_HAS_FLAG_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t flags;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = PURPLE_BLIST_NODE_HAS_FLAG(node, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_BLIST_NODE_SHOULD_SAVE_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = PURPLE_BLIST_NODE_SHOULD_SAVE(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_CONNECTION_IS_CONNECTED_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t connection_ID;
	PurpleConnection *connection;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &connection_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(connection, connection_ID, PurpleConnection, error_DBUS);
	RESULT = PURPLE_CONNECTION_IS_CONNECTED(connection);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_CONNECTION_IS_VALID_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t connection_ID;
	PurpleConnection *connection;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &connection_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(connection, connection_ID, PurpleConnection, error_DBUS);
	RESULT = PURPLE_CONNECTION_IS_VALID(connection);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_CONV_IM_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conversation_ID;
	PurpleConversation *conversation;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conversation_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conversation, conversation_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, PURPLE_CONV_IM(conversation), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
PURPLE_CONV_CHAT_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conversation_ID;
	PurpleConversation *conversation;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conversation_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conversation, conversation_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, PURPLE_CONV_CHAT(conversation), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *username;
	const char *protocol_id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &username, DBUS_TYPE_STRING, &protocol_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	username = (username && username[0]) ? username : NULL;
	protocol_id = (protocol_id && protocol_id[0]) ? protocol_id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_new(username, protocol_id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_destroy(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_connect_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_connect(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_register_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_register(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_disconnect_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_disconnect(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_notify_added_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *remote_user;
	const char *id;
	const char *alias;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &remote_user, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &alias, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	remote_user = (remote_user && remote_user[0]) ? remote_user : NULL;
	id = (id && id[0]) ? id : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_account_notify_added(account, remote_user, id, alias, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_request_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *remote_user;
	const char *id;
	const char *alias;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &remote_user, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &alias, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	remote_user = (remote_user && remote_user[0]) ? remote_user : NULL;
	id = (id && id[0]) ? id : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_account_request_add(account, remote_user, id, alias, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_request_close_with_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_request_close_with_account(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_request_close_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ui_handle_NULL;
	void *ui_handle;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ui_handle_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ui_handle = NULL;
	purple_account_request_close(ui_handle);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_request_change_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_request_change_password(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_request_change_user_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_request_change_user_info(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_username_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *username;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &username, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	username = (username && username[0]) ? username : NULL;
	purple_account_set_username(account, username);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *password;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &password, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	password = (password && password[0]) ? password : NULL;
	purple_account_set_password(account, password);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	purple_account_set_alias(account, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_user_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *user_info;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &user_info, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	user_info = (user_info && user_info[0]) ? user_info : NULL;
	purple_account_set_user_info(account, user_info);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_buddy_icon_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *path;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &path, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	path = (path && path[0]) ? path : NULL;
	purple_account_set_buddy_icon_path(account, path);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_protocol_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *protocol_id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &protocol_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	protocol_id = (protocol_id && protocol_id[0]) ? protocol_id : NULL;
	purple_account_set_protocol_id(account, protocol_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_connection_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	purple_account_set_connection(account, gc);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_remember_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_set_remember_password(account, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_check_mail_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_set_check_mail(account, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_enabled_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	purple_account_set_enabled(account, ui, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_proxy_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t info_ID;
	PurpleProxyInfo *info;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &info_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(info, info_ID, PurpleProxyInfo, error_DBUS);
	purple_account_set_proxy_info(account, info);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_privacy_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t privacy_type;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &privacy_type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_set_privacy_type(account, privacy_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_status_types_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t status_types_NULL;
	GList *status_types;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &status_types_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	status_types = NULL;
	purple_account_set_status_types(account, status_types);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_status_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *status_id;
	dbus_int32_t active;
	dbus_int32_t attrs_NULL;
	GList *attrs;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INT32, &active, DBUS_TYPE_INT32, &attrs_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	attrs = NULL;
	purple_account_set_status_list(account, status_id, active, attrs);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_silence_suppression_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_get_silence_suppression(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_silence_suppression_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_set_silence_suppression(account, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_clear_settings_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_clear_settings(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_remove_setting_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *setting;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &setting, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	setting = (setting && setting[0]) ? setting : NULL;
	purple_account_remove_setting(account, setting);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_account_set_int(account, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_account_set_string(account, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_account_set_bool(account, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_ui_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	name = (name && name[0]) ? name : NULL;
	purple_account_set_ui_int(account, ui, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_ui_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	const char *name;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	name = (name && name[0]) ? name : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_account_set_ui_string(account, ui, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_set_ui_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	name = (name && name[0]) ? name : NULL;
	purple_account_set_ui_bool(account, ui, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_is_connected_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_is_connected(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_is_connecting_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_is_connecting(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_is_disconnected_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_is_disconnected(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_username_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_username(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_password(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_alias(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_user_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_user_info(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_buddy_icon_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_buddy_icon_path(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_protocol_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_protocol_id(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_protocol_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_protocol_name(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_connection_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_connection(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_name_for_display_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_account_get_name_for_display(account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_remember_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_get_remember_password(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_check_mail_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_get_check_mail(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_enabled_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	RESULT = purple_account_get_enabled(account, ui);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_proxy_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_proxy_info(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_privacy_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_account_get_privacy_type(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_active_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_active_status(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *status_id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_status(account, status_id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_status_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_status_type(account, id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_status_type_with_primitive_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t primitive;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &primitive, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_status_type_with_primitive(account, primitive), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_presence_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_presence(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_is_status_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *status_id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	RESULT = purple_account_is_status_active(account, status_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_status_types_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	list = purple_account_get_status_types(account);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t default_value;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &default_value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_account_get_int(account, name, default_value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	const char *default_value;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &default_value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	default_value = (default_value && default_value[0]) ? default_value : NULL;
	if ((RESULT = purple_account_get_string(account, name, default_value)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t default_value;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &default_value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_account_get_bool(account, name, default_value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_ui_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	const char *name;
	dbus_int32_t default_value;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &default_value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_account_get_ui_int(account, ui, name, default_value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_ui_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	const char *name;
	const char *default_value;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &default_value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	name = (name && name[0]) ? name : NULL;
	default_value = (default_value && default_value[0]) ? default_value : NULL;
	if ((RESULT = purple_account_get_ui_string(account, ui, name, default_value)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_ui_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ui;
	const char *name;
	dbus_int32_t default_value;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ui, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &default_value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_account_get_ui_bool(account, ui, name, default_value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_log_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t create;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &create, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_log(account, create), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_destroy_log_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_destroy_log(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_add_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	purple_account_add_buddy(account, buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_add_buddy_with_invite_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_account_add_buddy_with_invite(account, buddy, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_add_buddies_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddies_NULL;
	GList *buddies;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddies_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	buddies = NULL;
	purple_account_add_buddies(account, buddies);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_add_buddies_with_invite_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddies_NULL;
	GList *buddies;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddies_NULL, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	buddies = NULL;
	message = (message && message[0]) ? message : NULL;
	purple_account_add_buddies_with_invite(account, buddies, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_remove_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	purple_account_remove_buddy(account, buddy, group);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_remove_buddies_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddies_NULL;
	GList *buddies;
	dbus_int32_t groups_NULL;
	GList *groups;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddies_NULL, DBUS_TYPE_INT32, &groups_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	buddies = NULL;
	groups = NULL;
	purple_account_remove_buddies(account, buddies, groups);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_remove_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	purple_account_remove_group(account, group);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_change_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *orig_pw;
	const char *new_pw;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &orig_pw, DBUS_TYPE_STRING, &new_pw, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	orig_pw = (orig_pw && orig_pw[0]) ? orig_pw : NULL;
	new_pw = (new_pw && new_pw[0]) ? new_pw : NULL;
	purple_account_change_password(account, orig_pw, new_pw);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_supports_offline_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	RESULT = purple_account_supports_offline_message(account, buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_get_current_error_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_get_current_error(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_clear_current_error_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_account_clear_current_error(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_accounts_add(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_remove_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_accounts_remove(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_delete_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_accounts_delete(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_reorder_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t new_index;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &new_index, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_accounts_reorder(account, new_index);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_get_all_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_accounts_get_all();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_get_all_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_accounts_get_all_active();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_find_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *protocol;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &protocol, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	protocol = (protocol && protocol[0]) ? protocol : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_accounts_find(name, protocol), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_restore_current_statuses_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_accounts_restore_current_statuses();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleAccountUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleAccountUiOps, error_DBUS);
	purple_accounts_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_accounts_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_prefs_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleAccountPrefsUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleAccountPrefsUiOps, error_DBUS);
	purple_account_prefs_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_account_prefs_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_account_prefs_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_accounts_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_accounts_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_accounts_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_new(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_set_blist_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t blist_ID;
	PurpleBuddyList *blist;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &blist_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(blist, blist_ID, PurpleBuddyList, error_DBUS);
	purple_set_blist(blist);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_get_blist_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_get_blist(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_get_root_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_get_root(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_get_buddies_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GSList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_blist_get_buddies();
	RESULT = purple_dbusify_GSList(list, FALSE, &RESULT_LEN);
	g_slist_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_next_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t offline;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INT32, &offline, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_node_next(node, offline), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_parent_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_node_get_parent(node), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_first_child_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_node_get_first_child(node), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_sibling_next_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_node_get_sibling_next(node), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_sibling_prev_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_node_get_sibling_prev(node), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_show_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_show();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_destroy();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_set_visible_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t show;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &show, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_set_visible(show);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_update_buddy_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t old_status_ID;
	PurpleStatus *old_status;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INT32, &old_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(old_status, old_status_ID, PurpleStatus, error_DBUS);
	purple_blist_update_buddy_status(buddy, old_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_update_node_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_update_node_icon(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_update_buddy_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	purple_blist_update_buddy_icon(buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_rename_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_blist_rename_buddy(buddy, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_alias_contact_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	purple_blist_alias_contact(contact, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_alias_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	purple_blist_alias_buddy(buddy, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_server_alias_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	purple_blist_server_alias_buddy(buddy, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_alias_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	purple_blist_alias_chat(chat, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_rename_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_blist_rename_group(group, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_chat_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *alias;
	DBusMessageIter components_ITER;
	GHashTable *components;
	dbus_int32_t RESULT;
	purple_dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_ARRAY, &components_ITER, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	components = purple_dbus_iter_hash_table(&components_ITER, error_DBUS);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_chat_new(account, alias, components), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	g_hash_table_destroy(components);
	return reply_DBUS;
}

static DBusMessage*
purple_chat_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	purple_chat_destroy(chat);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_add_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_add_chat(chat, group, node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	const char *alias;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_new(account, name, alias), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	purple_buddy_destroy(buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_set_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	purple_buddy_set_icon(buddy, icon);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_get_account(buddy), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_name(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_get_icon(buddy), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_contact_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_get_contact(buddy), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_presence_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_get_presence(buddy), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_media_caps_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	RESULT = purple_buddy_get_media_caps(buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_set_media_caps_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t media_caps;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INT32, &media_caps, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	purple_buddy_set_media_caps(buddy, media_caps);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_add_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_add_buddy(buddy, contact, group, node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_group_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_group_new(name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_group_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	purple_group_destroy(group);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_add_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_add_group(group, node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_contact_new(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	purple_contact_destroy(contact);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_get_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_contact_get_group(contact), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_add_contact_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_add_contact(contact, group, node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_merge_contact_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t source_ID;
	PurpleContact *source;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &source_ID, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(source, source_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_merge_contact(source, node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_get_priority_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_contact_get_priority_buddy(contact), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_set_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	purple_contact_set_alias(contact, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_get_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	if ((RESULT = purple_contact_get_alias(contact)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_on_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_contact_on_account(contact, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_contact_invalidate_priority_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	purple_contact_invalidate_priority_buddy(contact);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_remove_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	purple_blist_remove_buddy(buddy);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_remove_contact_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	purple_blist_remove_contact(contact);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_remove_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	purple_blist_remove_chat(chat);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_remove_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	purple_blist_remove_group(group);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_alias_only_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_alias_only(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_server_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_server_alias(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_contact_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_contact_alias(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_local_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_local_alias(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_alias(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_local_buddy_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	if ((RESULT = purple_buddy_get_local_buddy_alias(buddy)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_chat_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	if ((RESULT = purple_chat_get_name(chat)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_find_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_find_buddy(account, name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_find_buddy_in_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_find_buddy_in_group(account, name, group), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_find_buddies_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t RESULT_LEN;
	GSList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	list = purple_find_buddies(account, name);
	RESULT = purple_dbusify_GSList(list, FALSE, &RESULT_LEN);
	g_slist_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_find_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_find_group(name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_find_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_find_chat(account, name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_chat_get_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_chat_get_group(chat), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_chat_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleChat *chat;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleChat, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_chat_get_account(chat), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_get_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_get_group(buddy), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_group_get_accounts_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t g_ID;
	PurpleGroup *g;
	dbus_int32_t RESULT_LEN;
	GSList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &g_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(g, g_ID, PurpleGroup, error_DBUS);
	list = purple_group_get_accounts(g);
	RESULT = purple_dbusify_GSList(list, FALSE, &RESULT_LEN);
	g_slist_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_group_on_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t g_ID;
	PurpleGroup *g;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &g_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(g, g_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_group_on_account(g, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_group_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	if ((RESULT = purple_group_get_name(group)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_add_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_blist_add_account(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_remove_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_blist_remove_account(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_get_group_size_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t offline;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INT32, &offline, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	RESULT = purple_blist_get_group_size(group, offline);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_get_group_online_count_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	RESULT = purple_blist_get_group_online_count(group);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_load_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_load();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_schedule_save_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_schedule_save();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_request_add_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *username;
	const char *group;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &username, DBUS_TYPE_STRING, &group, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	username = (username && username[0]) ? username : NULL;
	group = (group && group[0]) ? group : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	purple_blist_request_add_buddy(account, username, group, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_request_add_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t group_ID;
	PurpleGroup *group;
	const char *alias;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &group_ID, DBUS_TYPE_STRING, &alias, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(group, group_ID, PurpleGroup, error_DBUS);
	alias = (alias && alias[0]) ? alias : NULL;
	name = (name && name[0]) ? name : NULL;
	purple_blist_request_add_chat(account, group, alias, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_request_add_group_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_request_add_group();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_set_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	purple_blist_node_set_bool(node, key, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	RESULT = purple_blist_node_get_bool(node, key);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_set_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	purple_blist_node_set_int(node, key, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	RESULT = purple_blist_node_get_int(node, key);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_set_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_blist_node_set_string(node, key, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	if ((RESULT = purple_blist_node_get_string(node, key)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_remove_setting_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *key;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	purple_blist_node_remove_setting(node, key);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_set_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t flags;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	purple_blist_node_set_flags(node, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = purple_blist_node_get_flags(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = purple_blist_node_get_type(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_node_get_extended_menu_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t n_ID;
	PurpleBlistNode *n;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &n_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(n, n_ID, PurpleBlistNode, error_DBUS);
	list = purple_blist_node_get_extended_menu(n);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleBlistUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleBlistUiOps, error_DBUS);
	purple_blist_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_blist_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_blist_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_blist_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *username;
	dbus_int32_t icon_data_NULL;
	void *icon_data;
	dbus_int32_t icon_len;
	const char *checksum;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &username, DBUS_TYPE_INT32, &icon_data_NULL, DBUS_TYPE_INT32, &icon_len, DBUS_TYPE_STRING, &checksum, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	username = (username && username[0]) ? username : NULL;
	icon_data = NULL;
	checksum = (checksum && checksum[0]) ? checksum : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icon_new(account, username, icon_data, icon_len, checksum), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_ref_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icon_ref(icon), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_unref_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icon_unref(icon), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_update_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	purple_buddy_icon_update(icon);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_set_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_int32_t data_NULL;
	guchar *data;
	dbus_int32_t len;
	const char *checksum;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INT32, &data_NULL, DBUS_TYPE_INT32, &len, DBUS_TYPE_STRING, &checksum, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	data = NULL;
	checksum = (checksum && checksum[0]) ? checksum : NULL;
	purple_buddy_icon_set_data(icon, data, len, checksum);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icon_get_account(icon), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_username_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	if ((RESULT = purple_buddy_icon_get_username(icon)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_checksum_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	if ((RESULT = purple_buddy_icon_get_checksum(icon)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	size_t len = 0;
	gconstpointer RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	RESULT = purple_buddy_icon_get_data(icon, &len);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &RESULT, len, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_extension_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	if ((RESULT = purple_buddy_icon_get_extension(icon)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_full_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	if ((RESULT = purple_buddy_icon_get_full_path(icon)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_set_for_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *username;
	dbus_int32_t icon_data_NULL;
	void *icon_data;
	dbus_int32_t icon_len;
	const char *checksum;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &username, DBUS_TYPE_INT32, &icon_data_NULL, DBUS_TYPE_INT32, &icon_len, DBUS_TYPE_STRING, &checksum, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	username = (username && username[0]) ? username : NULL;
	icon_data = NULL;
	checksum = (checksum && checksum[0]) ? checksum : NULL;
	purple_buddy_icons_set_for_user(account, username, icon_data, icon_len, checksum);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_find_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *username;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &username, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	username = (username && username[0]) ? username : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_find(account, username), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_find_account_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_find_account_icon(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_set_account_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t icon_data_NULL;
	guchar *icon_data;
	dbus_int32_t icon_len;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &icon_data_NULL, DBUS_TYPE_INT32, &icon_len, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	icon_data = NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_set_account_icon(account, icon_data, icon_len), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_get_account_icon_timestamp_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_buddy_icons_get_account_icon_timestamp(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_node_has_custom_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	RESULT = purple_buddy_icons_node_has_custom_icon(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_node_find_custom_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_node_find_custom_icon(node), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_node_set_custom_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	dbus_int32_t icon_data_NULL;
	guchar *icon_data;
	dbus_int32_t icon_len;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INT32, &icon_data_NULL, DBUS_TYPE_INT32, &icon_len, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	icon_data = NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_node_set_custom_icon(node, icon_data, icon_len), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_node_set_custom_icon_from_file_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	PurpleBlistNode *node;
	const char *filename;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &filename, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, PurpleBlistNode, error_DBUS);
	filename = (filename && filename[0]) ? filename : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_node_set_custom_icon_from_file(node, filename), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_has_custom_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	RESULT = purple_buddy_icons_has_custom_icon(contact);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_find_custom_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_find_custom_icon(contact), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_set_custom_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t contact_ID;
	PurpleContact *contact;
	dbus_int32_t icon_data_NULL;
	guchar *icon_data;
	dbus_int32_t icon_len;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &contact_ID, DBUS_TYPE_INT32, &icon_data_NULL, DBUS_TYPE_INT32, &icon_len, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(contact, contact_ID, PurpleContact, error_DBUS);
	icon_data = NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_buddy_icons_set_custom_icon(contact, icon_data, icon_len), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_set_caching_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t caching;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &caching, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_buddy_icons_set_caching(caching);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_is_caching_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_buddy_icons_is_caching();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_set_cache_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *cache_dir;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &cache_dir, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	cache_dir = (cache_dir && cache_dir[0]) ? cache_dir : NULL;
	purple_buddy_icons_set_cache_dir(cache_dir);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_get_cache_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_buddy_icons_get_cache_dir()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_buddy_icons_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icons_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_buddy_icons_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_buddy_icon_get_scale_size_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t spec_ID;
	PurpleBuddyIconSpec *spec;
	dbus_int32_t width_NULL;
	int *width;
	dbus_int32_t height_NULL;
	int *height;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &spec_ID, DBUS_TYPE_INT32, &width_NULL, DBUS_TYPE_INT32, &height_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(spec, spec_ID, PurpleBuddyIconSpec, error_DBUS);
	width = NULL;
	height = NULL;
	purple_buddy_icon_get_scale_size(spec, width, height);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t regist;
	const char *password;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &regist, DBUS_TYPE_STRING, &password, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	password = (password && password[0]) ? password : NULL;
	purple_connection_new(account, regist, password);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	purple_connection_destroy(gc);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_set_state_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t state;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	purple_connection_set_state(gc, state);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_set_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_connection_set_account(gc, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_set_display_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_connection_set_display_name(gc, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_set_protocol_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t connection_ID;
	PurpleConnection *connection;
	dbus_int32_t proto_data_NULL;
	void *proto_data;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &connection_ID, DBUS_TYPE_INT32, &proto_data_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(connection, connection_ID, PurpleConnection, error_DBUS);
	proto_data = NULL;
	purple_connection_set_protocol_data(connection, proto_data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_get_state_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	RESULT = purple_connection_get_state(gc);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_connection_get_account(gc), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_get_prpl_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_connection_get_prpl(gc), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_get_password_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	if ((RESULT = purple_connection_get_password(gc)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_get_display_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	if ((RESULT = purple_connection_get_display_name(gc)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_update_progress_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *text;
	dbus_int32_t step;
	dbus_int32_t count;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &text, DBUS_TYPE_INT32, &step, DBUS_TYPE_INT32, &count, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	text = (text && text[0]) ? text : NULL;
	purple_connection_update_progress(gc, text, step, count);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_notice_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *text;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &text, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	text = (text && text[0]) ? text : NULL;
	purple_connection_notice(gc, text);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_error_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *reason;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &reason, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	reason = (reason && reason[0]) ? reason : NULL;
	purple_connection_error(gc, reason);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_error_reason_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t reason;
	const char *description;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &reason, DBUS_TYPE_STRING, &description, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	description = (description && description[0]) ? description : NULL;
	purple_connection_error_reason(gc, reason, description);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_ssl_error_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t ssl_error;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &ssl_error, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	purple_connection_ssl_error(gc, ssl_error);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connection_error_is_fatal_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t reason;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &reason, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_connection_error_is_fatal(reason);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_disconnect_all_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_connections_disconnect_all();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_get_all_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_connections_get_all();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_get_connecting_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_connections_get_connecting();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleConnectionUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleConnectionUiOps, error_DBUS);
	purple_connections_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_connections_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_connections_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_connections_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_connections_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conversation_new(type, account, name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_destroy(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_present_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_present(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	RESULT = purple_conversation_get_type(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t ops_ID;
	PurpleConversationUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleConversationUiOps, error_DBUS);
	purple_conversation_set_ui_ops(conv, ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversations_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleConversationUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleConversationUiOps, error_DBUS);
	purple_conversations_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conversation_get_ui_ops(conv), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_set_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_conversation_set_account(conv, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conversation_get_account(conv), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_gc_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conversation_get_gc(conv), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_set_title_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *title;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_STRING, &title, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	title = (title && title[0]) ? title : NULL;
	purple_conversation_set_title(conv, title);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_title_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	if ((RESULT = purple_conversation_get_title(conv)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_autoset_title_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_autoset_title(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_set_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_conversation_set_name(conv, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	if ((RESULT = purple_conversation_get_name(conv)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_get_attribute_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t cb_ID;
	PurpleConvChatBuddy *cb;
	const char *key;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &cb_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(cb, cb_ID, PurpleConvChatBuddy, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	if ((RESULT = purple_conv_chat_cb_get_attribute(cb, key)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_get_attribute_keys_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t cb_ID;
	PurpleConvChatBuddy *cb;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &cb_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(cb, cb_ID, PurpleConvChatBuddy, error_DBUS);
	list = purple_conv_chat_cb_get_attribute_keys(cb);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_set_attribute_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t cb_ID;
	PurpleConvChatBuddy *cb;
	const char *key;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &cb_ID, DBUS_TYPE_STRING, &key, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(cb, cb_ID, PurpleConvChatBuddy, error_DBUS);
	key = (key && key[0]) ? key : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_conv_chat_cb_set_attribute(chat, cb, key, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_set_attributes_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t cb_ID;
	PurpleConvChatBuddy *cb;
	dbus_int32_t keys_NULL;
	GList *keys;
	dbus_int32_t values_NULL;
	GList *values;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &cb_ID, DBUS_TYPE_INT32, &keys_NULL, DBUS_TYPE_INT32, &values_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(cb, cb_ID, PurpleConvChatBuddy, error_DBUS);
	keys = NULL;
	values = NULL;
	purple_conv_chat_cb_set_attributes(chat, cb, keys, values);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_set_logging_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t log;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INT32, &log, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_set_logging(conv, log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_is_logging_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	RESULT = purple_conversation_is_logging(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_im_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conversation_get_im_data(conv), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_chat_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conversation_get_chat_data(conv), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_get_conversations_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_get_conversations();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_get_ims_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_get_ims();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_get_chats_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_get_chats();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_find_conversation_with_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_find_conversation_with_account(type, name, account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_write_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *who;
	const char *message;
	dbus_int32_t flags;
	dbus_int32_t mtime;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INT32, &mtime, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_conversation_write(conv, who, message, flags, mtime);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_set_features_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t features;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INT32, &features, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_set_features(conv, features);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_features_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	RESULT = purple_conversation_get_features(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_has_focus_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	RESULT = purple_conversation_has_focus(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_update_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t type;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_update(conv, type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_message_history_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	list = purple_conversation_get_message_history(conv);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_clear_message_history_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	purple_conversation_clear_message_history(conv);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_message_get_sender_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t msg_ID;
	PurpleConvMessage *msg;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &msg_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(msg, msg_ID, PurpleConvMessage, error_DBUS);
	if ((RESULT = purple_conversation_message_get_sender(msg)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_message_get_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t msg_ID;
	PurpleConvMessage *msg;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &msg_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(msg, msg_ID, PurpleConvMessage, error_DBUS);
	if ((RESULT = purple_conversation_message_get_message(msg)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_message_get_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t msg_ID;
	PurpleConvMessage *msg;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &msg_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(msg, msg_ID, PurpleConvMessage, error_DBUS);
	RESULT = purple_conversation_message_get_flags(msg);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_message_get_timestamp_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t msg_ID;
	PurpleConvMessage *msg;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &msg_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(msg, msg_ID, PurpleConvMessage, error_DBUS);
	RESULT = purple_conversation_message_get_timestamp(msg);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_get_conversation_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conv_im_get_conversation(im), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_set_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t icon_ID;
	PurpleBuddyIcon *icon;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INT32, &icon_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(icon, icon_ID, PurpleBuddyIcon, error_DBUS);
	purple_conv_im_set_icon(im, icon);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_get_icon_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conv_im_get_icon(im), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_set_typing_state_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t state;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_set_typing_state(im, state);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_get_typing_state_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	RESULT = purple_conv_im_get_typing_state(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_start_typing_timeout_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t timeout;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INT32, &timeout, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_start_typing_timeout(im, timeout);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_stop_typing_timeout_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_stop_typing_timeout(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_get_typing_timeout_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	RESULT = purple_conv_im_get_typing_timeout(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_set_type_again_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_uint32_t val;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_UINT32, &val, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_set_type_again(im, val);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_get_type_again_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	RESULT = purple_conv_im_get_type_again(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_start_send_typed_timeout_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_start_send_typed_timeout(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_stop_send_typed_timeout_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_stop_send_typed_timeout(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_get_send_typed_timeout_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	RESULT = purple_conv_im_get_send_typed_timeout(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_update_typing_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	purple_conv_im_update_typing(im);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_write_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	const char *who;
	const char *message;
	dbus_int32_t flags;
	dbus_int32_t mtime;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INT32, &mtime, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_conv_im_write(im, who, message, flags, mtime);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_present_error_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *who;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *what;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &what, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	who = (who && who[0]) ? who : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	what = (what && what[0]) ? what : NULL;
	RESULT = purple_conv_present_error(who, account, what);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_send_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_conv_im_send(im, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_send_confirm_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_conv_send_confirm(conv, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_im_send_with_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t im_ID;
	PurpleConvIm *im;
	const char *message;
	dbus_int32_t flags;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &im_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(im, im_ID, PurpleConvIm, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_conv_im_send_with_flags(im, message, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_custom_smiley_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *smile;
	const char *cksum_type;
	const char *chksum;
	dbus_int32_t remote;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_STRING, &smile, DBUS_TYPE_STRING, &cksum_type, DBUS_TYPE_STRING, &chksum, DBUS_TYPE_INT32, &remote, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	smile = (smile && smile[0]) ? smile : NULL;
	cksum_type = (cksum_type && cksum_type[0]) ? cksum_type : NULL;
	chksum = (chksum && chksum[0]) ? chksum : NULL;
	RESULT = purple_conv_custom_smiley_add(conv, smile, cksum_type, chksum, remote);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_custom_smiley_close_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	const char *smile;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_STRING, &smile, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	smile = (smile && smile[0]) ? smile : NULL;
	purple_conv_custom_smiley_close(conv, smile);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_conversation_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conv_chat_get_conversation(chat), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_set_users_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t users_NULL;
	GList *users;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &users_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	users = NULL;
	list = purple_conv_chat_set_users(chat, users);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_users_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	list = purple_conv_chat_get_users(chat);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_ignore_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_conv_chat_ignore(chat, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_unignore_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_conv_chat_unignore(chat, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_set_ignored_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t ignored_NULL;
	GList *ignored;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &ignored_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	ignored = NULL;
	list = purple_conv_chat_set_ignored(chat, ignored);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_ignored_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	list = purple_conv_chat_get_ignored(chat);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_ignored_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	if ((RESULT = purple_conv_chat_get_ignored_user(chat, user)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_is_user_ignored_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	RESULT = purple_conv_chat_is_user_ignored(chat, user);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_set_topic_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *who;
	const char *topic;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &topic, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	topic = (topic && topic[0]) ? topic : NULL;
	purple_conv_chat_set_topic(chat, who, topic);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_topic_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	if ((RESULT = purple_conv_chat_get_topic(chat)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_set_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	purple_conv_chat_set_id(chat, id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	RESULT = purple_conv_chat_get_id(chat);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_write_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *who;
	const char *message;
	dbus_int32_t flags;
	dbus_int32_t mtime;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INT32, &mtime, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_conv_chat_write(chat, who, message, flags, mtime);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_send_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_conv_chat_send(chat, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_send_with_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *message;
	dbus_int32_t flags;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_conv_chat_send_with_flags(chat, message, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_add_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	const char *extra_msg;
	dbus_int32_t flags;
	dbus_int32_t new_arrival;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_STRING, &extra_msg, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INT32, &new_arrival, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	extra_msg = (extra_msg && extra_msg[0]) ? extra_msg : NULL;
	purple_conv_chat_add_user(chat, user, extra_msg, flags, new_arrival);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_add_users_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t users_NULL;
	GList *users;
	dbus_int32_t extra_msgs_NULL;
	GList *extra_msgs;
	dbus_int32_t flags_NULL;
	GList *flags;
	dbus_int32_t new_arrivals;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &users_NULL, DBUS_TYPE_INT32, &extra_msgs_NULL, DBUS_TYPE_INT32, &flags_NULL, DBUS_TYPE_INT32, &new_arrivals, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	users = NULL;
	extra_msgs = NULL;
	flags = NULL;
	purple_conv_chat_add_users(chat, users, extra_msgs, flags, new_arrivals);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_rename_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *old_user;
	const char *new_user;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &old_user, DBUS_TYPE_STRING, &new_user, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	old_user = (old_user && old_user[0]) ? old_user : NULL;
	new_user = (new_user && new_user[0]) ? new_user : NULL;
	purple_conv_chat_rename_user(chat, old_user, new_user);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_remove_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	const char *reason;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_STRING, &reason, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	reason = (reason && reason[0]) ? reason : NULL;
	purple_conv_chat_remove_user(chat, user, reason);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_remove_users_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t users_NULL;
	GList *users;
	const char *reason;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INT32, &users_NULL, DBUS_TYPE_STRING, &reason, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	users = NULL;
	reason = (reason && reason[0]) ? reason : NULL;
	purple_conv_chat_remove_users(chat, users, reason);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_find_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	RESULT = purple_conv_chat_find_user(chat, user);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_user_set_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	dbus_int32_t flags;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	purple_conv_chat_user_set_flags(chat, user, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_user_get_flags_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	RESULT = purple_conv_chat_user_get_flags(chat, user);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_clear_users_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	purple_conv_chat_clear_users(chat);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_set_nick_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *nick;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &nick, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	nick = (nick && nick[0]) ? nick : NULL;
	purple_conv_chat_set_nick(chat, nick);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_get_nick_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	if ((RESULT = purple_conv_chat_get_nick(chat)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_find_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_find_chat(gc, id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_left_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	purple_conv_chat_left(chat);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_invite_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *user;
	const char *message;
	dbus_int32_t confirm;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &user, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &confirm, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	user = (user && user[0]) ? user : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_conv_chat_invite_user(chat, user, message, confirm);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_has_left_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	RESULT = purple_conv_chat_has_left(chat);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *alias;
	dbus_int32_t flags;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conv_chat_cb_new(name, alias, flags), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_find_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t chat_ID;
	PurpleConvChat *chat;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &chat_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(chat, chat_ID, PurpleConvChat, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_conv_chat_cb_find(chat, name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t cb_ID;
	PurpleConvChatBuddy *cb;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &cb_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(cb, cb_ID, PurpleConvChatBuddy, error_DBUS);
	if ((RESULT = purple_conv_chat_cb_get_name(cb)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conv_chat_cb_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t cb_ID;
	PurpleConvChatBuddy *cb;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &cb_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(cb, cb_ID, PurpleConvChatBuddy, error_DBUS);
	purple_conv_chat_cb_destroy(cb);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversation_get_extended_menu_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	list = purple_conversation_get_extended_menu(conv);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_conversations_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_conversations_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_conversations_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_conversations_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *ui;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &ui, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ui = (ui && ui[0]) ? ui : NULL;
	RESULT = purple_core_init(ui);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_quit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_core_quit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_get_version_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_core_get_version()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_get_ui_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_core_get_ui()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_get_core_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_get_core(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleCoreUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleCoreUiOps, error_DBUS);
	purple_core_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_core_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_migrate_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_core_migrate();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_core_ensure_single_instance_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_core_ensure_single_instance();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t type;
	const char *who;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &who, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_xfer_new(account, type, who), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfers_get_all_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_xfers_get_all();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_ref_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_ref(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_unref_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_unref(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_request_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_request(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_request_accepted_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *filename;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_STRING, &filename, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	filename = (filename && filename[0]) ? filename : NULL;
	purple_xfer_request_accepted(xfer, filename);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_request_denied_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_request_denied(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_type(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_xfer_get_account(xfer), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_remote_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	if ((RESULT = purple_xfer_get_remote_user(xfer)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_status(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_is_canceled_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_is_canceled(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_is_completed_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_is_completed(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_filename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	if ((RESULT = purple_xfer_get_filename(xfer)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_local_filename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	if ((RESULT = purple_xfer_get_local_filename(xfer)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_bytes_sent_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_bytes_sent(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_bytes_remaining_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_bytes_remaining(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_size_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_size(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_local_port_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_uint32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_local_port(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_UINT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_remote_ip_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	if ((RESULT = purple_xfer_get_remote_ip(xfer)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_remote_port_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_uint32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_remote_port(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_UINT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_start_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_start_time(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_end_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_end_time(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_set_completed_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t completed;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INT32, &completed, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_set_completed(xfer, completed);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_set_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_xfer_set_message(xfer, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_set_filename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *filename;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_STRING, &filename, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	filename = (filename && filename[0]) ? filename : NULL;
	purple_xfer_set_filename(xfer, filename);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_set_local_filename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *filename;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_STRING, &filename, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	filename = (filename && filename[0]) ? filename : NULL;
	purple_xfer_set_local_filename(xfer, filename);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_set_size_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t size;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INT32, &size, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_set_size(xfer, size);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_set_bytes_sent_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t bytes_sent;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INT32, &bytes_sent, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_set_bytes_sent(xfer, bytes_sent);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_xfer_get_ui_ops(xfer), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_start_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_int32_t fd;
	const char *ip;
	dbus_uint32_t port;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INT32, &fd, DBUS_TYPE_STRING, &ip, DBUS_TYPE_UINT32, &port, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	ip = (ip && ip[0]) ? ip : NULL;
	purple_xfer_start(xfer, fd, ip, port);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_end_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_end(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_add(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_cancel_local_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_cancel_local(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_cancel_remote_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_cancel_remote(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_error_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *who;
	const char *msg;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &msg, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	msg = (msg && msg[0]) ? msg : NULL;
	purple_xfer_error(type, account, who, msg);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_update_progress_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	purple_xfer_update_progress(xfer);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_thumbnail_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	size_t len = 0;
	gconstpointer RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	RESULT = purple_xfer_get_thumbnail(xfer, &len);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &RESULT, len, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_get_thumbnail_mimetype_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	if ((RESULT = purple_xfer_get_thumbnail_mimetype(xfer)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfer_prepare_thumbnail_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t xfer_ID;
	PurpleXfer *xfer;
	const char *formats;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &xfer_ID, DBUS_TYPE_STRING, &formats, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(xfer, xfer_ID, PurpleXfer, error_DBUS);
	formats = (formats && formats[0]) ? formats : NULL;
	purple_xfer_prepare_thumbnail(xfer, formats);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfers_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_xfers_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfers_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_xfers_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfers_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleXferUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleXferUiOps, error_DBUS);
	purple_xfers_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_xfers_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_xfers_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_free_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	purple_log_free(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_write_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t type;
	const char *from;
	dbus_int32_t time;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &from, DBUS_TYPE_INT32, &time, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	from = (from && from[0]) ? from : NULL;
	message = (message && message[0]) ? message : NULL;
	purple_log_write(log, type, from, time, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_get_logs_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	list = purple_log_get_logs(type, name, account);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_log_get_system_logs_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	list = purple_log_get_system_logs(account);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_log_get_size_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	RESULT = purple_log_get_size(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_get_total_size_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_log_get_total_size(type, name, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_get_activity_score_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	RESULT = purple_log_get_activity_score(type, name, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_is_deletable_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	RESULT = purple_log_is_deletable(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_delete_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	RESULT = purple_log_delete(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_get_log_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	if ((RESULT = purple_log_get_log_dir(type, name, account)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_log_set_free_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t set_ID;
	PurpleLogSet *set;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &set_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(set, set_ID, PurpleLogSet, error_DBUS);
	purple_log_set_free(set);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_common_writer_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	const char *ext;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_STRING, &ext, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	ext = (ext && ext[0]) ? ext : NULL;
	purple_log_common_writer(log, ext);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_common_lister_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ext;
	dbus_int32_t logger_ID;
	PurpleLogLogger *logger;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ext, DBUS_TYPE_INT32, &logger_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ext = (ext && ext[0]) ? ext : NULL;
	PURPLE_DBUS_ID_TO_POINTER(logger, logger_ID, PurpleLogLogger, error_DBUS);
	list = purple_log_common_lister(type, name, account, ext, logger);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_log_common_total_sizer_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *ext;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &ext, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	ext = (ext && ext[0]) ? ext : NULL;
	RESULT = purple_log_common_total_sizer(type, name, account, ext);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_common_sizer_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	RESULT = purple_log_common_sizer(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_common_deleter_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	RESULT = purple_log_common_deleter(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_common_is_deletable_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t log_ID;
	PurpleLog *log;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &log_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(log, log_ID, PurpleLog, error_DBUS);
	RESULT = purple_log_common_is_deletable(log);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_logger_free_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t logger_ID;
	PurpleLogLogger *logger;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &logger_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(logger, logger_ID, PurpleLogLogger, error_DBUS);
	purple_log_logger_free(logger);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_logger_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t logger_ID;
	PurpleLogLogger *logger;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &logger_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(logger, logger_ID, PurpleLogLogger, error_DBUS);
	purple_log_logger_add(logger);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_logger_remove_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t logger_ID;
	PurpleLogLogger *logger;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &logger_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(logger, logger_ID, PurpleLogLogger, error_DBUS);
	purple_log_logger_remove(logger);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_logger_set_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t logger_ID;
	PurpleLogLogger *logger;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &logger_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(logger, logger_ID, PurpleLogLogger, error_DBUS);
	purple_log_logger_set(logger);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_logger_get_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_log_logger_get(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_logger_get_options_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_log_logger_get_options();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_log_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_log_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_log_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_log_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_free_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	purple_notify_searchresults_free(results);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_new_rows_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_int32_t data_NULL;
	void *data;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_INT32, &data_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	data = NULL;
	purple_notify_searchresults_new_rows(gc, results, data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_notify_searchresults_new(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_column_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *title;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &title, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	title = (title && title[0]) ? title : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_notify_searchresults_column_new(title), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_column_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_int32_t column_ID;
	PurpleNotifySearchColumn *column;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_INT32, &column_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(column, column_ID, PurpleNotifySearchColumn, error_DBUS);
	purple_notify_searchresults_column_add(results, column);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_row_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_int32_t row_NULL;
	GList *row;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_INT32, &row_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	row = NULL;
	purple_notify_searchresults_row_add(results, row);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_get_rows_count_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	RESULT = purple_notify_searchresults_get_rows_count(results);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_get_columns_count_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	RESULT = purple_notify_searchresults_get_columns_count(results);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_row_get_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_uint32_t row_id;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_UINT32, &row_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	list = purple_notify_searchresults_row_get(results, row_id);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_searchresults_column_get_title_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t results_ID;
	PurpleNotifySearchResults *results;
	dbus_uint32_t column_id;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &results_ID, DBUS_TYPE_UINT32, &column_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(results, results_ID, PurpleNotifySearchResults, error_DBUS);
	if ((RESULT = purple_notify_searchresults_column_get_title(results, column_id)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_notify_user_info_new(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	purple_notify_user_info_destroy(user_info);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_get_entries_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	list = purple_notify_user_info_get_entries(user_info);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_get_text_with_newline_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	const char *newline;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_STRING, &newline, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	newline = (newline && newline[0]) ? newline : NULL;
	if ((RESULT = purple_notify_user_info_get_text_with_newline(user_info, newline)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_add_pair_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	const char *label;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_STRING, &label, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	label = (label && label[0]) ? label : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_notify_user_info_add_pair(user_info, label, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_add_pair_plaintext_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	const char *label;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_STRING, &label, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	label = (label && label[0]) ? label : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_notify_user_info_add_pair_plaintext(user_info, label, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_prepend_pair_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	const char *label;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_STRING, &label, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	label = (label && label[0]) ? label : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_notify_user_info_prepend_pair(user_info, label, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_remove_entry_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	purple_notify_user_info_remove_entry(user_info, user_info_entry);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *label;
	const char *value;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &label, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	label = (label && label[0]) ? label : NULL;
	value = (value && value[0]) ? value : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_notify_user_info_entry_new(label, value), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_add_section_break_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	purple_notify_user_info_add_section_break(user_info);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_prepend_section_break_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	purple_notify_user_info_prepend_section_break(user_info);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_add_section_header_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	const char *label;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_STRING, &label, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	label = (label && label[0]) ? label : NULL;
	purple_notify_user_info_add_section_header(user_info, label);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_prepend_section_header_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	const char *label;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_STRING, &label, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	label = (label && label[0]) ? label : NULL;
	purple_notify_user_info_prepend_section_header(user_info, label);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_remove_last_item_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_ID;
	PurpleNotifyUserInfo *user_info;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info, user_info_ID, PurpleNotifyUserInfo, error_DBUS);
	purple_notify_user_info_remove_last_item(user_info);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_get_label_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	if ((RESULT = purple_notify_user_info_entry_get_label(user_info_entry)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_set_label_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	const char *label;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_STRING, &label, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	label = (label && label[0]) ? label : NULL;
	purple_notify_user_info_entry_set_label(user_info_entry, label);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_get_value_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	if ((RESULT = purple_notify_user_info_entry_get_value(user_info_entry)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_set_value_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	value = (value && value[0]) ? value : NULL;
	purple_notify_user_info_entry_set_value(user_info_entry, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	RESULT = purple_notify_user_info_entry_get_type(user_info_entry);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_user_info_entry_set_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t user_info_entry_ID;
	PurpleNotifyUserInfoEntry *user_info_entry;
	dbus_int32_t type;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &user_info_entry_ID, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(user_info_entry, user_info_entry_ID, PurpleNotifyUserInfoEntry, error_DBUS);
	purple_notify_user_info_entry_set_type(user_info_entry, type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_close_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	dbus_int32_t ui_handle_NULL;
	void *ui_handle;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_INT32, &ui_handle_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ui_handle = NULL;
	purple_notify_close(type, ui_handle);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_close_with_handle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t handle_NULL;
	void *handle;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &handle_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	handle = NULL;
	purple_notify_close_with_handle(handle);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleNotifyUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleNotifyUiOps, error_DBUS);
	purple_notify_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_notify_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_notify_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_notify_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_notify_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurplePrefsUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurplePrefsUiOps, error_DBUS);
	purple_prefs_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_prefs_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_prefs_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_prefs_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_none_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_add_none(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_add_bool(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_add_int(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_prefs_add_string(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_string_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value_NULL;
	GList *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = NULL;
	purple_prefs_add_string_list(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_prefs_add_path(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_add_path_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value_NULL;
	GList *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = NULL;
	purple_prefs_add_path_list(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_remove_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_remove(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_rename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *oldname;
	const char *newname;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &oldname, DBUS_TYPE_STRING, &newname, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	oldname = (oldname && oldname[0]) ? oldname : NULL;
	newname = (newname && newname[0]) ? newname : NULL;
	purple_prefs_rename(oldname, newname);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_rename_boolean_toggle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *oldname;
	const char *newname;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &oldname, DBUS_TYPE_STRING, &newname, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	oldname = (oldname && oldname[0]) ? oldname : NULL;
	newname = (newname && newname[0]) ? newname : NULL;
	purple_prefs_rename_boolean_toggle(oldname, newname);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_prefs_destroy();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_set_bool(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_set_int(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_prefs_set_string(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_string_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value_NULL;
	GList *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = NULL;
	purple_prefs_set_string_list(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_prefs_set_path(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_set_path_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t value_NULL;
	GList *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	value = NULL;
	purple_prefs_set_path_list(name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_exists_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_prefs_exists(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_prefs_get_type(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_bool_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_prefs_get_bool(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = purple_prefs_get_int(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	if ((RESULT = purple_prefs_get_string(name)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_string_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT_LEN;
	GList *list;
	char **RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	list = purple_prefs_get_string_list(name);
	RESULT = (char **)purple_GList_to_array(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	g_list_foreach(list, (GFunc)g_free, NULL);
	g_list_free(list);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	if ((RESULT = purple_prefs_get_path(name)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_path_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT_LEN;
	GList *list;
	char **RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	list = purple_prefs_get_path_list(name);
	RESULT = (char **)purple_GList_to_array(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	g_list_foreach(list, (GFunc)g_free, NULL);
	g_list_free(list);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_get_children_names_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_int32_t RESULT_LEN;
	GList *list;
	char **RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	list = purple_prefs_get_children_names(name);
	RESULT = (char **)purple_GList_to_array(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	g_list_foreach(list, (GFunc)g_free, NULL);
	g_list_free(list);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_disconnect_callback_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t callback_id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &callback_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_prefs_disconnect_callback(callback_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_disconnect_by_handle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t handle_NULL;
	void *handle;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &handle_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	handle = NULL;
	purple_prefs_disconnect_by_handle(handle);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_trigger_callback_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prefs_trigger_callback(name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_trigger_callback_object_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t data_ID;
	PurplePrefCallbackData *data;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &data_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(data, data_ID, PurplePrefCallbackData, error_DBUS);
	purple_prefs_trigger_callback_object(data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_load_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_prefs_load();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prefs_update_old_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_prefs_update_old();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_show_with_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_roomlist_show_with_account(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_roomlist_new(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_ref_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	purple_roomlist_ref(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_unref_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	purple_roomlist_unref(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_set_fields_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_int32_t fields_NULL;
	GList *fields;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INT32, &fields_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	fields = NULL;
	purple_roomlist_set_fields(list, fields);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_set_in_progress_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_int32_t in_progress;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INT32, &in_progress, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	purple_roomlist_set_in_progress(list, in_progress);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_get_in_progress_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	RESULT = purple_roomlist_get_in_progress(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_add_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_int32_t room_ID;
	PurpleRoomlistRoom *room;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INT32, &room_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(room, room_ID, PurpleRoomlistRoom, error_DBUS);
	purple_roomlist_room_add(list, room);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_get_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_roomlist_get_list(gc), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_cancel_get_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	purple_roomlist_cancel_get_list(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_expand_category_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_int32_t category_ID;
	PurpleRoomlistRoom *category;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INT32, &category_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(category, category_ID, PurpleRoomlistRoom, error_DBUS);
	purple_roomlist_expand_category(list, category);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_get_fields_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t roomlist_ID;
	PurpleRoomlist *roomlist;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &roomlist_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(roomlist, roomlist_ID, PurpleRoomlist, error_DBUS);
	list = purple_roomlist_get_fields(roomlist);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *name;
	dbus_int32_t parent_ID;
	PurpleRoomlistRoom *parent;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &parent_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(parent, parent_ID, PurpleRoomlistRoom, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_roomlist_room_new(type, name, parent), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_join_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t list_ID;
	PurpleRoomlist *list;
	dbus_int32_t room_ID;
	PurpleRoomlistRoom *room;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &list_ID, DBUS_TYPE_INT32, &room_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(list, list_ID, PurpleRoomlist, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(room, room_ID, PurpleRoomlistRoom, error_DBUS);
	purple_roomlist_room_join(list, room);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t room_ID;
	PurpleRoomlistRoom *room;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &room_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(room, room_ID, PurpleRoomlistRoom, error_DBUS);
	RESULT = purple_roomlist_room_get_type(room);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t room_ID;
	PurpleRoomlistRoom *room;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &room_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(room, room_ID, PurpleRoomlistRoom, error_DBUS);
	if ((RESULT = purple_roomlist_room_get_name(room)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_get_parent_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t room_ID;
	PurpleRoomlistRoom *room;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &room_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(room, room_ID, PurpleRoomlistRoom, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_roomlist_room_get_parent(room), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_room_get_fields_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t room_ID;
	PurpleRoomlistRoom *room;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &room_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(room, room_ID, PurpleRoomlistRoom, error_DBUS);
	list = purple_roomlist_room_get_fields(room);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_field_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *label;
	const char *name;
	dbus_int32_t hidden;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &label, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &hidden, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	label = (label && label[0]) ? label : NULL;
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_roomlist_field_new(type, label, name, hidden), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_field_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t field_ID;
	PurpleRoomlistField *field;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &field_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(field, field_ID, PurpleRoomlistField, error_DBUS);
	RESULT = purple_roomlist_field_get_type(field);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_field_get_label_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t field_ID;
	PurpleRoomlistField *field;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &field_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(field, field_ID, PurpleRoomlistField, error_DBUS);
	if ((RESULT = purple_roomlist_field_get_label(field)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_field_get_hidden_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t field_ID;
	PurpleRoomlistField *field;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &field_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(field, field_ID, PurpleRoomlistField, error_DBUS);
	RESULT = purple_roomlist_field_get_hidden(field);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_set_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t ops_ID;
	PurpleRoomlistUiOps *ops;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &ops_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(ops, ops_ID, PurpleRoomlistUiOps, error_DBUS);
	purple_roomlist_set_ui_ops(ops);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_roomlist_get_ui_ops_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_roomlist_get_ui_ops(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *title;
	dbus_int32_t type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &title, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	title = (title && title[0]) ? title : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_new(title, type), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_set_title_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleSavedStatus *status;
	const char *title;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &title, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleSavedStatus, error_DBUS);
	title = (title && title[0]) ? title : NULL;
	purple_savedstatus_set_title(status, title);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_set_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleSavedStatus *status;
	dbus_int32_t type;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleSavedStatus, error_DBUS);
	purple_savedstatus_set_type(status, type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_set_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleSavedStatus *status;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleSavedStatus, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_savedstatus_set_message(status, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_set_substatus_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleSavedStatus *status;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t type_ID;
	PurpleStatusType *type;
	const char *message;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleSavedStatus, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleStatusType, error_DBUS);
	message = (message && message[0]) ? message : NULL;
	purple_savedstatus_set_substatus(status, account, type, message);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_unset_substatus_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_savedstatus_unset_substatus(saved_status, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_delete_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *title;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &title, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	title = (title && title[0]) ? title : NULL;
	RESULT = purple_savedstatus_delete(title);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_delete_by_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	purple_savedstatus_delete_by_status(saved_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatuses_get_all_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_savedstatuses_get_all();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatuses_get_popular_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_uint32_t how_many;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_UINT32, &how_many, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_savedstatuses_get_popular(how_many);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_current_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_get_current(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_default_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_get_default(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_idleaway_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_get_idleaway(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_is_idleaway_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_savedstatus_is_idleaway();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_set_idleaway_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t idleaway;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &idleaway, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_savedstatus_set_idleaway(idleaway);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_startup_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_get_startup(), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_find_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *title;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &title, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	title = (title && title[0]) ? title : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_find(title), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_find_by_creation_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t creation_time;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &creation_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_find_by_creation_time(creation_time), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_find_transient_by_type_and_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *message;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_STRING, &message, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	message = (message && message[0]) ? message : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_find_transient_by_type_and_message(type, message), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_is_transient_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	RESULT = purple_savedstatus_is_transient(saved_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_title_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	if ((RESULT = purple_savedstatus_get_title(saved_status)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	RESULT = purple_savedstatus_get_type(saved_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	if ((RESULT = purple_savedstatus_get_message(saved_status)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_creation_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	RESULT = purple_savedstatus_get_creation_time(saved_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_has_substatuses_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	RESULT = purple_savedstatus_has_substatuses(saved_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_get_substatus_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_get_substatus(saved_status, account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_substatus_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t substatus_ID;
	PurpleSavedStatusSub *substatus;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &substatus_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(substatus, substatus_ID, PurpleSavedStatusSub, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_savedstatus_substatus_get_type(substatus), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_substatus_get_message_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t substatus_ID;
	PurpleSavedStatusSub *substatus;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &substatus_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(substatus, substatus_ID, PurpleSavedStatusSub, error_DBUS);
	if ((RESULT = purple_savedstatus_substatus_get_message(substatus)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_activate_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	purple_savedstatus_activate(saved_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatus_activate_for_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t saved_status_ID;
	PurpleSavedStatus *saved_status;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &saved_status_ID, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(saved_status, saved_status_ID, PurpleSavedStatus, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_savedstatus_activate_for_account(saved_status, account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatuses_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_savedstatuses_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_savedstatuses_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_savedstatuses_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t img_ID;
	PurpleStoredImage *img;
	const char *shortcut;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &img_ID, DBUS_TYPE_STRING, &shortcut, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(img, img_ID, PurpleStoredImage, error_DBUS);
	shortcut = (shortcut && shortcut[0]) ? shortcut : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_smiley_new(img, shortcut), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_new_from_file_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *shortcut;
	const char *filepath;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &shortcut, DBUS_TYPE_STRING, &filepath, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	shortcut = (shortcut && shortcut[0]) ? shortcut : NULL;
	filepath = (filepath && filepath[0]) ? filepath : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_smiley_new_from_file(shortcut, filepath), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_delete_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	purple_smiley_delete(smiley);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_set_shortcut_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	const char *shortcut;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_STRING, &shortcut, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	shortcut = (shortcut && shortcut[0]) ? shortcut : NULL;
	RESULT = purple_smiley_set_shortcut(smiley, shortcut);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_set_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	dbus_int32_t smiley_data_NULL;
	guchar *smiley_data;
	dbus_int32_t smiley_data_len;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INT32, &smiley_data_NULL, DBUS_TYPE_INT32, &smiley_data_len, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	smiley_data = NULL;
	purple_smiley_set_data(smiley, smiley_data, smiley_data_len);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_get_shortcut_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	if ((RESULT = purple_smiley_get_shortcut(smiley)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_get_checksum_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	if ((RESULT = purple_smiley_get_checksum(smiley)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_get_stored_image_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_smiley_get_stored_image(smiley), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_get_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	size_t len = 0;
	gconstpointer RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	RESULT = purple_smiley_get_data(smiley, &len);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &RESULT, len, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_get_extension_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	if ((RESULT = purple_smiley_get_extension(smiley)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smiley_get_full_path_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t smiley_ID;
	PurpleSmiley *smiley;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &smiley_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(smiley, smiley_ID, PurpleSmiley, error_DBUS);
	if ((RESULT = purple_smiley_get_full_path(smiley)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_smileys_get_all_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	list = purple_smileys_get_all();
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_smileys_find_by_shortcut_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *shortcut;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &shortcut, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	shortcut = (shortcut && shortcut[0]) ? shortcut : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_smileys_find_by_shortcut(shortcut), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smileys_find_by_checksum_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *checksum;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &checksum, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	checksum = (checksum && checksum[0]) ? checksum : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_smileys_find_by_checksum(checksum), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smileys_get_storing_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_smileys_get_storing_dir()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smileys_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_smileys_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_smileys_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_smileys_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_primitive_get_id_from_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_primitive_get_id_from_type(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_primitive_get_name_from_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_primitive_get_name_from_type(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_primitive_get_type_from_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	id = (id && id[0]) ? id : NULL;
	RESULT = purple_primitive_get_type_from_id(id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_new_full_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t primitive;
	const char *id;
	const char *name;
	dbus_int32_t saveable;
	dbus_int32_t user_settable;
	dbus_int32_t independent;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &primitive, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &saveable, DBUS_TYPE_INT32, &user_settable, DBUS_TYPE_INT32, &independent, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	id = (id && id[0]) ? id : NULL;
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_type_new_full(primitive, id, name, saveable, user_settable, independent), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t primitive;
	const char *id;
	const char *name;
	dbus_int32_t user_settable;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &primitive, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &user_settable, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	id = (id && id[0]) ? id : NULL;
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_type_new(primitive, id, name, user_settable), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	purple_status_type_destroy(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_set_primary_attr_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	const char *attr_id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_STRING, &attr_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	attr_id = (attr_id && attr_id[0]) ? attr_id : NULL;
	purple_status_type_set_primary_attr(status_type, attr_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_add_attr_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	const char *id;
	const char *name;
	dbus_int32_t value_ID;
	PurpleValue *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(value, value_ID, PurpleValue, error_DBUS);
	purple_status_type_add_attr(status_type, id, name, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_get_primitive_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	RESULT = purple_status_type_get_primitive(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_get_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	if ((RESULT = purple_status_type_get_id(status_type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	if ((RESULT = purple_status_type_get_name(status_type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_is_saveable_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	RESULT = purple_status_type_is_saveable(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_is_user_settable_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	RESULT = purple_status_type_is_user_settable(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_is_independent_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	RESULT = purple_status_type_is_independent(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_is_exclusive_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	RESULT = purple_status_type_is_exclusive(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_is_available_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	RESULT = purple_status_type_is_available(status_type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_get_primary_attr_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleStatusType *type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleStatusType, error_DBUS);
	if ((RESULT = purple_status_type_get_primary_attr(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_get_attr_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_type_get_attr(status_type, id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_get_attrs_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	list = purple_status_type_get_attrs(status_type);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_status_type_find_with_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_types_NULL;
	GList *status_types;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_types_NULL, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	status_types = NULL;
	id = (id && id[0]) ? id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_type_find_with_id(status_types, id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_attr_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *id;
	const char *name;
	dbus_int32_t value_type_ID;
	PurpleValue *value_type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &value_type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	id = (id && id[0]) ? id : NULL;
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_ID_TO_POINTER(value_type, value_type_ID, PurpleValue, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_attr_new(id, name, value_type), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_attr_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t attr_ID;
	PurpleStatusAttr *attr;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &attr_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(attr, attr_ID, PurpleStatusAttr, error_DBUS);
	purple_status_attr_destroy(attr);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_attr_get_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t attr_ID;
	PurpleStatusAttr *attr;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &attr_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(attr, attr_ID, PurpleStatusAttr, error_DBUS);
	if ((RESULT = purple_status_attr_get_id(attr)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_attr_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t attr_ID;
	PurpleStatusAttr *attr;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &attr_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(attr, attr_ID, PurpleStatusAttr, error_DBUS);
	if ((RESULT = purple_status_attr_get_name(attr)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_attr_get_value_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t attr_ID;
	PurpleStatusAttr *attr;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &attr_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(attr, attr_ID, PurpleStatusAttr, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_attr_get_value(attr), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_type_ID;
	PurpleStatusType *status_type;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_type_ID, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status_type, status_type_ID, PurpleStatusType, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_new(status_type, presence), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	purple_status_destroy(status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_set_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t active;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INT32, &active, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	purple_status_set_active(status, active);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_set_active_with_attrs_list_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t active;
	dbus_int32_t attrs_NULL;
	GList *attrs;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INT32, &active, DBUS_TYPE_INT32, &attrs_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	attrs = NULL;
	purple_status_set_active_with_attrs_list(status, active, attrs);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_set_attr_boolean_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	purple_status_set_attr_boolean(status, id, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_set_attr_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	dbus_int32_t value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INT32, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	purple_status_set_attr_int(status, id, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_set_attr_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	value = (value && value[0]) ? value : NULL;
	purple_status_set_attr_string(status, id, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_type_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_get_type(status), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_presence_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_get_presence(status), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_id_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	if ((RESULT = purple_status_get_id(status)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	if ((RESULT = purple_status_get_name(status)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_is_independent_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	RESULT = purple_status_is_independent(status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_is_exclusive_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	RESULT = purple_status_is_exclusive(status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_is_available_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	RESULT = purple_status_is_available(status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_is_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	RESULT = purple_status_is_active(status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_is_online_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	RESULT = purple_status_is_online(status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_attr_value_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_status_get_attr_value(status, id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_attr_boolean_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	RESULT = purple_status_get_attr_boolean(status, id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_attr_int_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	RESULT = purple_status_get_attr_int(status, id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_get_attr_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	const char *id;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	id = (id && id[0]) ? id : NULL;
	if ((RESULT = purple_status_get_attr_string(status, id)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_compare_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t status1_ID;
	PurpleStatus *status1;
	dbus_int32_t status2_ID;
	PurpleStatus *status2;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &status1_ID, DBUS_TYPE_INT32, &status2_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status1, status1_ID, PurpleStatus, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status2, status2_ID, PurpleStatus, error_DBUS);
	RESULT = purple_status_compare(status1, status2);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t context;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &context, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_new(context), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_new_for_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_new_for_account(account), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_new_for_conv_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t conv_ID;
	PurpleConversation *conv;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &conv_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(conv, conv_ID, PurpleConversation, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_new_for_conv(conv), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_new_for_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t buddy_ID;
	PurpleBuddy *buddy;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &buddy_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(buddy, buddy_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_new_for_buddy(buddy), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_destroy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	purple_presence_destroy(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_add_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t status_ID;
	PurpleStatus *status;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INT32, &status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(status, status_ID, PurpleStatus, error_DBUS);
	purple_presence_add_status(presence, status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_set_status_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	const char *status_id;
	dbus_int32_t active;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INT32, &active, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	purple_presence_set_status_active(presence, status_id, active);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_switch_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	const char *status_id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	purple_presence_switch_status(presence, status_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_set_idle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t idle;
	dbus_int32_t idle_time;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INT32, &idle, DBUS_TYPE_INT32, &idle_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	purple_presence_set_idle(presence, idle, idle_time);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_set_login_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t login_time;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INT32, &login_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	purple_presence_set_login_time(presence, login_time);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_context_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_get_context(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_account_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_get_account(presence), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_conversation_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_get_conversation(presence), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_chat_user_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	if ((RESULT = purple_presence_get_chat_user(presence)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_get_buddy(presence), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_statuses_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	list = purple_presence_get_statuses(presence);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	const char *status_id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_get_status(presence, status_id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_active_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_presence_get_active_status(presence), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_is_available_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_is_available(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_is_online_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_is_online(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_is_status_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	const char *status_id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	RESULT = purple_presence_is_status_active(presence, status_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_is_status_primitive_active_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t primitive;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INT32, &primitive, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_is_status_primitive_active(presence, primitive);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_is_idle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_is_idle(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_idle_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_get_idle_time(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_get_login_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_get_login_time(presence);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_presence_compare_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t presence1_ID;
	PurplePresence *presence1;
	dbus_int32_t presence2_ID;
	PurplePresence *presence2;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &presence1_ID, DBUS_TYPE_INT32, &presence2_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence1, presence1_ID, PurplePresence, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence2, presence2_ID, PurplePresence, error_DBUS);
	RESULT = purple_presence_compare(presence1, presence2);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_status_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_status_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_status_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_send_typing_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *name;
	dbus_int32_t state;
	dbus_uint32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	RESULT = serv_send_typing(gc, name, state);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_UINT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_move_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleBuddy *param0;
	dbus_int32_t param1_ID;
	PurpleGroup *param1;
	dbus_int32_t param2_ID;
	PurpleGroup *param2;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INT32, &param1_ID, DBUS_TYPE_INT32, &param2_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleBuddy, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param1, param1_ID, PurpleGroup, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param2, param2_ID, PurpleGroup, error_DBUS);
	serv_move_buddy(param0, param1, param2);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_send_im_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	const char *param2;
	dbus_int32_t flags;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_STRING, &param2, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	param2 = (param2 && param2[0]) ? param2 : NULL;
	RESULT = serv_send_im(param0, param1, param2, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_get_attention_type_from_code_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t type_code;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &type_code, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_get_attention_type_from_code(account, type_code), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_send_attention_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	dbus_int32_t type_code;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &type_code, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	serv_send_attention(gc, who, type_code);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_got_attention_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	dbus_int32_t type_code;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &type_code, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	serv_got_attention(gc, who, type_code);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_get_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	serv_get_info(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_set_info_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	serv_set_info(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_add_permit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	serv_add_permit(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_add_deny_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	serv_add_deny(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_rem_permit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	serv_rem_permit(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_rem_deny_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	const char *param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_STRING, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param1 = (param1 && param1[0]) ? param1 : NULL;
	serv_rem_deny(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_set_permit_deny_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	serv_set_permit_deny(param0);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_chat_invite_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	dbus_int32_t param1;
	const char *param2;
	const char *param3;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INT32, &param1, DBUS_TYPE_STRING, &param2, DBUS_TYPE_STRING, &param3, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param2 = (param2 && param2[0]) ? param2 : NULL;
	param3 = (param3 && param3[0]) ? param3 : NULL;
	serv_chat_invite(param0, param1, param2, param3);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_chat_leave_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	dbus_int32_t param1;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INT32, &param1, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	serv_chat_leave(param0, param1);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_chat_whisper_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	dbus_int32_t param1;
	const char *param2;
	const char *param3;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INT32, &param1, DBUS_TYPE_STRING, &param2, DBUS_TYPE_STRING, &param3, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param2 = (param2 && param2[0]) ? param2 : NULL;
	param3 = (param3 && param3[0]) ? param3 : NULL;
	serv_chat_whisper(param0, param1, param2, param3);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_chat_send_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	dbus_int32_t param1;
	const char *param2;
	dbus_int32_t flags;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INT32, &param1, DBUS_TYPE_STRING, &param2, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	param2 = (param2 && param2[0]) ? param2 : NULL;
	RESULT = serv_chat_send(param0, param1, param2, flags);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_alias_buddy_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleBuddy *param0;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleBuddy, error_DBUS);
	serv_alias_buddy(param0);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_got_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	serv_got_alias(gc, who, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_serv_got_private_alias_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	const char *alias;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &alias, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	alias = (alias && alias[0]) ? alias : NULL;
	purple_serv_got_private_alias(gc, who, alias);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_got_typing_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *name;
	dbus_int32_t timeout;
	dbus_int32_t state;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &timeout, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	serv_got_typing(gc, name, timeout, state);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_got_typing_stopped_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	serv_got_typing_stopped(gc, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_got_im_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	const char *msg;
	dbus_int32_t flags;
	dbus_int32_t mtime;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &msg, DBUS_TYPE_INT32, &flags, DBUS_TYPE_INT32, &mtime, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	msg = (msg && msg[0]) ? msg : NULL;
	serv_got_im(gc, who, msg, flags, mtime);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_join_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	DBusMessageIter data_ITER;
	GHashTable *data;
	purple_dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_ARRAY, &data_ITER, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	data = purple_dbus_iter_hash_table(&data_ITER, error_DBUS);
	CHECK_ERROR(error_DBUS);
	serv_join_chat(param0, data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	g_hash_table_destroy(data);
	return reply_DBUS;
}

static DBusMessage*
serv_reject_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t param0_ID;
	PurpleConnection *param0;
	DBusMessageIter data_ITER;
	GHashTable *data;
	purple_dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &param0_ID, DBUS_TYPE_ARRAY, &data_ITER, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(param0, param0_ID, PurpleConnection, error_DBUS);
	data = purple_dbus_iter_hash_table(&data_ITER, error_DBUS);
	CHECK_ERROR(error_DBUS);
	serv_reject_chat(param0, data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	g_hash_table_destroy(data);
	return reply_DBUS;
}

static DBusMessage*
serv_got_chat_invite_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *name;
	const char *who;
	const char *message;
	DBusMessageIter data_ITER;
	GHashTable *data;
	purple_dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &message, DBUS_TYPE_ARRAY, &data_ITER, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	who = (who && who[0]) ? who : NULL;
	message = (message && message[0]) ? message : NULL;
	data = purple_dbus_iter_hash_table(&data_ITER, error_DBUS);
	CHECK_ERROR(error_DBUS);
	serv_got_chat_invite(gc, name, who, message, data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	g_hash_table_destroy(data);
	return reply_DBUS;
}

static DBusMessage*
serv_got_joined_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t id;
	const char *name;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, serv_got_joined_chat(gc, id, name), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_serv_got_join_chat_failed_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	DBusMessageIter data_ITER;
	GHashTable *data;
	purple_dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_ARRAY, &data_ITER, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	data = purple_dbus_iter_hash_table(&data_ITER, error_DBUS);
	CHECK_ERROR(error_DBUS);
	purple_serv_got_join_chat_failed(gc, data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	g_hash_table_destroy(data);
	return reply_DBUS;
}

static DBusMessage*
serv_got_chat_left_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t g_ID;
	PurpleConnection *g;
	dbus_int32_t id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &g_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(g, g_ID, PurpleConnection, error_DBUS);
	serv_got_chat_left(g, id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_got_chat_in_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t g_ID;
	PurpleConnection *g;
	dbus_int32_t id;
	const char *who;
	dbus_int32_t flags;
	const char *message;
	dbus_int32_t mtime;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &g_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &flags, DBUS_TYPE_STRING, &message, DBUS_TYPE_INT32, &mtime, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(g, g_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	message = (message && message[0]) ? message : NULL;
	serv_got_chat_in(g, id, who, flags, message, mtime);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_send_file_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	const char *file;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_STRING, &file, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	file = (file && file[0]) ? file : NULL;
	serv_send_file(gc, who, file);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
serv_chat_send_file_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t id;
	const char *file;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_STRING, &file, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	file = (file && file[0]) ? file : NULL;
	serv_chat_send_file(gc, id, file);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_menu_action_free_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t act_ID;
	PurpleMenuAction *act;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &act_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(act, act_ID, PurpleMenuAction, error_DBUS);
	purple_menu_action_free(act);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_set_current_song_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *title;
	const char *artist;
	const char *album;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &title, DBUS_TYPE_STRING, &artist, DBUS_TYPE_STRING, &album, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	title = (title && title[0]) ? title : NULL;
	artist = (artist && artist[0]) ? artist : NULL;
	album = (album && album[0]) ? album : NULL;
	purple_util_set_current_song(title, artist, album);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_init_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_util_init();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_uninit_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_util_uninit();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_mime_decode_field_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_mime_decode_field(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_time_build_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t year;
	dbus_int32_t month;
	dbus_int32_t day;
	dbus_int32_t hour;
	dbus_int32_t min;
	dbus_int32_t sec;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &year, DBUS_TYPE_INT32, &month, DBUS_TYPE_INT32, &day, DBUS_TYPE_INT32, &hour, DBUS_TYPE_INT32, &min, DBUS_TYPE_INT32, &sec, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_time_build(year, month, day, hour, min, sec);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_escape_text_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *text;
	dbus_int32_t length;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &text, DBUS_TYPE_INT32, &length, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	text = (text && text[0]) ? text : NULL;
	if ((RESULT = purple_markup_escape_text(text, length)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_strip_html_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_markup_strip_html(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_linkify_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_markup_linkify(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_unescape_text_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *text;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &text, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	text = (text && text[0]) ? text : NULL;
	if ((RESULT = purple_unescape_text(text)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_unescape_html_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *html;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &html, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	html = (html && html[0]) ? html : NULL;
	if ((RESULT = purple_unescape_html(html)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_slice_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	dbus_int32_t x;
	dbus_int32_t y;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INT32, &x, DBUS_TYPE_INT32, &y, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_markup_slice(str, x, y)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_get_tag_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *tag;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &tag, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	tag = (tag && tag[0]) ? tag : NULL;
	if ((RESULT = purple_markup_get_tag_name(tag)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_unescape_entity_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *text;
	dbus_int32_t length_NULL;
	int *length;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &text, DBUS_TYPE_INT32, &length_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	text = (text && text[0]) ? text : NULL;
	length = NULL;
	if ((RESULT = purple_markup_unescape_entity(text, length)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_get_css_property_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *style;
	const char *opt;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &style, DBUS_TYPE_STRING, &opt, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	style = (style && style[0]) ? style : NULL;
	opt = (opt && opt[0]) ? opt : NULL;
	if ((RESULT = purple_markup_get_css_property(style, opt)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_markup_is_rtl_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *html;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &html, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	html = (html && html[0]) ? html : NULL;
	RESULT = purple_markup_is_rtl(html);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_home_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_home_dir()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_user_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_user_dir()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_set_user_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *dir;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &dir, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	dir = (dir && dir[0]) ? dir : NULL;
	purple_util_set_user_dir(dir);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_build_dir_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *path;
	dbus_int32_t mode;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &path, DBUS_TYPE_INT32, &mode, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	path = (path && path[0]) ? path : NULL;
	RESULT = purple_build_dir(path, mode);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_write_data_to_file_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *filename;
	const char *data;
	dbus_int32_t size;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &filename, DBUS_TYPE_STRING, &data, DBUS_TYPE_INT32, &size, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	filename = (filename && filename[0]) ? filename : NULL;
	data = (data && data[0]) ? data : NULL;
	RESULT = purple_util_write_data_to_file(filename, data, size);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_write_data_to_file_absolute_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *filename_full;
	const char *data;
	dbus_int32_t size;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &filename_full, DBUS_TYPE_STRING, &data, DBUS_TYPE_INT32, &size, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	filename_full = (filename_full && filename_full[0]) ? filename_full : NULL;
	data = (data && data[0]) ? data : NULL;
	RESULT = purple_util_write_data_to_file_absolute(filename_full, data, size);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_program_is_valid_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *program;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &program, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	program = (program && program[0]) ? program : NULL;
	RESULT = purple_program_is_valid(program);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_running_gnome_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_running_gnome();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_running_kde_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_running_kde();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_running_osx_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_running_osx();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_fd_get_ip_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t fd;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &fd, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_fd_get_ip(fd)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_socket_get_family_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t fd;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &fd, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_socket_get_family(fd);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_socket_speaks_ipv4_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t fd;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &fd, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	RESULT = purple_socket_speaks_ipv4(fd);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_strequal_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *left;
	const char *right;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &left, DBUS_TYPE_STRING, &right, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	left = (left && left[0]) ? left : NULL;
	right = (right && right[0]) ? right : NULL;
	RESULT = purple_strequal(left, right);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_normalize_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *str;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_normalize(account, str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_normalize_nocase_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *str;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_normalize_nocase(account, str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_str_has_prefix_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *s;
	const char *p;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &s, DBUS_TYPE_STRING, &p, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	s = (s && s[0]) ? s : NULL;
	p = (p && p[0]) ? p : NULL;
	RESULT = purple_str_has_prefix(s, p);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_str_has_suffix_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *s;
	const char *x;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &s, DBUS_TYPE_STRING, &x, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	s = (s && s[0]) ? s : NULL;
	x = (x && x[0]) ? x : NULL;
	RESULT = purple_str_has_suffix(s, x);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_strdup_withhtml_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *src;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &src, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	src = (src && src[0]) ? src : NULL;
	if ((RESULT = purple_strdup_withhtml(src)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_str_add_cr_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_str_add_cr(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_strreplace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *string;
	const char *delimiter;
	const char *replacement;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &string, DBUS_TYPE_STRING, &delimiter, DBUS_TYPE_STRING, &replacement, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	string = (string && string[0]) ? string : NULL;
	delimiter = (delimiter && delimiter[0]) ? delimiter : NULL;
	replacement = (replacement && replacement[0]) ? replacement : NULL;
	if ((RESULT = purple_strreplace(string, delimiter, replacement)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_ncr_encode_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *in;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &in, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	in = (in && in[0]) ? in : NULL;
	if ((RESULT = purple_utf8_ncr_encode(in)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_ncr_decode_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *in;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &in, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	in = (in && in[0]) ? in : NULL;
	if ((RESULT = purple_utf8_ncr_decode(in)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_strcasereplace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *string;
	const char *delimiter;
	const char *replacement;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &string, DBUS_TYPE_STRING, &delimiter, DBUS_TYPE_STRING, &replacement, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	string = (string && string[0]) ? string : NULL;
	delimiter = (delimiter && delimiter[0]) ? delimiter : NULL;
	replacement = (replacement && replacement[0]) ? replacement : NULL;
	if ((RESULT = purple_strcasereplace(string, delimiter, replacement)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_strcasestr_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *haystack;
	const char *needle;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &haystack, DBUS_TYPE_STRING, &needle, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	haystack = (haystack && haystack[0]) ? haystack : NULL;
	needle = (needle && needle[0]) ? needle : NULL;
	if ((RESULT = purple_strcasestr(haystack, needle)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_str_size_to_units_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t size;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &size, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_str_size_to_units(size)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_str_seconds_to_string_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t sec;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &sec, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_str_seconds_to_string(sec)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_str_binary_to_ascii_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const unsigned char *binary;
	dbus_int32_t len;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &binary, DBUS_TYPE_INT32, &len, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	binary = (binary && binary[0]) ? binary : NULL;
	if ((RESULT = purple_str_binary_to_ascii(binary, len)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_got_protocol_handler_uri_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *uri;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &uri, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	uri = (uri && uri[0]) ? uri : NULL;
	purple_got_protocol_handler_uri(uri);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_util_fetch_url_cancel_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t url_data_ID;
	PurpleUtilFetchUrlData *url_data;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &url_data_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(url_data, url_data_ID, PurpleUtilFetchUrlData, error_DBUS);
	purple_util_fetch_url_cancel(url_data);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_url_decode_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_url_decode(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_url_encode_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_url_encode(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_email_is_valid_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *address;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &address, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	address = (address && address[0]) ? address : NULL;
	RESULT = purple_email_is_valid(address);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_ip_address_is_valid_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *ip;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &ip, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ip = (ip && ip[0]) ? ip : NULL;
	RESULT = purple_ip_address_is_valid(ip);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_ipv4_address_is_valid_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *ip;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &ip, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ip = (ip && ip[0]) ? ip : NULL;
	RESULT = purple_ipv4_address_is_valid(ip);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_ipv6_address_is_valid_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *ip;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &ip, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ip = (ip && ip[0]) ? ip : NULL;
	RESULT = purple_ipv6_address_is_valid(ip);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_uri_list_extract_uris_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *uri_list;
	dbus_int32_t RESULT_LEN;
	GList *list;
	char **RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &uri_list, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	uri_list = (uri_list && uri_list[0]) ? uri_list : NULL;
	list = purple_uri_list_extract_uris(uri_list);
	RESULT = (char **)purple_GList_to_array(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	g_list_foreach(list, (GFunc)g_free, NULL);
	g_list_free(list);
	return reply_DBUS;
}

static DBusMessage*
purple_uri_list_extract_filenames_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *uri_list;
	dbus_int32_t RESULT_LEN;
	GList *list;
	char **RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &uri_list, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	uri_list = (uri_list && uri_list[0]) ? uri_list : NULL;
	list = purple_uri_list_extract_filenames(uri_list);
	RESULT = (char **)purple_GList_to_array(list, FALSE, &RESULT_LEN);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	g_list_foreach(list, (GFunc)g_free, NULL);
	g_list_free(list);
	return reply_DBUS;
}

static DBusMessage*
purple_uri_escape_for_open_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *unescaped;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &unescaped, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	unescaped = (unescaped && unescaped[0]) ? unescaped : NULL;
	if ((RESULT = purple_uri_escape_for_open(unescaped)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_try_convert_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_utf8_try_convert(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_salvage_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_utf8_salvage(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_strip_unprintables_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_utf8_strip_unprintables(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_gai_strerror_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t errnum;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &errnum, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_gai_strerror(errnum)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_strcasecmp_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *a;
	const char *b;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &a, DBUS_TYPE_STRING, &b, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	a = (a && a[0]) ? a : NULL;
	b = (b && b[0]) ? b : NULL;
	RESULT = purple_utf8_strcasecmp(a, b);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_utf8_has_word_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *haystack;
	const char *needle;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &haystack, DBUS_TYPE_STRING, &needle, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	haystack = (haystack && haystack[0]) ? haystack : NULL;
	needle = (needle && needle[0]) ? needle : NULL;
	RESULT = purple_utf8_has_word(haystack, needle);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_text_strip_mnemonic_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *in;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &in, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	in = (in && in[0]) ? in : NULL;
	if ((RESULT = purple_text_strip_mnemonic(in)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_unescape_filename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_unescape_filename(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_escape_filename_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *str;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	str = (str && str[0]) ? str : NULL;
	if ((RESULT = purple_escape_filename(str)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
_purple_oscar_convert_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *act;
	const char *protocol;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &act, DBUS_TYPE_STRING, &protocol, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	act = (act && act[0]) ? act : NULL;
	protocol = (protocol && protocol[0]) ? protocol : NULL;
	if ((RESULT = _purple_oscar_convert(act, protocol)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_restore_default_signal_handlers_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	purple_restore_default_signal_handlers();
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_get_host_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_get_host_name()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_uuid_random_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	if ((RESULT = purple_uuid_random()) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_insert_child_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t parent_ID;
	xmlnode *parent;
	dbus_int32_t child_ID;
	xmlnode *child;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &parent_ID, DBUS_TYPE_INT32, &child_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(parent, parent_ID, xmlnode, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(child, child_ID, xmlnode, error_DBUS);
	xmlnode_insert_child(parent, child);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_insert_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *data;
	dbus_int32_t size;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &data, DBUS_TYPE_INT32, &size, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	data = (data && data[0]) ? data : NULL;
	xmlnode_insert_data(node, data, size);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_get_data_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	if ((RESULT = xmlnode_get_data(node)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_get_data_unescaped_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	if ((RESULT = xmlnode_get_data_unescaped(node)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_set_attrib_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	value = (value && value[0]) ? value : NULL;
	xmlnode_set_attrib(node, attr, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_set_attrib_with_prefix_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *prefix;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_STRING, &prefix, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	prefix = (prefix && prefix[0]) ? prefix : NULL;
	value = (value && value[0]) ? value : NULL;
	xmlnode_set_attrib_with_prefix(node, attr, prefix, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_set_attrib_with_namespace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *xmlns;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_STRING, &xmlns, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	xmlns = (xmlns && xmlns[0]) ? xmlns : NULL;
	value = (value && value[0]) ? value : NULL;
	xmlnode_set_attrib_with_namespace(node, attr, xmlns, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_set_attrib_full_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *xmlns;
	const char *prefix;
	const char *value;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_STRING, &xmlns, DBUS_TYPE_STRING, &prefix, DBUS_TYPE_STRING, &value, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	xmlns = (xmlns && xmlns[0]) ? xmlns : NULL;
	prefix = (prefix && prefix[0]) ? prefix : NULL;
	value = (value && value[0]) ? value : NULL;
	xmlnode_set_attrib_full(node, attr, xmlns, prefix, value);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_get_attrib_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	if ((RESULT = xmlnode_get_attrib(node, attr)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_get_attrib_with_namespace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *xmlns;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_STRING, &xmlns, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	xmlns = (xmlns && xmlns[0]) ? xmlns : NULL;
	if ((RESULT = xmlnode_get_attrib_with_namespace(node, attr, xmlns)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_remove_attrib_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	xmlnode_remove_attrib(node, attr);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_remove_attrib_with_namespace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *attr;
	const char *xmlns;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &attr, DBUS_TYPE_STRING, &xmlns, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	attr = (attr && attr[0]) ? attr : NULL;
	xmlns = (xmlns && xmlns[0]) ? xmlns : NULL;
	xmlnode_remove_attrib_with_namespace(node, attr, xmlns);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_set_namespace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *xmlns;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &xmlns, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	xmlns = (xmlns && xmlns[0]) ? xmlns : NULL;
	xmlnode_set_namespace(node, xmlns);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_get_namespace_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	if ((RESULT = xmlnode_get_namespace(node)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_set_prefix_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *prefix;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_STRING, &prefix, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	prefix = (prefix && prefix[0]) ? prefix : NULL;
	xmlnode_set_prefix(node, prefix);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_get_prefix_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	if ((RESULT = xmlnode_get_prefix(node)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_to_str_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	dbus_int32_t len_NULL;
	int *len;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INT32, &len_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	len = NULL;
	if ((RESULT = xmlnode_to_str(node, len)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_to_formatted_str_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	dbus_int32_t len_NULL;
	int *len;
	char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INT32, &len_NULL, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	len = NULL;
	if ((RESULT = xmlnode_to_formatted_str(node, len)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
xmlnode_free_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t node_ID;
	xmlnode *node;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &node_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(node, node_ID, xmlnode, error_DBUS);
	xmlnode_free(node);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_new_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *ulname;
	const char *name;
	const char *inc_desc;
	const char *out_desc;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &ulname, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &inc_desc, DBUS_TYPE_STRING, &out_desc, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	ulname = (ulname && ulname[0]) ? ulname : NULL;
	name = (name && name[0]) ? name : NULL;
	inc_desc = (inc_desc && inc_desc[0]) ? inc_desc : NULL;
	out_desc = (out_desc && out_desc[0]) ? out_desc : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_attention_type_new(ulname, name, inc_desc, out_desc), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_set_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_attention_type_set_name(type, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_set_incoming_desc_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *desc;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_STRING, &desc, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	desc = (desc && desc[0]) ? desc : NULL;
	purple_attention_type_set_incoming_desc(type, desc);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_set_outgoing_desc_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *desc;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_STRING, &desc, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	desc = (desc && desc[0]) ? desc : NULL;
	purple_attention_type_set_outgoing_desc(type, desc);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_set_icon_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *name;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_attention_type_set_icon_name(type, name);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_set_unlocalized_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *ulname;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_STRING, &ulname, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	ulname = (ulname && ulname[0]) ? ulname : NULL;
	purple_attention_type_set_unlocalized_name(type, ulname);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_get_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	if ((RESULT = purple_attention_type_get_name(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_get_incoming_desc_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	if ((RESULT = purple_attention_type_get_incoming_desc(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_get_outgoing_desc_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	if ((RESULT = purple_attention_type_get_outgoing_desc(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_get_icon_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	if ((RESULT = purple_attention_type_get_icon_name(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_attention_type_get_unlocalized_name_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t type_ID;
	PurpleAttentionType *type;
	const char *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &type_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(type, type_ID, PurpleAttentionType, error_DBUS);
	if ((RESULT = purple_attention_type_get_unlocalized_name(type)) == NULL)
		RESULT = "";
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_STRING, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_account_idle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t idle;
	dbus_int32_t idle_time;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &idle, DBUS_TYPE_INT32, &idle_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_prpl_got_account_idle(account, idle, idle_time);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_account_login_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t login_time;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &login_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_prpl_got_account_login_time(account, login_time);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_account_actions_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	purple_prpl_got_account_actions(account);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_user_idle_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t idle;
	dbus_int32_t idle_time;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &idle, DBUS_TYPE_INT32, &idle_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prpl_got_user_idle(account, name, idle, idle_time);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_user_login_time_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	dbus_int32_t login_time;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_INT32, &login_time, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	purple_prpl_got_user_login_time(account, name, login_time);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_user_status_deactive_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *name;
	const char *status_id;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &status_id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	name = (name && name[0]) ? name : NULL;
	status_id = (status_id && status_id[0]) ? status_id : NULL;
	purple_prpl_got_user_status_deactive(account, name, status_id);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_change_account_status_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t old_status_ID;
	PurpleStatus *old_status;
	dbus_int32_t new_status_ID;
	PurpleStatus *new_status;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &old_status_ID, DBUS_TYPE_INT32, &new_status_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(old_status, old_status_ID, PurpleStatus, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(new_status, new_status_ID, PurpleStatus, error_DBUS);
	purple_prpl_change_account_status(account, old_status, new_status);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_get_statuses_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	dbus_int32_t presence_ID;
	PurplePresence *presence;
	dbus_int32_t RESULT_LEN;
	GList *list;
	dbus_int32_t *RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_INT32, &presence_ID, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(presence, presence_ID, PurplePresence, error_DBUS);
	list = purple_prpl_get_statuses(account, presence);
	RESULT = purple_dbusify_GList(list, FALSE, &RESULT_LEN);
	g_list_free(list);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_ARRAY, DBUS_TYPE_INT32, &RESULT, RESULT_LEN, DBUS_TYPE_INVALID);
	g_free(RESULT);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_send_attention_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	dbus_int32_t type_code;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &type_code, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	purple_prpl_send_attention(gc, who, type_code);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_attention_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	const char *who;
	dbus_int32_t type_code;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &type_code, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	purple_prpl_got_attention(gc, who, type_code);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_attention_in_chat_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t gc_ID;
	PurpleConnection *gc;
	dbus_int32_t id;
	const char *who;
	dbus_int32_t type_code;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &gc_ID, DBUS_TYPE_INT32, &id, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &type_code, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(gc, gc_ID, PurpleConnection, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	purple_prpl_got_attention_in_chat(gc, id, who, type_code);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_get_media_caps_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *who;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	RESULT = purple_prpl_get_media_caps(account, who);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_initiate_media_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *who;
	dbus_int32_t type;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	RESULT = purple_prpl_initiate_media(account, who, type);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_prpl_got_media_caps_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	dbus_int32_t account_ID;
	PurpleAccount *account;
	const char *who;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_INT32, &account_ID, DBUS_TYPE_STRING, &who, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	PURPLE_DBUS_ID_TO_POINTER(account, account_ID, PurpleAccount, error_DBUS);
	who = (who && who[0]) ? who : NULL;
	purple_prpl_got_media_caps(account, who);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static DBusMessage*
purple_find_prpl_DBUS(DBusMessage *message_DBUS, DBusError *error_DBUS) {
	DBusMessage *reply_DBUS;
	const char *id;
	dbus_int32_t RESULT;
	dbus_message_get_args(message_DBUS, error_DBUS, DBUS_TYPE_STRING, &id, DBUS_TYPE_INVALID);
	CHECK_ERROR(error_DBUS);
	id = (id && id[0]) ? id : NULL;
	PURPLE_DBUS_POINTER_TO_ID(RESULT, purple_find_prpl(id), error_DBUS);
	reply_DBUS = dbus_message_new_method_return (message_DBUS);
	dbus_message_append_args(reply_DBUS, DBUS_TYPE_INT32, &RESULT, DBUS_TYPE_INVALID);
	return reply_DBUS;
}

static PurpleDBusBinding bindings_DBUS[] = { 
{"PurpleAccountsFindAny", "in\0s\0name\0in\0s\0protocol\0out\0i\0RESULT\0", purple_accounts_find_any_DBUS},
{"PurpleAccountsFindConnected", "in\0s\0name\0in\0s\0protocol\0out\0i\0RESULT\0", purple_accounts_find_connected_DBUS},
{"PurpleBlistNodeIsChat", "in\0i\0node\0out\0i\0RESULT\0", PURPLE_BLIST_NODE_IS_CHAT_DBUS},
{"PurpleBlistNodeIsBuddy", "in\0i\0node\0out\0i\0RESULT\0", PURPLE_BLIST_NODE_IS_BUDDY_DBUS},
{"PurpleBlistNodeIsContact", "in\0i\0node\0out\0i\0RESULT\0", PURPLE_BLIST_NODE_IS_CONTACT_DBUS},
{"PurpleBlistNodeIsGroup", "in\0i\0node\0out\0i\0RESULT\0", PURPLE_BLIST_NODE_IS_GROUP_DBUS},
{"PurpleBuddyIsOnline", "in\0i\0buddy\0out\0i\0RESULT\0", PURPLE_BUDDY_IS_ONLINE_DBUS},
{"PurpleBlistNodeHasFlag", "in\0i\0node\0in\0i\0flags\0out\0i\0RESULT\0", PURPLE_BLIST_NODE_HAS_FLAG_DBUS},
{"PurpleBlistNodeShouldSave", "in\0i\0node\0out\0i\0RESULT\0", PURPLE_BLIST_NODE_SHOULD_SAVE_DBUS},
{"PurpleConnectionIsConnected", "in\0i\0connection\0out\0i\0RESULT\0", PURPLE_CONNECTION_IS_CONNECTED_DBUS},
{"PurpleConnectionIsValid", "in\0i\0connection\0out\0i\0RESULT\0", PURPLE_CONNECTION_IS_VALID_DBUS},
{"PurpleConvIm", "in\0i\0conversation\0out\0i\0RESULT\0", PURPLE_CONV_IM_DBUS},
{"PurpleConvChat", "in\0i\0conversation\0out\0i\0RESULT\0", PURPLE_CONV_CHAT_DBUS},
{"PurpleAccountNew", "in\0s\0username\0in\0s\0protocol_id\0out\0i\0RESULT\0", purple_account_new_DBUS},
{"PurpleAccountDestroy", "in\0i\0account\0", purple_account_destroy_DBUS},
{"PurpleAccountConnect", "in\0i\0account\0", purple_account_connect_DBUS},
{"PurpleAccountRegister", "in\0i\0account\0", purple_account_register_DBUS},
{"PurpleAccountDisconnect", "in\0i\0account\0", purple_account_disconnect_DBUS},
{"PurpleAccountNotifyAdded", "in\0i\0account\0in\0s\0remote_user\0in\0s\0id\0in\0s\0alias\0in\0s\0message\0", purple_account_notify_added_DBUS},
{"PurpleAccountRequestAdd", "in\0i\0account\0in\0s\0remote_user\0in\0s\0id\0in\0s\0alias\0in\0s\0message\0", purple_account_request_add_DBUS},
{"PurpleAccountRequestCloseWithAccount", "in\0i\0account\0", purple_account_request_close_with_account_DBUS},
{"PurpleAccountRequestClose", "in\0i\0ui_handle\0", purple_account_request_close_DBUS},
{"PurpleAccountRequestChangePassword", "in\0i\0account\0", purple_account_request_change_password_DBUS},
{"PurpleAccountRequestChangeUserInfo", "in\0i\0account\0", purple_account_request_change_user_info_DBUS},
{"PurpleAccountSetUsername", "in\0i\0account\0in\0s\0username\0", purple_account_set_username_DBUS},
{"PurpleAccountSetPassword", "in\0i\0account\0in\0s\0password\0", purple_account_set_password_DBUS},
{"PurpleAccountSetAlias", "in\0i\0account\0in\0s\0alias\0", purple_account_set_alias_DBUS},
{"PurpleAccountSetUserInfo", "in\0i\0account\0in\0s\0user_info\0", purple_account_set_user_info_DBUS},
{"PurpleAccountSetBuddyIconPath", "in\0i\0account\0in\0s\0path\0", purple_account_set_buddy_icon_path_DBUS},
{"PurpleAccountSetProtocolId", "in\0i\0account\0in\0s\0protocol_id\0", purple_account_set_protocol_id_DBUS},
{"PurpleAccountSetConnection", "in\0i\0account\0in\0i\0gc\0", purple_account_set_connection_DBUS},
{"PurpleAccountSetRememberPassword", "in\0i\0account\0in\0i\0value\0", purple_account_set_remember_password_DBUS},
{"PurpleAccountSetCheckMail", "in\0i\0account\0in\0i\0value\0", purple_account_set_check_mail_DBUS},
{"PurpleAccountSetEnabled", "in\0i\0account\0in\0s\0ui\0in\0i\0value\0", purple_account_set_enabled_DBUS},
{"PurpleAccountSetProxyInfo", "in\0i\0account\0in\0i\0info\0", purple_account_set_proxy_info_DBUS},
{"PurpleAccountSetPrivacyType", "in\0i\0account\0in\0i\0privacy_type\0", purple_account_set_privacy_type_DBUS},
{"PurpleAccountSetStatusTypes", "in\0i\0account\0in\0i\0status_types\0", purple_account_set_status_types_DBUS},
{"PurpleAccountSetStatusList", "in\0i\0account\0in\0s\0status_id\0in\0i\0active\0in\0i\0attrs\0", purple_account_set_status_list_DBUS},
{"PurpleAccountGetSilenceSuppression", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_silence_suppression_DBUS},
{"PurpleAccountSetSilenceSuppression", "in\0i\0account\0in\0i\0value\0", purple_account_set_silence_suppression_DBUS},
{"PurpleAccountClearSettings", "in\0i\0account\0", purple_account_clear_settings_DBUS},
{"PurpleAccountRemoveSetting", "in\0i\0account\0in\0s\0setting\0", purple_account_remove_setting_DBUS},
{"PurpleAccountSetInt", "in\0i\0account\0in\0s\0name\0in\0i\0value\0", purple_account_set_int_DBUS},
{"PurpleAccountSetString", "in\0i\0account\0in\0s\0name\0in\0s\0value\0", purple_account_set_string_DBUS},
{"PurpleAccountSetBool", "in\0i\0account\0in\0s\0name\0in\0i\0value\0", purple_account_set_bool_DBUS},
{"PurpleAccountSetUiInt", "in\0i\0account\0in\0s\0ui\0in\0s\0name\0in\0i\0value\0", purple_account_set_ui_int_DBUS},
{"PurpleAccountSetUiString", "in\0i\0account\0in\0s\0ui\0in\0s\0name\0in\0s\0value\0", purple_account_set_ui_string_DBUS},
{"PurpleAccountSetUiBool", "in\0i\0account\0in\0s\0ui\0in\0s\0name\0in\0i\0value\0", purple_account_set_ui_bool_DBUS},
{"PurpleAccountIsConnected", "in\0i\0account\0out\0i\0RESULT\0", purple_account_is_connected_DBUS},
{"PurpleAccountIsConnecting", "in\0i\0account\0out\0i\0RESULT\0", purple_account_is_connecting_DBUS},
{"PurpleAccountIsDisconnected", "in\0i\0account\0out\0i\0RESULT\0", purple_account_is_disconnected_DBUS},
{"PurpleAccountGetUsername", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_username_DBUS},
{"PurpleAccountGetPassword", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_password_DBUS},
{"PurpleAccountGetAlias", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_alias_DBUS},
{"PurpleAccountGetUserInfo", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_user_info_DBUS},
{"PurpleAccountGetBuddyIconPath", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_buddy_icon_path_DBUS},
{"PurpleAccountGetProtocolId", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_protocol_id_DBUS},
{"PurpleAccountGetProtocolName", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_protocol_name_DBUS},
{"PurpleAccountGetConnection", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_connection_DBUS},
{"PurpleAccountGetNameForDisplay", "in\0i\0account\0out\0s\0RESULT\0", purple_account_get_name_for_display_DBUS},
{"PurpleAccountGetRememberPassword", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_remember_password_DBUS},
{"PurpleAccountGetCheckMail", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_check_mail_DBUS},
{"PurpleAccountGetEnabled", "in\0i\0account\0in\0s\0ui\0out\0i\0RESULT\0", purple_account_get_enabled_DBUS},
{"PurpleAccountGetProxyInfo", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_proxy_info_DBUS},
{"PurpleAccountGetPrivacyType", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_privacy_type_DBUS},
{"PurpleAccountGetActiveStatus", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_active_status_DBUS},
{"PurpleAccountGetStatus", "in\0i\0account\0in\0s\0status_id\0out\0i\0RESULT\0", purple_account_get_status_DBUS},
{"PurpleAccountGetStatusType", "in\0i\0account\0in\0s\0id\0out\0i\0RESULT\0", purple_account_get_status_type_DBUS},
{"PurpleAccountGetStatusTypeWithPrimitive", "in\0i\0account\0in\0i\0primitive\0out\0i\0RESULT\0", purple_account_get_status_type_with_primitive_DBUS},
{"PurpleAccountGetPresence", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_presence_DBUS},
{"PurpleAccountIsStatusActive", "in\0i\0account\0in\0s\0status_id\0out\0i\0RESULT\0", purple_account_is_status_active_DBUS},
{"PurpleAccountGetStatusTypes", "in\0i\0account\0out\0ai\0RESULT\0", purple_account_get_status_types_DBUS},
{"PurpleAccountGetInt", "in\0i\0account\0in\0s\0name\0in\0i\0default_value\0out\0i\0RESULT\0", purple_account_get_int_DBUS},
{"PurpleAccountGetString", "in\0i\0account\0in\0s\0name\0in\0s\0default_value\0out\0s\0RESULT\0", purple_account_get_string_DBUS},
{"PurpleAccountGetBool", "in\0i\0account\0in\0s\0name\0in\0i\0default_value\0out\0i\0RESULT\0", purple_account_get_bool_DBUS},
{"PurpleAccountGetUiInt", "in\0i\0account\0in\0s\0ui\0in\0s\0name\0in\0i\0default_value\0out\0i\0RESULT\0", purple_account_get_ui_int_DBUS},
{"PurpleAccountGetUiString", "in\0i\0account\0in\0s\0ui\0in\0s\0name\0in\0s\0default_value\0out\0s\0RESULT\0", purple_account_get_ui_string_DBUS},
{"PurpleAccountGetUiBool", "in\0i\0account\0in\0s\0ui\0in\0s\0name\0in\0i\0default_value\0out\0i\0RESULT\0", purple_account_get_ui_bool_DBUS},
{"PurpleAccountGetLog", "in\0i\0account\0in\0i\0create\0out\0i\0RESULT\0", purple_account_get_log_DBUS},
{"PurpleAccountDestroyLog", "in\0i\0account\0", purple_account_destroy_log_DBUS},
{"PurpleAccountAddBuddy", "in\0i\0account\0in\0i\0buddy\0", purple_account_add_buddy_DBUS},
{"PurpleAccountAddBuddyWithInvite", "in\0i\0account\0in\0i\0buddy\0in\0s\0message\0", purple_account_add_buddy_with_invite_DBUS},
{"PurpleAccountAddBuddies", "in\0i\0account\0in\0i\0buddies\0", purple_account_add_buddies_DBUS},
{"PurpleAccountAddBuddiesWithInvite", "in\0i\0account\0in\0i\0buddies\0in\0s\0message\0", purple_account_add_buddies_with_invite_DBUS},
{"PurpleAccountRemoveBuddy", "in\0i\0account\0in\0i\0buddy\0in\0i\0group\0", purple_account_remove_buddy_DBUS},
{"PurpleAccountRemoveBuddies", "in\0i\0account\0in\0i\0buddies\0in\0i\0groups\0", purple_account_remove_buddies_DBUS},
{"PurpleAccountRemoveGroup", "in\0i\0account\0in\0i\0group\0", purple_account_remove_group_DBUS},
{"PurpleAccountChangePassword", "in\0i\0account\0in\0s\0orig_pw\0in\0s\0new_pw\0", purple_account_change_password_DBUS},
{"PurpleAccountSupportsOfflineMessage", "in\0i\0account\0in\0i\0buddy\0out\0i\0RESULT\0", purple_account_supports_offline_message_DBUS},
{"PurpleAccountGetCurrentError", "in\0i\0account\0out\0i\0RESULT\0", purple_account_get_current_error_DBUS},
{"PurpleAccountClearCurrentError", "in\0i\0account\0", purple_account_clear_current_error_DBUS},
{"PurpleAccountsAdd", "in\0i\0account\0", purple_accounts_add_DBUS},
{"PurpleAccountsRemove", "in\0i\0account\0", purple_accounts_remove_DBUS},
{"PurpleAccountsDelete", "in\0i\0account\0", purple_accounts_delete_DBUS},
{"PurpleAccountsReorder", "in\0i\0account\0in\0i\0new_index\0", purple_accounts_reorder_DBUS},
{"PurpleAccountsGetAll", "out\0ai\0RESULT\0", purple_accounts_get_all_DBUS},
{"PurpleAccountsGetAllActive", "out\0ai\0RESULT\0", purple_accounts_get_all_active_DBUS},
{"PurpleAccountsFind", "in\0s\0name\0in\0s\0protocol\0out\0i\0RESULT\0", purple_accounts_find_DBUS},
{"PurpleAccountsRestoreCurrentStatuses", "", purple_accounts_restore_current_statuses_DBUS},
{"PurpleAccountsSetUiOps", "in\0i\0ops\0", purple_accounts_set_ui_ops_DBUS},
{"PurpleAccountsGetUiOps", "out\0i\0RESULT\0", purple_accounts_get_ui_ops_DBUS},
{"PurpleAccountPrefsSetUiOps", "in\0i\0ops\0", purple_account_prefs_set_ui_ops_DBUS},
{"PurpleAccountPrefsGetUiOps", "out\0i\0RESULT\0", purple_account_prefs_get_ui_ops_DBUS},
{"PurpleAccountsInit", "", purple_accounts_init_DBUS},
{"PurpleAccountsUninit", "", purple_accounts_uninit_DBUS},
{"PurpleBlistNew", "out\0i\0RESULT\0", purple_blist_new_DBUS},
{"PurpleSetBlist", "in\0i\0blist\0", purple_set_blist_DBUS},
{"PurpleGetBlist", "out\0i\0RESULT\0", purple_get_blist_DBUS},
{"PurpleBlistGetRoot", "out\0i\0RESULT\0", purple_blist_get_root_DBUS},
{"PurpleBlistGetBuddies", "out\0ai\0RESULT\0", purple_blist_get_buddies_DBUS},
{"PurpleBlistNodeNext", "in\0i\0node\0in\0i\0offline\0out\0i\0RESULT\0", purple_blist_node_next_DBUS},
{"PurpleBlistNodeGetParent", "in\0i\0node\0out\0i\0RESULT\0", purple_blist_node_get_parent_DBUS},
{"PurpleBlistNodeGetFirstChild", "in\0i\0node\0out\0i\0RESULT\0", purple_blist_node_get_first_child_DBUS},
{"PurpleBlistNodeGetSiblingNext", "in\0i\0node\0out\0i\0RESULT\0", purple_blist_node_get_sibling_next_DBUS},
{"PurpleBlistNodeGetSiblingPrev", "in\0i\0node\0out\0i\0RESULT\0", purple_blist_node_get_sibling_prev_DBUS},
{"PurpleBlistShow", "", purple_blist_show_DBUS},
{"PurpleBlistDestroy", "", purple_blist_destroy_DBUS},
{"PurpleBlistSetVisible", "in\0i\0show\0", purple_blist_set_visible_DBUS},
{"PurpleBlistUpdateBuddyStatus", "in\0i\0buddy\0in\0i\0old_status\0", purple_blist_update_buddy_status_DBUS},
{"PurpleBlistUpdateNodeIcon", "in\0i\0node\0", purple_blist_update_node_icon_DBUS},
{"PurpleBlistUpdateBuddyIcon", "in\0i\0buddy\0", purple_blist_update_buddy_icon_DBUS},
{"PurpleBlistRenameBuddy", "in\0i\0buddy\0in\0s\0name\0", purple_blist_rename_buddy_DBUS},
{"PurpleBlistAliasContact", "in\0i\0contact\0in\0s\0alias\0", purple_blist_alias_contact_DBUS},
{"PurpleBlistAliasBuddy", "in\0i\0buddy\0in\0s\0alias\0", purple_blist_alias_buddy_DBUS},
{"PurpleBlistServerAliasBuddy", "in\0i\0buddy\0in\0s\0alias\0", purple_blist_server_alias_buddy_DBUS},
{"PurpleBlistAliasChat", "in\0i\0chat\0in\0s\0alias\0", purple_blist_alias_chat_DBUS},
{"PurpleBlistRenameGroup", "in\0i\0group\0in\0s\0name\0", purple_blist_rename_group_DBUS},
{"PurpleChatNew", "in\0i\0account\0in\0s\0alias\0in\0a{ss}\0components\0out\0i\0RESULT\0", purple_chat_new_DBUS},
{"PurpleChatDestroy", "in\0i\0chat\0", purple_chat_destroy_DBUS},
{"PurpleBlistAddChat", "in\0i\0chat\0in\0i\0group\0in\0i\0node\0", purple_blist_add_chat_DBUS},
{"PurpleBuddyNew", "in\0i\0account\0in\0s\0name\0in\0s\0alias\0out\0i\0RESULT\0", purple_buddy_new_DBUS},
{"PurpleBuddyDestroy", "in\0i\0buddy\0", purple_buddy_destroy_DBUS},
{"PurpleBuddySetIcon", "in\0i\0buddy\0in\0i\0icon\0", purple_buddy_set_icon_DBUS},
{"PurpleBuddyGetAccount", "in\0i\0buddy\0out\0i\0RESULT\0", purple_buddy_get_account_DBUS},
{"PurpleBuddyGetName", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_name_DBUS},
{"PurpleBuddyGetIcon", "in\0i\0buddy\0out\0i\0RESULT\0", purple_buddy_get_icon_DBUS},
{"PurpleBuddyGetContact", "in\0i\0buddy\0out\0i\0RESULT\0", purple_buddy_get_contact_DBUS},
{"PurpleBuddyGetPresence", "in\0i\0buddy\0out\0i\0RESULT\0", purple_buddy_get_presence_DBUS},
{"PurpleBuddyGetMediaCaps", "in\0i\0buddy\0out\0i\0RESULT\0", purple_buddy_get_media_caps_DBUS},
{"PurpleBuddySetMediaCaps", "in\0i\0buddy\0in\0i\0media_caps\0", purple_buddy_set_media_caps_DBUS},
{"PurpleBlistAddBuddy", "in\0i\0buddy\0in\0i\0contact\0in\0i\0group\0in\0i\0node\0", purple_blist_add_buddy_DBUS},
{"PurpleGroupNew", "in\0s\0name\0out\0i\0RESULT\0", purple_group_new_DBUS},
{"PurpleGroupDestroy", "in\0i\0group\0", purple_group_destroy_DBUS},
{"PurpleBlistAddGroup", "in\0i\0group\0in\0i\0node\0", purple_blist_add_group_DBUS},
{"PurpleContactNew", "out\0i\0RESULT\0", purple_contact_new_DBUS},
{"PurpleContactDestroy", "in\0i\0contact\0", purple_contact_destroy_DBUS},
{"PurpleContactGetGroup", "in\0i\0contact\0out\0i\0RESULT\0", purple_contact_get_group_DBUS},
{"PurpleBlistAddContact", "in\0i\0contact\0in\0i\0group\0in\0i\0node\0", purple_blist_add_contact_DBUS},
{"PurpleBlistMergeContact", "in\0i\0source\0in\0i\0node\0", purple_blist_merge_contact_DBUS},
{"PurpleContactGetPriorityBuddy", "in\0i\0contact\0out\0i\0RESULT\0", purple_contact_get_priority_buddy_DBUS},
{"PurpleContactSetAlias", "in\0i\0contact\0in\0s\0alias\0", purple_contact_set_alias_DBUS},
{"PurpleContactGetAlias", "in\0i\0contact\0out\0s\0RESULT\0", purple_contact_get_alias_DBUS},
{"PurpleContactOnAccount", "in\0i\0contact\0in\0i\0account\0out\0i\0RESULT\0", purple_contact_on_account_DBUS},
{"PurpleContactInvalidatePriorityBuddy", "in\0i\0contact\0", purple_contact_invalidate_priority_buddy_DBUS},
{"PurpleBlistRemoveBuddy", "in\0i\0buddy\0", purple_blist_remove_buddy_DBUS},
{"PurpleBlistRemoveContact", "in\0i\0contact\0", purple_blist_remove_contact_DBUS},
{"PurpleBlistRemoveChat", "in\0i\0chat\0", purple_blist_remove_chat_DBUS},
{"PurpleBlistRemoveGroup", "in\0i\0group\0", purple_blist_remove_group_DBUS},
{"PurpleBuddyGetAliasOnly", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_alias_only_DBUS},
{"PurpleBuddyGetServerAlias", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_server_alias_DBUS},
{"PurpleBuddyGetContactAlias", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_contact_alias_DBUS},
{"PurpleBuddyGetLocalAlias", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_local_alias_DBUS},
{"PurpleBuddyGetAlias", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_alias_DBUS},
{"PurpleBuddyGetLocalBuddyAlias", "in\0i\0buddy\0out\0s\0RESULT\0", purple_buddy_get_local_buddy_alias_DBUS},
{"PurpleChatGetName", "in\0i\0chat\0out\0s\0RESULT\0", purple_chat_get_name_DBUS},
{"PurpleFindBuddy", "in\0i\0account\0in\0s\0name\0out\0i\0RESULT\0", purple_find_buddy_DBUS},
{"PurpleFindBuddyInGroup", "in\0i\0account\0in\0s\0name\0in\0i\0group\0out\0i\0RESULT\0", purple_find_buddy_in_group_DBUS},
{"PurpleFindBuddies", "in\0i\0account\0in\0s\0name\0out\0ai\0RESULT\0", purple_find_buddies_DBUS},
{"PurpleFindGroup", "in\0s\0name\0out\0i\0RESULT\0", purple_find_group_DBUS},
{"PurpleBlistFindChat", "in\0i\0account\0in\0s\0name\0out\0i\0RESULT\0", purple_blist_find_chat_DBUS},
{"PurpleChatGetGroup", "in\0i\0chat\0out\0i\0RESULT\0", purple_chat_get_group_DBUS},
{"PurpleChatGetAccount", "in\0i\0chat\0out\0i\0RESULT\0", purple_chat_get_account_DBUS},
{"PurpleBuddyGetGroup", "in\0i\0buddy\0out\0i\0RESULT\0", purple_buddy_get_group_DBUS},
{"PurpleGroupGetAccounts", "in\0i\0g\0out\0ai\0RESULT\0", purple_group_get_accounts_DBUS},
{"PurpleGroupOnAccount", "in\0i\0g\0in\0i\0account\0out\0i\0RESULT\0", purple_group_on_account_DBUS},
{"PurpleGroupGetName", "in\0i\0group\0out\0s\0RESULT\0", purple_group_get_name_DBUS},
{"PurpleBlistAddAccount", "in\0i\0account\0", purple_blist_add_account_DBUS},
{"PurpleBlistRemoveAccount", "in\0i\0account\0", purple_blist_remove_account_DBUS},
{"PurpleBlistGetGroupSize", "in\0i\0group\0in\0i\0offline\0out\0i\0RESULT\0", purple_blist_get_group_size_DBUS},
{"PurpleBlistGetGroupOnlineCount", "in\0i\0group\0out\0i\0RESULT\0", purple_blist_get_group_online_count_DBUS},
{"PurpleBlistLoad", "", purple_blist_load_DBUS},
{"PurpleBlistScheduleSave", "", purple_blist_schedule_save_DBUS},
{"PurpleBlistRequestAddBuddy", "in\0i\0account\0in\0s\0username\0in\0s\0group\0in\0s\0alias\0", purple_blist_request_add_buddy_DBUS},
{"PurpleBlistRequestAddChat", "in\0i\0account\0in\0i\0group\0in\0s\0alias\0in\0s\0name\0", purple_blist_request_add_chat_DBUS},
{"PurpleBlistRequestAddGroup", "", purple_blist_request_add_group_DBUS},
{"PurpleBlistNodeSetBool", "in\0i\0node\0in\0s\0key\0in\0i\0value\0", purple_blist_node_set_bool_DBUS},
{"PurpleBlistNodeGetBool", "in\0i\0node\0in\0s\0key\0out\0i\0RESULT\0", purple_blist_node_get_bool_DBUS},
{"PurpleBlistNodeSetInt", "in\0i\0node\0in\0s\0key\0in\0i\0value\0", purple_blist_node_set_int_DBUS},
{"PurpleBlistNodeGetInt", "in\0i\0node\0in\0s\0key\0out\0i\0RESULT\0", purple_blist_node_get_int_DBUS},
{"PurpleBlistNodeSetString", "in\0i\0node\0in\0s\0key\0in\0s\0value\0", purple_blist_node_set_string_DBUS},
{"PurpleBlistNodeGetString", "in\0i\0node\0in\0s\0key\0out\0s\0RESULT\0", purple_blist_node_get_string_DBUS},
{"PurpleBlistNodeRemoveSetting", "in\0i\0node\0in\0s\0key\0", purple_blist_node_remove_setting_DBUS},
{"PurpleBlistNodeSetFlags", "in\0i\0node\0in\0i\0flags\0", purple_blist_node_set_flags_DBUS},
{"PurpleBlistNodeGetFlags", "in\0i\0node\0out\0i\0RESULT\0", purple_blist_node_get_flags_DBUS},
{"PurpleBlistNodeGetType", "in\0i\0node\0out\0i\0RESULT\0", purple_blist_node_get_type_DBUS},
{"PurpleBlistNodeGetExtendedMenu", "in\0i\0n\0out\0ai\0RESULT\0", purple_blist_node_get_extended_menu_DBUS},
{"PurpleBlistSetUiOps", "in\0i\0ops\0", purple_blist_set_ui_ops_DBUS},
{"PurpleBlistGetUiOps", "out\0i\0RESULT\0", purple_blist_get_ui_ops_DBUS},
{"PurpleBlistInit", "", purple_blist_init_DBUS},
{"PurpleBlistUninit", "", purple_blist_uninit_DBUS},
{"PurpleBuddyIconNew", "in\0i\0account\0in\0s\0username\0in\0i\0icon_data\0in\0i\0icon_len\0in\0s\0checksum\0out\0i\0RESULT\0", purple_buddy_icon_new_DBUS},
{"PurpleBuddyIconRef", "in\0i\0icon\0out\0i\0RESULT\0", purple_buddy_icon_ref_DBUS},
{"PurpleBuddyIconUnref", "in\0i\0icon\0out\0i\0RESULT\0", purple_buddy_icon_unref_DBUS},
{"PurpleBuddyIconUpdate", "in\0i\0icon\0", purple_buddy_icon_update_DBUS},
{"PurpleBuddyIconSetData", "in\0i\0icon\0in\0i\0data\0in\0i\0len\0in\0s\0checksum\0", purple_buddy_icon_set_data_DBUS},
{"PurpleBuddyIconGetAccount", "in\0i\0icon\0out\0i\0RESULT\0", purple_buddy_icon_get_account_DBUS},
{"PurpleBuddyIconGetUsername", "in\0i\0icon\0out\0s\0RESULT\0", purple_buddy_icon_get_username_DBUS},
{"PurpleBuddyIconGetChecksum", "in\0i\0icon\0out\0s\0RESULT\0", purple_buddy_icon_get_checksum_DBUS},
{"PurpleBuddyIconGetData", "in\0i\0icon\0out\0ay\0RESULT\0", purple_buddy_icon_get_data_DBUS},
{"PurpleBuddyIconGetExtension", "in\0i\0icon\0out\0s\0RESULT\0", purple_buddy_icon_get_extension_DBUS},
{"PurpleBuddyIconGetFullPath", "in\0i\0icon\0out\0s\0RESULT\0", purple_buddy_icon_get_full_path_DBUS},
{"PurpleBuddyIconsSetForUser", "in\0i\0account\0in\0s\0username\0in\0i\0icon_data\0in\0i\0icon_len\0in\0s\0checksum\0", purple_buddy_icons_set_for_user_DBUS},
{"PurpleBuddyIconsFind", "in\0i\0account\0in\0s\0username\0out\0i\0RESULT\0", purple_buddy_icons_find_DBUS},
{"PurpleBuddyIconsFindAccountIcon", "in\0i\0account\0out\0i\0RESULT\0", purple_buddy_icons_find_account_icon_DBUS},
{"PurpleBuddyIconsSetAccountIcon", "in\0i\0account\0in\0i\0icon_data\0in\0i\0icon_len\0out\0i\0RESULT\0", purple_buddy_icons_set_account_icon_DBUS},
{"PurpleBuddyIconsGetAccountIconTimestamp", "in\0i\0account\0out\0i\0RESULT\0", purple_buddy_icons_get_account_icon_timestamp_DBUS},
{"PurpleBuddyIconsNodeHasCustomIcon", "in\0i\0node\0out\0i\0RESULT\0", purple_buddy_icons_node_has_custom_icon_DBUS},
{"PurpleBuddyIconsNodeFindCustomIcon", "in\0i\0node\0out\0i\0RESULT\0", purple_buddy_icons_node_find_custom_icon_DBUS},
{"PurpleBuddyIconsNodeSetCustomIcon", "in\0i\0node\0in\0i\0icon_data\0in\0i\0icon_len\0out\0i\0RESULT\0", purple_buddy_icons_node_set_custom_icon_DBUS},
{"PurpleBuddyIconsNodeSetCustomIconFromFile", "in\0i\0node\0in\0s\0filename\0out\0i\0RESULT\0", purple_buddy_icons_node_set_custom_icon_from_file_DBUS},
{"PurpleBuddyIconsHasCustomIcon", "in\0i\0contact\0out\0i\0RESULT\0", purple_buddy_icons_has_custom_icon_DBUS},
{"PurpleBuddyIconsFindCustomIcon", "in\0i\0contact\0out\0i\0RESULT\0", purple_buddy_icons_find_custom_icon_DBUS},
{"PurpleBuddyIconsSetCustomIcon", "in\0i\0contact\0in\0i\0icon_data\0in\0i\0icon_len\0out\0i\0RESULT\0", purple_buddy_icons_set_custom_icon_DBUS},
{"PurpleBuddyIconsSetCaching", "in\0i\0caching\0", purple_buddy_icons_set_caching_DBUS},
{"PurpleBuddyIconsIsCaching", "out\0i\0RESULT\0", purple_buddy_icons_is_caching_DBUS},
{"PurpleBuddyIconsSetCacheDir", "in\0s\0cache_dir\0", purple_buddy_icons_set_cache_dir_DBUS},
{"PurpleBuddyIconsGetCacheDir", "out\0s\0RESULT\0", purple_buddy_icons_get_cache_dir_DBUS},
{"PurpleBuddyIconsInit", "", purple_buddy_icons_init_DBUS},
{"PurpleBuddyIconsUninit", "", purple_buddy_icons_uninit_DBUS},
{"PurpleBuddyIconGetScaleSize", "in\0i\0spec\0in\0i\0width\0in\0i\0height\0", purple_buddy_icon_get_scale_size_DBUS},
{"PurpleConnectionNew", "in\0i\0account\0in\0i\0regist\0in\0s\0password\0", purple_connection_new_DBUS},
{"PurpleConnectionDestroy", "in\0i\0gc\0", purple_connection_destroy_DBUS},
{"PurpleConnectionSetState", "in\0i\0gc\0in\0i\0state\0", purple_connection_set_state_DBUS},
{"PurpleConnectionSetAccount", "in\0i\0gc\0in\0i\0account\0", purple_connection_set_account_DBUS},
{"PurpleConnectionSetDisplayName", "in\0i\0gc\0in\0s\0name\0", purple_connection_set_display_name_DBUS},
{"PurpleConnectionSetProtocolData", "in\0i\0connection\0in\0i\0proto_data\0", purple_connection_set_protocol_data_DBUS},
{"PurpleConnectionGetState", "in\0i\0gc\0out\0i\0RESULT\0", purple_connection_get_state_DBUS},
{"PurpleConnectionGetAccount", "in\0i\0gc\0out\0i\0RESULT\0", purple_connection_get_account_DBUS},
{"PurpleConnectionGetPrpl", "in\0i\0gc\0out\0i\0RESULT\0", purple_connection_get_prpl_DBUS},
{"PurpleConnectionGetPassword", "in\0i\0gc\0out\0s\0RESULT\0", purple_connection_get_password_DBUS},
{"PurpleConnectionGetDisplayName", "in\0i\0gc\0out\0s\0RESULT\0", purple_connection_get_display_name_DBUS},
{"PurpleConnectionUpdateProgress", "in\0i\0gc\0in\0s\0text\0in\0i\0step\0in\0i\0count\0", purple_connection_update_progress_DBUS},
{"PurpleConnectionNotice", "in\0i\0gc\0in\0s\0text\0", purple_connection_notice_DBUS},
{"PurpleConnectionError", "in\0i\0gc\0in\0s\0reason\0", purple_connection_error_DBUS},
{"PurpleConnectionErrorReason", "in\0i\0gc\0in\0i\0reason\0in\0s\0description\0", purple_connection_error_reason_DBUS},
{"PurpleConnectionSslError", "in\0i\0gc\0in\0i\0ssl_error\0", purple_connection_ssl_error_DBUS},
{"PurpleConnectionErrorIsFatal", "in\0i\0reason\0out\0i\0RESULT\0", purple_connection_error_is_fatal_DBUS},
{"PurpleConnectionsDisconnectAll", "", purple_connections_disconnect_all_DBUS},
{"PurpleConnectionsGetAll", "out\0ai\0RESULT\0", purple_connections_get_all_DBUS},
{"PurpleConnectionsGetConnecting", "out\0ai\0RESULT\0", purple_connections_get_connecting_DBUS},
{"PurpleConnectionsSetUiOps", "in\0i\0ops\0", purple_connections_set_ui_ops_DBUS},
{"PurpleConnectionsGetUiOps", "out\0i\0RESULT\0", purple_connections_get_ui_ops_DBUS},
{"PurpleConnectionsInit", "", purple_connections_init_DBUS},
{"PurpleConnectionsUninit", "", purple_connections_uninit_DBUS},
{"PurpleConversationNew", "in\0i\0type\0in\0i\0account\0in\0s\0name\0out\0i\0RESULT\0", purple_conversation_new_DBUS},
{"PurpleConversationDestroy", "in\0i\0conv\0", purple_conversation_destroy_DBUS},
{"PurpleConversationPresent", "in\0i\0conv\0", purple_conversation_present_DBUS},
{"PurpleConversationGetType", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_type_DBUS},
{"PurpleConversationSetUiOps", "in\0i\0conv\0in\0i\0ops\0", purple_conversation_set_ui_ops_DBUS},
{"PurpleConversationsSetUiOps", "in\0i\0ops\0", purple_conversations_set_ui_ops_DBUS},
{"PurpleConversationGetUiOps", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_ui_ops_DBUS},
{"PurpleConversationSetAccount", "in\0i\0conv\0in\0i\0account\0", purple_conversation_set_account_DBUS},
{"PurpleConversationGetAccount", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_account_DBUS},
{"PurpleConversationGetGc", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_gc_DBUS},
{"PurpleConversationSetTitle", "in\0i\0conv\0in\0s\0title\0", purple_conversation_set_title_DBUS},
{"PurpleConversationGetTitle", "in\0i\0conv\0out\0s\0RESULT\0", purple_conversation_get_title_DBUS},
{"PurpleConversationAutosetTitle", "in\0i\0conv\0", purple_conversation_autoset_title_DBUS},
{"PurpleConversationSetName", "in\0i\0conv\0in\0s\0name\0", purple_conversation_set_name_DBUS},
{"PurpleConversationGetName", "in\0i\0conv\0out\0s\0RESULT\0", purple_conversation_get_name_DBUS},
{"PurpleConvChatCbGetAttribute", "in\0i\0cb\0in\0s\0key\0out\0s\0RESULT\0", purple_conv_chat_cb_get_attribute_DBUS},
{"PurpleConvChatCbGetAttributeKeys", "in\0i\0cb\0out\0ai\0RESULT\0", purple_conv_chat_cb_get_attribute_keys_DBUS},
{"PurpleConvChatCbSetAttribute", "in\0i\0chat\0in\0i\0cb\0in\0s\0key\0in\0s\0value\0", purple_conv_chat_cb_set_attribute_DBUS},
{"PurpleConvChatCbSetAttributes", "in\0i\0chat\0in\0i\0cb\0in\0i\0keys\0in\0i\0values\0", purple_conv_chat_cb_set_attributes_DBUS},
{"PurpleConversationSetLogging", "in\0i\0conv\0in\0i\0log\0", purple_conversation_set_logging_DBUS},
{"PurpleConversationIsLogging", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_is_logging_DBUS},
{"PurpleConversationGetImData", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_im_data_DBUS},
{"PurpleConversationGetChatData", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_chat_data_DBUS},
{"PurpleGetConversations", "out\0ai\0RESULT\0", purple_get_conversations_DBUS},
{"PurpleGetIms", "out\0ai\0RESULT\0", purple_get_ims_DBUS},
{"PurpleGetChats", "out\0ai\0RESULT\0", purple_get_chats_DBUS},
{"PurpleFindConversationWithAccount", "in\0i\0type\0in\0s\0name\0in\0i\0account\0out\0i\0RESULT\0", purple_find_conversation_with_account_DBUS},
{"PurpleConversationWrite", "in\0i\0conv\0in\0s\0who\0in\0s\0message\0in\0i\0flags\0in\0i\0mtime\0", purple_conversation_write_DBUS},
{"PurpleConversationSetFeatures", "in\0i\0conv\0in\0i\0features\0", purple_conversation_set_features_DBUS},
{"PurpleConversationGetFeatures", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_get_features_DBUS},
{"PurpleConversationHasFocus", "in\0i\0conv\0out\0i\0RESULT\0", purple_conversation_has_focus_DBUS},
{"PurpleConversationUpdate", "in\0i\0conv\0in\0i\0type\0", purple_conversation_update_DBUS},
{"PurpleConversationGetMessageHistory", "in\0i\0conv\0out\0ai\0RESULT\0", purple_conversation_get_message_history_DBUS},
{"PurpleConversationClearMessageHistory", "in\0i\0conv\0", purple_conversation_clear_message_history_DBUS},
{"PurpleConversationMessageGetSender", "in\0i\0msg\0out\0s\0RESULT\0", purple_conversation_message_get_sender_DBUS},
{"PurpleConversationMessageGetMessage", "in\0i\0msg\0out\0s\0RESULT\0", purple_conversation_message_get_message_DBUS},
{"PurpleConversationMessageGetFlags", "in\0i\0msg\0out\0i\0RESULT\0", purple_conversation_message_get_flags_DBUS},
{"PurpleConversationMessageGetTimestamp", "in\0i\0msg\0out\0i\0RESULT\0", purple_conversation_message_get_timestamp_DBUS},
{"PurpleConvImGetConversation", "in\0i\0im\0out\0i\0RESULT\0", purple_conv_im_get_conversation_DBUS},
{"PurpleConvImSetIcon", "in\0i\0im\0in\0i\0icon\0", purple_conv_im_set_icon_DBUS},
{"PurpleConvImGetIcon", "in\0i\0im\0out\0i\0RESULT\0", purple_conv_im_get_icon_DBUS},
{"PurpleConvImSetTypingState", "in\0i\0im\0in\0i\0state\0", purple_conv_im_set_typing_state_DBUS},
{"PurpleConvImGetTypingState", "in\0i\0im\0out\0i\0RESULT\0", purple_conv_im_get_typing_state_DBUS},
{"PurpleConvImStartTypingTimeout", "in\0i\0im\0in\0i\0timeout\0", purple_conv_im_start_typing_timeout_DBUS},
{"PurpleConvImStopTypingTimeout", "in\0i\0im\0", purple_conv_im_stop_typing_timeout_DBUS},
{"PurpleConvImGetTypingTimeout", "in\0i\0im\0out\0i\0RESULT\0", purple_conv_im_get_typing_timeout_DBUS},
{"PurpleConvImSetTypeAgain", "in\0i\0im\0in\0u\0val\0", purple_conv_im_set_type_again_DBUS},
{"PurpleConvImGetTypeAgain", "in\0i\0im\0out\0i\0RESULT\0", purple_conv_im_get_type_again_DBUS},
{"PurpleConvImStartSendTypedTimeout", "in\0i\0im\0", purple_conv_im_start_send_typed_timeout_DBUS},
{"PurpleConvImStopSendTypedTimeout", "in\0i\0im\0", purple_conv_im_stop_send_typed_timeout_DBUS},
{"PurpleConvImGetSendTypedTimeout", "in\0i\0im\0out\0i\0RESULT\0", purple_conv_im_get_send_typed_timeout_DBUS},
{"PurpleConvImUpdateTyping", "in\0i\0im\0", purple_conv_im_update_typing_DBUS},
{"PurpleConvImWrite", "in\0i\0im\0in\0s\0who\0in\0s\0message\0in\0i\0flags\0in\0i\0mtime\0", purple_conv_im_write_DBUS},
{"PurpleConvPresentError", "in\0s\0who\0in\0i\0account\0in\0s\0what\0out\0i\0RESULT\0", purple_conv_present_error_DBUS},
{"PurpleConvImSend", "in\0i\0im\0in\0s\0message\0", purple_conv_im_send_DBUS},
{"PurpleConvSendConfirm", "in\0i\0conv\0in\0s\0message\0", purple_conv_send_confirm_DBUS},
{"PurpleConvImSendWithFlags", "in\0i\0im\0in\0s\0message\0in\0i\0flags\0", purple_conv_im_send_with_flags_DBUS},
{"PurpleConvCustomSmileyAdd", "in\0i\0conv\0in\0s\0smile\0in\0s\0cksum_type\0in\0s\0chksum\0in\0i\0remote\0out\0i\0RESULT\0", purple_conv_custom_smiley_add_DBUS},
{"PurpleConvCustomSmileyClose", "in\0i\0conv\0in\0s\0smile\0", purple_conv_custom_smiley_close_DBUS},
{"PurpleConvChatGetConversation", "in\0i\0chat\0out\0i\0RESULT\0", purple_conv_chat_get_conversation_DBUS},
{"PurpleConvChatSetUsers", "in\0i\0chat\0in\0i\0users\0out\0ai\0RESULT\0", purple_conv_chat_set_users_DBUS},
{"PurpleConvChatGetUsers", "in\0i\0chat\0out\0ai\0RESULT\0", purple_conv_chat_get_users_DBUS},
{"PurpleConvChatIgnore", "in\0i\0chat\0in\0s\0name\0", purple_conv_chat_ignore_DBUS},
{"PurpleConvChatUnignore", "in\0i\0chat\0in\0s\0name\0", purple_conv_chat_unignore_DBUS},
{"PurpleConvChatSetIgnored", "in\0i\0chat\0in\0i\0ignored\0out\0ai\0RESULT\0", purple_conv_chat_set_ignored_DBUS},
{"PurpleConvChatGetIgnored", "in\0i\0chat\0out\0ai\0RESULT\0", purple_conv_chat_get_ignored_DBUS},
{"PurpleConvChatGetIgnoredUser", "in\0i\0chat\0in\0s\0user\0out\0s\0RESULT\0", purple_conv_chat_get_ignored_user_DBUS},
{"PurpleConvChatIsUserIgnored", "in\0i\0chat\0in\0s\0user\0out\0i\0RESULT\0", purple_conv_chat_is_user_ignored_DBUS},
{"PurpleConvChatSetTopic", "in\0i\0chat\0in\0s\0who\0in\0s\0topic\0", purple_conv_chat_set_topic_DBUS},
{"PurpleConvChatGetTopic", "in\0i\0chat\0out\0s\0RESULT\0", purple_conv_chat_get_topic_DBUS},
{"PurpleConvChatSetId", "in\0i\0chat\0in\0i\0id\0", purple_conv_chat_set_id_DBUS},
{"PurpleConvChatGetId", "in\0i\0chat\0out\0i\0RESULT\0", purple_conv_chat_get_id_DBUS},
{"PurpleConvChatWrite", "in\0i\0chat\0in\0s\0who\0in\0s\0message\0in\0i\0flags\0in\0i\0mtime\0", purple_conv_chat_write_DBUS},
{"PurpleConvChatSend", "in\0i\0chat\0in\0s\0message\0", purple_conv_chat_send_DBUS},
{"PurpleConvChatSendWithFlags", "in\0i\0chat\0in\0s\0message\0in\0i\0flags\0", purple_conv_chat_send_with_flags_DBUS},
{"PurpleConvChatAddUser", "in\0i\0chat\0in\0s\0user\0in\0s\0extra_msg\0in\0i\0flags\0in\0i\0new_arrival\0", purple_conv_chat_add_user_DBUS},
{"PurpleConvChatAddUsers", "in\0i\0chat\0in\0i\0users\0in\0i\0extra_msgs\0in\0i\0flags\0in\0i\0new_arrivals\0", purple_conv_chat_add_users_DBUS},
{"PurpleConvChatRenameUser", "in\0i\0chat\0in\0s\0old_user\0in\0s\0new_user\0", purple_conv_chat_rename_user_DBUS},
{"PurpleConvChatRemoveUser", "in\0i\0chat\0in\0s\0user\0in\0s\0reason\0", purple_conv_chat_remove_user_DBUS},
{"PurpleConvChatRemoveUsers", "in\0i\0chat\0in\0i\0users\0in\0s\0reason\0", purple_conv_chat_remove_users_DBUS},
{"PurpleConvChatFindUser", "in\0i\0chat\0in\0s\0user\0out\0i\0RESULT\0", purple_conv_chat_find_user_DBUS},
{"PurpleConvChatUserSetFlags", "in\0i\0chat\0in\0s\0user\0in\0i\0flags\0", purple_conv_chat_user_set_flags_DBUS},
{"PurpleConvChatUserGetFlags", "in\0i\0chat\0in\0s\0user\0out\0i\0RESULT\0", purple_conv_chat_user_get_flags_DBUS},
{"PurpleConvChatClearUsers", "in\0i\0chat\0", purple_conv_chat_clear_users_DBUS},
{"PurpleConvChatSetNick", "in\0i\0chat\0in\0s\0nick\0", purple_conv_chat_set_nick_DBUS},
{"PurpleConvChatGetNick", "in\0i\0chat\0out\0s\0RESULT\0", purple_conv_chat_get_nick_DBUS},
{"PurpleFindChat", "in\0i\0gc\0in\0i\0id\0out\0i\0RESULT\0", purple_find_chat_DBUS},
{"PurpleConvChatLeft", "in\0i\0chat\0", purple_conv_chat_left_DBUS},
{"PurpleConvChatInviteUser", "in\0i\0chat\0in\0s\0user\0in\0s\0message\0in\0i\0confirm\0", purple_conv_chat_invite_user_DBUS},
{"PurpleConvChatHasLeft", "in\0i\0chat\0out\0i\0RESULT\0", purple_conv_chat_has_left_DBUS},
{"PurpleConvChatCbNew", "in\0s\0name\0in\0s\0alias\0in\0i\0flags\0out\0i\0RESULT\0", purple_conv_chat_cb_new_DBUS},
{"PurpleConvChatCbFind", "in\0i\0chat\0in\0s\0name\0out\0i\0RESULT\0", purple_conv_chat_cb_find_DBUS},
{"PurpleConvChatCbGetName", "in\0i\0cb\0out\0s\0RESULT\0", purple_conv_chat_cb_get_name_DBUS},
{"PurpleConvChatCbDestroy", "in\0i\0cb\0", purple_conv_chat_cb_destroy_DBUS},
{"PurpleConversationGetExtendedMenu", "in\0i\0conv\0out\0ai\0RESULT\0", purple_conversation_get_extended_menu_DBUS},
{"PurpleConversationsInit", "", purple_conversations_init_DBUS},
{"PurpleConversationsUninit", "", purple_conversations_uninit_DBUS},
{"PurpleCoreInit", "in\0s\0ui\0out\0i\0RESULT\0", purple_core_init_DBUS},
{"PurpleCoreQuit", "", purple_core_quit_DBUS},
{"PurpleCoreGetVersion", "out\0s\0RESULT\0", purple_core_get_version_DBUS},
{"PurpleCoreGetUi", "out\0s\0RESULT\0", purple_core_get_ui_DBUS},
{"PurpleGetCore", "out\0i\0RESULT\0", purple_get_core_DBUS},
{"PurpleCoreSetUiOps", "in\0i\0ops\0", purple_core_set_ui_ops_DBUS},
{"PurpleCoreGetUiOps", "out\0i\0RESULT\0", purple_core_get_ui_ops_DBUS},
{"PurpleCoreMigrate", "out\0i\0RESULT\0", purple_core_migrate_DBUS},
{"PurpleCoreEnsureSingleInstance", "out\0i\0RESULT\0", purple_core_ensure_single_instance_DBUS},
{"PurpleXferNew", "in\0i\0account\0in\0i\0type\0in\0s\0who\0out\0i\0RESULT\0", purple_xfer_new_DBUS},
{"PurpleXfersGetAll", "out\0ai\0RESULT\0", purple_xfers_get_all_DBUS},
{"PurpleXferRef", "in\0i\0xfer\0", purple_xfer_ref_DBUS},
{"PurpleXferUnref", "in\0i\0xfer\0", purple_xfer_unref_DBUS},
{"PurpleXferRequest", "in\0i\0xfer\0", purple_xfer_request_DBUS},
{"PurpleXferRequestAccepted", "in\0i\0xfer\0in\0s\0filename\0", purple_xfer_request_accepted_DBUS},
{"PurpleXferRequestDenied", "in\0i\0xfer\0", purple_xfer_request_denied_DBUS},
{"PurpleXferGetType", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_type_DBUS},
{"PurpleXferGetAccount", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_account_DBUS},
{"PurpleXferGetRemoteUser", "in\0i\0xfer\0out\0s\0RESULT\0", purple_xfer_get_remote_user_DBUS},
{"PurpleXferGetStatus", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_status_DBUS},
{"PurpleXferIsCanceled", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_is_canceled_DBUS},
{"PurpleXferIsCompleted", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_is_completed_DBUS},
{"PurpleXferGetFilename", "in\0i\0xfer\0out\0s\0RESULT\0", purple_xfer_get_filename_DBUS},
{"PurpleXferGetLocalFilename", "in\0i\0xfer\0out\0s\0RESULT\0", purple_xfer_get_local_filename_DBUS},
{"PurpleXferGetBytesSent", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_bytes_sent_DBUS},
{"PurpleXferGetBytesRemaining", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_bytes_remaining_DBUS},
{"PurpleXferGetSize", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_size_DBUS},
{"PurpleXferGetLocalPort", "in\0i\0xfer\0out\0u\0RESULT\0", purple_xfer_get_local_port_DBUS},
{"PurpleXferGetRemoteIp", "in\0i\0xfer\0out\0s\0RESULT\0", purple_xfer_get_remote_ip_DBUS},
{"PurpleXferGetRemotePort", "in\0i\0xfer\0out\0u\0RESULT\0", purple_xfer_get_remote_port_DBUS},
{"PurpleXferGetStartTime", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_start_time_DBUS},
{"PurpleXferGetEndTime", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_end_time_DBUS},
{"PurpleXferSetCompleted", "in\0i\0xfer\0in\0i\0completed\0", purple_xfer_set_completed_DBUS},
{"PurpleXferSetMessage", "in\0i\0xfer\0in\0s\0message\0", purple_xfer_set_message_DBUS},
{"PurpleXferSetFilename", "in\0i\0xfer\0in\0s\0filename\0", purple_xfer_set_filename_DBUS},
{"PurpleXferSetLocalFilename", "in\0i\0xfer\0in\0s\0filename\0", purple_xfer_set_local_filename_DBUS},
{"PurpleXferSetSize", "in\0i\0xfer\0in\0i\0size\0", purple_xfer_set_size_DBUS},
{"PurpleXferSetBytesSent", "in\0i\0xfer\0in\0i\0bytes_sent\0", purple_xfer_set_bytes_sent_DBUS},
{"PurpleXferGetUiOps", "in\0i\0xfer\0out\0i\0RESULT\0", purple_xfer_get_ui_ops_DBUS},
{"PurpleXferStart", "in\0i\0xfer\0in\0i\0fd\0in\0s\0ip\0in\0u\0port\0", purple_xfer_start_DBUS},
{"PurpleXferEnd", "in\0i\0xfer\0", purple_xfer_end_DBUS},
{"PurpleXferAdd", "in\0i\0xfer\0", purple_xfer_add_DBUS},
{"PurpleXferCancelLocal", "in\0i\0xfer\0", purple_xfer_cancel_local_DBUS},
{"PurpleXferCancelRemote", "in\0i\0xfer\0", purple_xfer_cancel_remote_DBUS},
{"PurpleXferError", "in\0i\0type\0in\0i\0account\0in\0s\0who\0in\0s\0msg\0", purple_xfer_error_DBUS},
{"PurpleXferUpdateProgress", "in\0i\0xfer\0", purple_xfer_update_progress_DBUS},
{"PurpleXferGetThumbnail", "in\0i\0xfer\0out\0ay\0RESULT\0", purple_xfer_get_thumbnail_DBUS},
{"PurpleXferGetThumbnailMimetype", "in\0i\0xfer\0out\0s\0RESULT\0", purple_xfer_get_thumbnail_mimetype_DBUS},
{"PurpleXferPrepareThumbnail", "in\0i\0xfer\0in\0s\0formats\0", purple_xfer_prepare_thumbnail_DBUS},
{"PurpleXfersInit", "", purple_xfers_init_DBUS},
{"PurpleXfersUninit", "", purple_xfers_uninit_DBUS},
{"PurpleXfersSetUiOps", "in\0i\0ops\0", purple_xfers_set_ui_ops_DBUS},
{"PurpleXfersGetUiOps", "out\0i\0RESULT\0", purple_xfers_get_ui_ops_DBUS},
{"PurpleLogFree", "in\0i\0log\0", purple_log_free_DBUS},
{"PurpleLogWrite", "in\0i\0log\0in\0i\0type\0in\0s\0from\0in\0i\0time\0in\0s\0message\0", purple_log_write_DBUS},
{"PurpleLogGetLogs", "in\0i\0type\0in\0s\0name\0in\0i\0account\0out\0ai\0RESULT\0", purple_log_get_logs_DBUS},
{"PurpleLogGetSystemLogs", "in\0i\0account\0out\0ai\0RESULT\0", purple_log_get_system_logs_DBUS},
{"PurpleLogGetSize", "in\0i\0log\0out\0i\0RESULT\0", purple_log_get_size_DBUS},
{"PurpleLogGetTotalSize", "in\0i\0type\0in\0s\0name\0in\0i\0account\0out\0i\0RESULT\0", purple_log_get_total_size_DBUS},
{"PurpleLogGetActivityScore", "in\0i\0type\0in\0s\0name\0in\0i\0account\0out\0i\0RESULT\0", purple_log_get_activity_score_DBUS},
{"PurpleLogIsDeletable", "in\0i\0log\0out\0i\0RESULT\0", purple_log_is_deletable_DBUS},
{"PurpleLogDelete", "in\0i\0log\0out\0i\0RESULT\0", purple_log_delete_DBUS},
{"PurpleLogGetLogDir", "in\0i\0type\0in\0s\0name\0in\0i\0account\0out\0s\0RESULT\0", purple_log_get_log_dir_DBUS},
{"PurpleLogSetFree", "in\0i\0set\0", purple_log_set_free_DBUS},
{"PurpleLogCommonWriter", "in\0i\0log\0in\0s\0ext\0", purple_log_common_writer_DBUS},
{"PurpleLogCommonLister", "in\0i\0type\0in\0s\0name\0in\0i\0account\0in\0s\0ext\0in\0i\0logger\0out\0ai\0RESULT\0", purple_log_common_lister_DBUS},
{"PurpleLogCommonTotalSizer", "in\0i\0type\0in\0s\0name\0in\0i\0account\0in\0s\0ext\0out\0i\0RESULT\0", purple_log_common_total_sizer_DBUS},
{"PurpleLogCommonSizer", "in\0i\0log\0out\0i\0RESULT\0", purple_log_common_sizer_DBUS},
{"PurpleLogCommonDeleter", "in\0i\0log\0out\0i\0RESULT\0", purple_log_common_deleter_DBUS},
{"PurpleLogCommonIsDeletable", "in\0i\0log\0out\0i\0RESULT\0", purple_log_common_is_deletable_DBUS},
{"PurpleLogLoggerFree", "in\0i\0logger\0", purple_log_logger_free_DBUS},
{"PurpleLogLoggerAdd", "in\0i\0logger\0", purple_log_logger_add_DBUS},
{"PurpleLogLoggerRemove", "in\0i\0logger\0", purple_log_logger_remove_DBUS},
{"PurpleLogLoggerSet", "in\0i\0logger\0", purple_log_logger_set_DBUS},
{"PurpleLogLoggerGet", "out\0i\0RESULT\0", purple_log_logger_get_DBUS},
{"PurpleLogLoggerGetOptions", "out\0ai\0RESULT\0", purple_log_logger_get_options_DBUS},
{"PurpleLogInit", "", purple_log_init_DBUS},
{"PurpleLogUninit", "", purple_log_uninit_DBUS},
{"PurpleNotifySearchresultsFree", "in\0i\0results\0", purple_notify_searchresults_free_DBUS},
{"PurpleNotifySearchresultsNewRows", "in\0i\0gc\0in\0i\0results\0in\0i\0data\0", purple_notify_searchresults_new_rows_DBUS},
{"PurpleNotifySearchresultsNew", "out\0i\0RESULT\0", purple_notify_searchresults_new_DBUS},
{"PurpleNotifySearchresultsColumnNew", "in\0s\0title\0out\0i\0RESULT\0", purple_notify_searchresults_column_new_DBUS},
{"PurpleNotifySearchresultsColumnAdd", "in\0i\0results\0in\0i\0column\0", purple_notify_searchresults_column_add_DBUS},
{"PurpleNotifySearchresultsRowAdd", "in\0i\0results\0in\0i\0row\0", purple_notify_searchresults_row_add_DBUS},
{"PurpleNotifySearchresultsGetRowsCount", "in\0i\0results\0out\0i\0RESULT\0", purple_notify_searchresults_get_rows_count_DBUS},
{"PurpleNotifySearchresultsGetColumnsCount", "in\0i\0results\0out\0i\0RESULT\0", purple_notify_searchresults_get_columns_count_DBUS},
{"PurpleNotifySearchresultsRowGet", "in\0i\0results\0in\0u\0row_id\0out\0ai\0RESULT\0", purple_notify_searchresults_row_get_DBUS},
{"PurpleNotifySearchresultsColumnGetTitle", "in\0i\0results\0in\0u\0column_id\0out\0s\0RESULT\0", purple_notify_searchresults_column_get_title_DBUS},
{"PurpleNotifyUserInfoNew", "out\0i\0RESULT\0", purple_notify_user_info_new_DBUS},
{"PurpleNotifyUserInfoDestroy", "in\0i\0user_info\0", purple_notify_user_info_destroy_DBUS},
{"PurpleNotifyUserInfoGetEntries", "in\0i\0user_info\0out\0ai\0RESULT\0", purple_notify_user_info_get_entries_DBUS},
{"PurpleNotifyUserInfoGetTextWithNewline", "in\0i\0user_info\0in\0s\0newline\0out\0s\0RESULT\0", purple_notify_user_info_get_text_with_newline_DBUS},
{"PurpleNotifyUserInfoAddPair", "in\0i\0user_info\0in\0s\0label\0in\0s\0value\0", purple_notify_user_info_add_pair_DBUS},
{"PurpleNotifyUserInfoAddPairPlaintext", "in\0i\0user_info\0in\0s\0label\0in\0s\0value\0", purple_notify_user_info_add_pair_plaintext_DBUS},
{"PurpleNotifyUserInfoPrependPair", "in\0i\0user_info\0in\0s\0label\0in\0s\0value\0", purple_notify_user_info_prepend_pair_DBUS},
{"PurpleNotifyUserInfoRemoveEntry", "in\0i\0user_info\0in\0i\0user_info_entry\0", purple_notify_user_info_remove_entry_DBUS},
{"PurpleNotifyUserInfoEntryNew", "in\0s\0label\0in\0s\0value\0out\0i\0RESULT\0", purple_notify_user_info_entry_new_DBUS},
{"PurpleNotifyUserInfoAddSectionBreak", "in\0i\0user_info\0", purple_notify_user_info_add_section_break_DBUS},
{"PurpleNotifyUserInfoPrependSectionBreak", "in\0i\0user_info\0", purple_notify_user_info_prepend_section_break_DBUS},
{"PurpleNotifyUserInfoAddSectionHeader", "in\0i\0user_info\0in\0s\0label\0", purple_notify_user_info_add_section_header_DBUS},
{"PurpleNotifyUserInfoPrependSectionHeader", "in\0i\0user_info\0in\0s\0label\0", purple_notify_user_info_prepend_section_header_DBUS},
{"PurpleNotifyUserInfoRemoveLastItem", "in\0i\0user_info\0", purple_notify_user_info_remove_last_item_DBUS},
{"PurpleNotifyUserInfoEntryGetLabel", "in\0i\0user_info_entry\0out\0s\0RESULT\0", purple_notify_user_info_entry_get_label_DBUS},
{"PurpleNotifyUserInfoEntrySetLabel", "in\0i\0user_info_entry\0in\0s\0label\0", purple_notify_user_info_entry_set_label_DBUS},
{"PurpleNotifyUserInfoEntryGetValue", "in\0i\0user_info_entry\0out\0s\0RESULT\0", purple_notify_user_info_entry_get_value_DBUS},
{"PurpleNotifyUserInfoEntrySetValue", "in\0i\0user_info_entry\0in\0s\0value\0", purple_notify_user_info_entry_set_value_DBUS},
{"PurpleNotifyUserInfoEntryGetType", "in\0i\0user_info_entry\0out\0i\0RESULT\0", purple_notify_user_info_entry_get_type_DBUS},
{"PurpleNotifyUserInfoEntrySetType", "in\0i\0user_info_entry\0in\0i\0type\0", purple_notify_user_info_entry_set_type_DBUS},
{"PurpleNotifyClose", "in\0i\0type\0in\0i\0ui_handle\0", purple_notify_close_DBUS},
{"PurpleNotifyCloseWithHandle", "in\0i\0handle\0", purple_notify_close_with_handle_DBUS},
{"PurpleNotifySetUiOps", "in\0i\0ops\0", purple_notify_set_ui_ops_DBUS},
{"PurpleNotifyGetUiOps", "out\0i\0RESULT\0", purple_notify_get_ui_ops_DBUS},
{"PurpleNotifyInit", "", purple_notify_init_DBUS},
{"PurpleNotifyUninit", "", purple_notify_uninit_DBUS},
{"PurplePrefsSetUiOps", "in\0i\0ops\0", purple_prefs_set_ui_ops_DBUS},
{"PurplePrefsGetUiOps", "out\0i\0RESULT\0", purple_prefs_get_ui_ops_DBUS},
{"PurplePrefsInit", "", purple_prefs_init_DBUS},
{"PurplePrefsUninit", "", purple_prefs_uninit_DBUS},
{"PurplePrefsAddNone", "in\0s\0name\0", purple_prefs_add_none_DBUS},
{"PurplePrefsAddBool", "in\0s\0name\0in\0i\0value\0", purple_prefs_add_bool_DBUS},
{"PurplePrefsAddInt", "in\0s\0name\0in\0i\0value\0", purple_prefs_add_int_DBUS},
{"PurplePrefsAddString", "in\0s\0name\0in\0s\0value\0", purple_prefs_add_string_DBUS},
{"PurplePrefsAddStringList", "in\0s\0name\0in\0i\0value\0", purple_prefs_add_string_list_DBUS},
{"PurplePrefsAddPath", "in\0s\0name\0in\0s\0value\0", purple_prefs_add_path_DBUS},
{"PurplePrefsAddPathList", "in\0s\0name\0in\0i\0value\0", purple_prefs_add_path_list_DBUS},
{"PurplePrefsRemove", "in\0s\0name\0", purple_prefs_remove_DBUS},
{"PurplePrefsRename", "in\0s\0oldname\0in\0s\0newname\0", purple_prefs_rename_DBUS},
{"PurplePrefsRenameBooleanToggle", "in\0s\0oldname\0in\0s\0newname\0", purple_prefs_rename_boolean_toggle_DBUS},
{"PurplePrefsDestroy", "", purple_prefs_destroy_DBUS},
{"PurplePrefsSetBool", "in\0s\0name\0in\0i\0value\0", purple_prefs_set_bool_DBUS},
{"PurplePrefsSetInt", "in\0s\0name\0in\0i\0value\0", purple_prefs_set_int_DBUS},
{"PurplePrefsSetString", "in\0s\0name\0in\0s\0value\0", purple_prefs_set_string_DBUS},
{"PurplePrefsSetStringList", "in\0s\0name\0in\0i\0value\0", purple_prefs_set_string_list_DBUS},
{"PurplePrefsSetPath", "in\0s\0name\0in\0s\0value\0", purple_prefs_set_path_DBUS},
{"PurplePrefsSetPathList", "in\0s\0name\0in\0i\0value\0", purple_prefs_set_path_list_DBUS},
{"PurplePrefsExists", "in\0s\0name\0out\0i\0RESULT\0", purple_prefs_exists_DBUS},
{"PurplePrefsGetType", "in\0s\0name\0out\0i\0RESULT\0", purple_prefs_get_type_DBUS},
{"PurplePrefsGetBool", "in\0s\0name\0out\0i\0RESULT\0", purple_prefs_get_bool_DBUS},
{"PurplePrefsGetInt", "in\0s\0name\0out\0i\0RESULT\0", purple_prefs_get_int_DBUS},
{"PurplePrefsGetString", "in\0s\0name\0out\0s\0RESULT\0", purple_prefs_get_string_DBUS},
{"PurplePrefsGetStringList", "in\0s\0name\0out\0as\0RESULT\0", purple_prefs_get_string_list_DBUS},
{"PurplePrefsGetPath", "in\0s\0name\0out\0s\0RESULT\0", purple_prefs_get_path_DBUS},
{"PurplePrefsGetPathList", "in\0s\0name\0out\0as\0RESULT\0", purple_prefs_get_path_list_DBUS},
{"PurplePrefsGetChildrenNames", "in\0s\0name\0out\0as\0RESULT\0", purple_prefs_get_children_names_DBUS},
{"PurplePrefsDisconnectCallback", "in\0i\0callback_id\0", purple_prefs_disconnect_callback_DBUS},
{"PurplePrefsDisconnectByHandle", "in\0i\0handle\0", purple_prefs_disconnect_by_handle_DBUS},
{"PurplePrefsTriggerCallback", "in\0s\0name\0", purple_prefs_trigger_callback_DBUS},
{"PurplePrefsTriggerCallbackObject", "in\0i\0data\0", purple_prefs_trigger_callback_object_DBUS},
{"PurplePrefsLoad", "out\0i\0RESULT\0", purple_prefs_load_DBUS},
{"PurplePrefsUpdateOld", "", purple_prefs_update_old_DBUS},
{"PurpleRoomlistShowWithAccount", "in\0i\0account\0", purple_roomlist_show_with_account_DBUS},
{"PurpleRoomlistNew", "in\0i\0account\0out\0i\0RESULT\0", purple_roomlist_new_DBUS},
{"PurpleRoomlistRef", "in\0i\0list\0", purple_roomlist_ref_DBUS},
{"PurpleRoomlistUnref", "in\0i\0list\0", purple_roomlist_unref_DBUS},
{"PurpleRoomlistSetFields", "in\0i\0list\0in\0i\0fields\0", purple_roomlist_set_fields_DBUS},
{"PurpleRoomlistSetInProgress", "in\0i\0list\0in\0i\0in_progress\0", purple_roomlist_set_in_progress_DBUS},
{"PurpleRoomlistGetInProgress", "in\0i\0list\0out\0i\0RESULT\0", purple_roomlist_get_in_progress_DBUS},
{"PurpleRoomlistRoomAdd", "in\0i\0list\0in\0i\0room\0", purple_roomlist_room_add_DBUS},
{"PurpleRoomlistGetList", "in\0i\0gc\0out\0i\0RESULT\0", purple_roomlist_get_list_DBUS},
{"PurpleRoomlistCancelGetList", "in\0i\0list\0", purple_roomlist_cancel_get_list_DBUS},
{"PurpleRoomlistExpandCategory", "in\0i\0list\0in\0i\0category\0", purple_roomlist_expand_category_DBUS},
{"PurpleRoomlistGetFields", "in\0i\0roomlist\0out\0ai\0RESULT\0", purple_roomlist_get_fields_DBUS},
{"PurpleRoomlistRoomNew", "in\0i\0type\0in\0s\0name\0in\0i\0parent\0out\0i\0RESULT\0", purple_roomlist_room_new_DBUS},
{"PurpleRoomlistRoomJoin", "in\0i\0list\0in\0i\0room\0", purple_roomlist_room_join_DBUS},
{"PurpleRoomlistRoomGetType", "in\0i\0room\0out\0i\0RESULT\0", purple_roomlist_room_get_type_DBUS},
{"PurpleRoomlistRoomGetName", "in\0i\0room\0out\0s\0RESULT\0", purple_roomlist_room_get_name_DBUS},
{"PurpleRoomlistRoomGetParent", "in\0i\0room\0out\0i\0RESULT\0", purple_roomlist_room_get_parent_DBUS},
{"PurpleRoomlistRoomGetFields", "in\0i\0room\0out\0ai\0RESULT\0", purple_roomlist_room_get_fields_DBUS},
{"PurpleRoomlistFieldNew", "in\0i\0type\0in\0s\0label\0in\0s\0name\0in\0i\0hidden\0out\0i\0RESULT\0", purple_roomlist_field_new_DBUS},
{"PurpleRoomlistFieldGetType", "in\0i\0field\0out\0i\0RESULT\0", purple_roomlist_field_get_type_DBUS},
{"PurpleRoomlistFieldGetLabel", "in\0i\0field\0out\0s\0RESULT\0", purple_roomlist_field_get_label_DBUS},
{"PurpleRoomlistFieldGetHidden", "in\0i\0field\0out\0i\0RESULT\0", purple_roomlist_field_get_hidden_DBUS},
{"PurpleRoomlistSetUiOps", "in\0i\0ops\0", purple_roomlist_set_ui_ops_DBUS},
{"PurpleRoomlistGetUiOps", "out\0i\0RESULT\0", purple_roomlist_get_ui_ops_DBUS},
{"PurpleSavedstatusNew", "in\0s\0title\0in\0i\0type\0out\0i\0RESULT\0", purple_savedstatus_new_DBUS},
{"PurpleSavedstatusSetTitle", "in\0i\0status\0in\0s\0title\0", purple_savedstatus_set_title_DBUS},
{"PurpleSavedstatusSetType", "in\0i\0status\0in\0i\0type\0", purple_savedstatus_set_type_DBUS},
{"PurpleSavedstatusSetMessage", "in\0i\0status\0in\0s\0message\0", purple_savedstatus_set_message_DBUS},
{"PurpleSavedstatusSetSubstatus", "in\0i\0status\0in\0i\0account\0in\0i\0type\0in\0s\0message\0", purple_savedstatus_set_substatus_DBUS},
{"PurpleSavedstatusUnsetSubstatus", "in\0i\0saved_status\0in\0i\0account\0", purple_savedstatus_unset_substatus_DBUS},
{"PurpleSavedstatusDelete", "in\0s\0title\0out\0i\0RESULT\0", purple_savedstatus_delete_DBUS},
{"PurpleSavedstatusDeleteByStatus", "in\0i\0saved_status\0", purple_savedstatus_delete_by_status_DBUS},
{"PurpleSavedstatusesGetAll", "out\0ai\0RESULT\0", purple_savedstatuses_get_all_DBUS},
{"PurpleSavedstatusesGetPopular", "in\0u\0how_many\0out\0ai\0RESULT\0", purple_savedstatuses_get_popular_DBUS},
{"PurpleSavedstatusGetCurrent", "out\0i\0RESULT\0", purple_savedstatus_get_current_DBUS},
{"PurpleSavedstatusGetDefault", "out\0i\0RESULT\0", purple_savedstatus_get_default_DBUS},
{"PurpleSavedstatusGetIdleaway", "out\0i\0RESULT\0", purple_savedstatus_get_idleaway_DBUS},
{"PurpleSavedstatusIsIdleaway", "out\0i\0RESULT\0", purple_savedstatus_is_idleaway_DBUS},
{"PurpleSavedstatusSetIdleaway", "in\0i\0idleaway\0", purple_savedstatus_set_idleaway_DBUS},
{"PurpleSavedstatusGetStartup", "out\0i\0RESULT\0", purple_savedstatus_get_startup_DBUS},
{"PurpleSavedstatusFind", "in\0s\0title\0out\0i\0RESULT\0", purple_savedstatus_find_DBUS},
{"PurpleSavedstatusFindByCreationTime", "in\0i\0creation_time\0out\0i\0RESULT\0", purple_savedstatus_find_by_creation_time_DBUS},
{"PurpleSavedstatusFindTransientByTypeAndMessage", "in\0i\0type\0in\0s\0message\0out\0i\0RESULT\0", purple_savedstatus_find_transient_by_type_and_message_DBUS},
{"PurpleSavedstatusIsTransient", "in\0i\0saved_status\0out\0i\0RESULT\0", purple_savedstatus_is_transient_DBUS},
{"PurpleSavedstatusGetTitle", "in\0i\0saved_status\0out\0s\0RESULT\0", purple_savedstatus_get_title_DBUS},
{"PurpleSavedstatusGetType", "in\0i\0saved_status\0out\0i\0RESULT\0", purple_savedstatus_get_type_DBUS},
{"PurpleSavedstatusGetMessage", "in\0i\0saved_status\0out\0s\0RESULT\0", purple_savedstatus_get_message_DBUS},
{"PurpleSavedstatusGetCreationTime", "in\0i\0saved_status\0out\0i\0RESULT\0", purple_savedstatus_get_creation_time_DBUS},
{"PurpleSavedstatusHasSubstatuses", "in\0i\0saved_status\0out\0i\0RESULT\0", purple_savedstatus_has_substatuses_DBUS},
{"PurpleSavedstatusGetSubstatus", "in\0i\0saved_status\0in\0i\0account\0out\0i\0RESULT\0", purple_savedstatus_get_substatus_DBUS},
{"PurpleSavedstatusSubstatusGetType", "in\0i\0substatus\0out\0i\0RESULT\0", purple_savedstatus_substatus_get_type_DBUS},
{"PurpleSavedstatusSubstatusGetMessage", "in\0i\0substatus\0out\0s\0RESULT\0", purple_savedstatus_substatus_get_message_DBUS},
{"PurpleSavedstatusActivate", "in\0i\0saved_status\0", purple_savedstatus_activate_DBUS},
{"PurpleSavedstatusActivateForAccount", "in\0i\0saved_status\0in\0i\0account\0", purple_savedstatus_activate_for_account_DBUS},
{"PurpleSavedstatusesInit", "", purple_savedstatuses_init_DBUS},
{"PurpleSavedstatusesUninit", "", purple_savedstatuses_uninit_DBUS},
{"PurpleSmileyNew", "in\0i\0img\0in\0s\0shortcut\0out\0i\0RESULT\0", purple_smiley_new_DBUS},
{"PurpleSmileyNewFromFile", "in\0s\0shortcut\0in\0s\0filepath\0out\0i\0RESULT\0", purple_smiley_new_from_file_DBUS},
{"PurpleSmileyDelete", "in\0i\0smiley\0", purple_smiley_delete_DBUS},
{"PurpleSmileySetShortcut", "in\0i\0smiley\0in\0s\0shortcut\0out\0i\0RESULT\0", purple_smiley_set_shortcut_DBUS},
{"PurpleSmileySetData", "in\0i\0smiley\0in\0i\0smiley_data\0in\0i\0smiley_data_len\0", purple_smiley_set_data_DBUS},
{"PurpleSmileyGetShortcut", "in\0i\0smiley\0out\0s\0RESULT\0", purple_smiley_get_shortcut_DBUS},
{"PurpleSmileyGetChecksum", "in\0i\0smiley\0out\0s\0RESULT\0", purple_smiley_get_checksum_DBUS},
{"PurpleSmileyGetStoredImage", "in\0i\0smiley\0out\0i\0RESULT\0", purple_smiley_get_stored_image_DBUS},
{"PurpleSmileyGetData", "in\0i\0smiley\0out\0ay\0RESULT\0", purple_smiley_get_data_DBUS},
{"PurpleSmileyGetExtension", "in\0i\0smiley\0out\0s\0RESULT\0", purple_smiley_get_extension_DBUS},
{"PurpleSmileyGetFullPath", "in\0i\0smiley\0out\0s\0RESULT\0", purple_smiley_get_full_path_DBUS},
{"PurpleSmileysGetAll", "out\0ai\0RESULT\0", purple_smileys_get_all_DBUS},
{"PurpleSmileysFindByShortcut", "in\0s\0shortcut\0out\0i\0RESULT\0", purple_smileys_find_by_shortcut_DBUS},
{"PurpleSmileysFindByChecksum", "in\0s\0checksum\0out\0i\0RESULT\0", purple_smileys_find_by_checksum_DBUS},
{"PurpleSmileysGetStoringDir", "out\0s\0RESULT\0", purple_smileys_get_storing_dir_DBUS},
{"PurpleSmileysInit", "", purple_smileys_init_DBUS},
{"PurpleSmileysUninit", "", purple_smileys_uninit_DBUS},
{"PurplePrimitiveGetIdFromType", "in\0i\0type\0out\0s\0RESULT\0", purple_primitive_get_id_from_type_DBUS},
{"PurplePrimitiveGetNameFromType", "in\0i\0type\0out\0s\0RESULT\0", purple_primitive_get_name_from_type_DBUS},
{"PurplePrimitiveGetTypeFromId", "in\0s\0id\0out\0i\0RESULT\0", purple_primitive_get_type_from_id_DBUS},
{"PurpleStatusTypeNewFull", "in\0i\0primitive\0in\0s\0id\0in\0s\0name\0in\0i\0saveable\0in\0i\0user_settable\0in\0i\0independent\0out\0i\0RESULT\0", purple_status_type_new_full_DBUS},
{"PurpleStatusTypeNew", "in\0i\0primitive\0in\0s\0id\0in\0s\0name\0in\0i\0user_settable\0out\0i\0RESULT\0", purple_status_type_new_DBUS},
{"PurpleStatusTypeDestroy", "in\0i\0status_type\0", purple_status_type_destroy_DBUS},
{"PurpleStatusTypeSetPrimaryAttr", "in\0i\0status_type\0in\0s\0attr_id\0", purple_status_type_set_primary_attr_DBUS},
{"PurpleStatusTypeAddAttr", "in\0i\0status_type\0in\0s\0id\0in\0s\0name\0in\0i\0value\0", purple_status_type_add_attr_DBUS},
{"PurpleStatusTypeGetPrimitive", "in\0i\0status_type\0out\0i\0RESULT\0", purple_status_type_get_primitive_DBUS},
{"PurpleStatusTypeGetId", "in\0i\0status_type\0out\0s\0RESULT\0", purple_status_type_get_id_DBUS},
{"PurpleStatusTypeGetName", "in\0i\0status_type\0out\0s\0RESULT\0", purple_status_type_get_name_DBUS},
{"PurpleStatusTypeIsSaveable", "in\0i\0status_type\0out\0i\0RESULT\0", purple_status_type_is_saveable_DBUS},
{"PurpleStatusTypeIsUserSettable", "in\0i\0status_type\0out\0i\0RESULT\0", purple_status_type_is_user_settable_DBUS},
{"PurpleStatusTypeIsIndependent", "in\0i\0status_type\0out\0i\0RESULT\0", purple_status_type_is_independent_DBUS},
{"PurpleStatusTypeIsExclusive", "in\0i\0status_type\0out\0i\0RESULT\0", purple_status_type_is_exclusive_DBUS},
{"PurpleStatusTypeIsAvailable", "in\0i\0status_type\0out\0i\0RESULT\0", purple_status_type_is_available_DBUS},
{"PurpleStatusTypeGetPrimaryAttr", "in\0i\0type\0out\0s\0RESULT\0", purple_status_type_get_primary_attr_DBUS},
{"PurpleStatusTypeGetAttr", "in\0i\0status_type\0in\0s\0id\0out\0i\0RESULT\0", purple_status_type_get_attr_DBUS},
{"PurpleStatusTypeGetAttrs", "in\0i\0status_type\0out\0ai\0RESULT\0", purple_status_type_get_attrs_DBUS},
{"PurpleStatusTypeFindWithId", "in\0i\0status_types\0in\0s\0id\0out\0i\0RESULT\0", purple_status_type_find_with_id_DBUS},
{"PurpleStatusAttrNew", "in\0s\0id\0in\0s\0name\0in\0i\0value_type\0out\0i\0RESULT\0", purple_status_attr_new_DBUS},
{"PurpleStatusAttrDestroy", "in\0i\0attr\0", purple_status_attr_destroy_DBUS},
{"PurpleStatusAttrGetId", "in\0i\0attr\0out\0s\0RESULT\0", purple_status_attr_get_id_DBUS},
{"PurpleStatusAttrGetName", "in\0i\0attr\0out\0s\0RESULT\0", purple_status_attr_get_name_DBUS},
{"PurpleStatusAttrGetValue", "in\0i\0attr\0out\0i\0RESULT\0", purple_status_attr_get_value_DBUS},
{"PurpleStatusNew", "in\0i\0status_type\0in\0i\0presence\0out\0i\0RESULT\0", purple_status_new_DBUS},
{"PurpleStatusDestroy", "in\0i\0status\0", purple_status_destroy_DBUS},
{"PurpleStatusSetActive", "in\0i\0status\0in\0i\0active\0", purple_status_set_active_DBUS},
{"PurpleStatusSetActiveWithAttrsList", "in\0i\0status\0in\0i\0active\0in\0i\0attrs\0", purple_status_set_active_with_attrs_list_DBUS},
{"PurpleStatusSetAttrBoolean", "in\0i\0status\0in\0s\0id\0in\0i\0value\0", purple_status_set_attr_boolean_DBUS},
{"PurpleStatusSetAttrInt", "in\0i\0status\0in\0s\0id\0in\0i\0value\0", purple_status_set_attr_int_DBUS},
{"PurpleStatusSetAttrString", "in\0i\0status\0in\0s\0id\0in\0s\0value\0", purple_status_set_attr_string_DBUS},
{"PurpleStatusGetType", "in\0i\0status\0out\0i\0RESULT\0", purple_status_get_type_DBUS},
{"PurpleStatusGetPresence", "in\0i\0status\0out\0i\0RESULT\0", purple_status_get_presence_DBUS},
{"PurpleStatusGetId", "in\0i\0status\0out\0s\0RESULT\0", purple_status_get_id_DBUS},
{"PurpleStatusGetName", "in\0i\0status\0out\0s\0RESULT\0", purple_status_get_name_DBUS},
{"PurpleStatusIsIndependent", "in\0i\0status\0out\0i\0RESULT\0", purple_status_is_independent_DBUS},
{"PurpleStatusIsExclusive", "in\0i\0status\0out\0i\0RESULT\0", purple_status_is_exclusive_DBUS},
{"PurpleStatusIsAvailable", "in\0i\0status\0out\0i\0RESULT\0", purple_status_is_available_DBUS},
{"PurpleStatusIsActive", "in\0i\0status\0out\0i\0RESULT\0", purple_status_is_active_DBUS},
{"PurpleStatusIsOnline", "in\0i\0status\0out\0i\0RESULT\0", purple_status_is_online_DBUS},
{"PurpleStatusGetAttrValue", "in\0i\0status\0in\0s\0id\0out\0i\0RESULT\0", purple_status_get_attr_value_DBUS},
{"PurpleStatusGetAttrBoolean", "in\0i\0status\0in\0s\0id\0out\0i\0RESULT\0", purple_status_get_attr_boolean_DBUS},
{"PurpleStatusGetAttrInt", "in\0i\0status\0in\0s\0id\0out\0i\0RESULT\0", purple_status_get_attr_int_DBUS},
{"PurpleStatusGetAttrString", "in\0i\0status\0in\0s\0id\0out\0s\0RESULT\0", purple_status_get_attr_string_DBUS},
{"PurpleStatusCompare", "in\0i\0status1\0in\0i\0status2\0out\0i\0RESULT\0", purple_status_compare_DBUS},
{"PurplePresenceNew", "in\0i\0context\0out\0i\0RESULT\0", purple_presence_new_DBUS},
{"PurplePresenceNewForAccount", "in\0i\0account\0out\0i\0RESULT\0", purple_presence_new_for_account_DBUS},
{"PurplePresenceNewForConv", "in\0i\0conv\0out\0i\0RESULT\0", purple_presence_new_for_conv_DBUS},
{"PurplePresenceNewForBuddy", "in\0i\0buddy\0out\0i\0RESULT\0", purple_presence_new_for_buddy_DBUS},
{"PurplePresenceDestroy", "in\0i\0presence\0", purple_presence_destroy_DBUS},
{"PurplePresenceAddStatus", "in\0i\0presence\0in\0i\0status\0", purple_presence_add_status_DBUS},
{"PurplePresenceSetStatusActive", "in\0i\0presence\0in\0s\0status_id\0in\0i\0active\0", purple_presence_set_status_active_DBUS},
{"PurplePresenceSwitchStatus", "in\0i\0presence\0in\0s\0status_id\0", purple_presence_switch_status_DBUS},
{"PurplePresenceSetIdle", "in\0i\0presence\0in\0i\0idle\0in\0i\0idle_time\0", purple_presence_set_idle_DBUS},
{"PurplePresenceSetLoginTime", "in\0i\0presence\0in\0i\0login_time\0", purple_presence_set_login_time_DBUS},
{"PurplePresenceGetContext", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_context_DBUS},
{"PurplePresenceGetAccount", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_account_DBUS},
{"PurplePresenceGetConversation", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_conversation_DBUS},
{"PurplePresenceGetChatUser", "in\0i\0presence\0out\0s\0RESULT\0", purple_presence_get_chat_user_DBUS},
{"PurplePresenceGetBuddy", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_buddy_DBUS},
{"PurplePresenceGetStatuses", "in\0i\0presence\0out\0ai\0RESULT\0", purple_presence_get_statuses_DBUS},
{"PurplePresenceGetStatus", "in\0i\0presence\0in\0s\0status_id\0out\0i\0RESULT\0", purple_presence_get_status_DBUS},
{"PurplePresenceGetActiveStatus", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_active_status_DBUS},
{"PurplePresenceIsAvailable", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_is_available_DBUS},
{"PurplePresenceIsOnline", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_is_online_DBUS},
{"PurplePresenceIsStatusActive", "in\0i\0presence\0in\0s\0status_id\0out\0i\0RESULT\0", purple_presence_is_status_active_DBUS},
{"PurplePresenceIsStatusPrimitiveActive", "in\0i\0presence\0in\0i\0primitive\0out\0i\0RESULT\0", purple_presence_is_status_primitive_active_DBUS},
{"PurplePresenceIsIdle", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_is_idle_DBUS},
{"PurplePresenceGetIdleTime", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_idle_time_DBUS},
{"PurplePresenceGetLoginTime", "in\0i\0presence\0out\0i\0RESULT\0", purple_presence_get_login_time_DBUS},
{"PurplePresenceCompare", "in\0i\0presence1\0in\0i\0presence2\0out\0i\0RESULT\0", purple_presence_compare_DBUS},
{"PurpleStatusInit", "", purple_status_init_DBUS},
{"PurpleStatusUninit", "", purple_status_uninit_DBUS},
{"ServSendTyping", "in\0i\0gc\0in\0s\0name\0in\0i\0state\0out\0u\0RESULT\0", serv_send_typing_DBUS},
{"ServMoveBuddy", "in\0i\0param0\0in\0i\0param1\0in\0i\0param2\0", serv_move_buddy_DBUS},
{"ServSendIm", "in\0i\0param0\0in\0s\0param1\0in\0s\0param2\0in\0i\0flags\0out\0i\0RESULT\0", serv_send_im_DBUS},
{"PurpleGetAttentionTypeFromCode", "in\0i\0account\0in\0i\0type_code\0out\0i\0RESULT\0", purple_get_attention_type_from_code_DBUS},
{"ServSendAttention", "in\0i\0gc\0in\0s\0who\0in\0i\0type_code\0", serv_send_attention_DBUS},
{"ServGotAttention", "in\0i\0gc\0in\0s\0who\0in\0i\0type_code\0", serv_got_attention_DBUS},
{"ServGetInfo", "in\0i\0param0\0in\0s\0param1\0", serv_get_info_DBUS},
{"ServSetInfo", "in\0i\0param0\0in\0s\0param1\0", serv_set_info_DBUS},
{"ServAddPermit", "in\0i\0param0\0in\0s\0param1\0", serv_add_permit_DBUS},
{"ServAddDeny", "in\0i\0param0\0in\0s\0param1\0", serv_add_deny_DBUS},
{"ServRemPermit", "in\0i\0param0\0in\0s\0param1\0", serv_rem_permit_DBUS},
{"ServRemDeny", "in\0i\0param0\0in\0s\0param1\0", serv_rem_deny_DBUS},
{"ServSetPermitDeny", "in\0i\0param0\0", serv_set_permit_deny_DBUS},
{"ServChatInvite", "in\0i\0param0\0in\0i\0param1\0in\0s\0param2\0in\0s\0param3\0", serv_chat_invite_DBUS},
{"ServChatLeave", "in\0i\0param0\0in\0i\0param1\0", serv_chat_leave_DBUS},
{"ServChatWhisper", "in\0i\0param0\0in\0i\0param1\0in\0s\0param2\0in\0s\0param3\0", serv_chat_whisper_DBUS},
{"ServChatSend", "in\0i\0param0\0in\0i\0param1\0in\0s\0param2\0in\0i\0flags\0out\0i\0RESULT\0", serv_chat_send_DBUS},
{"ServAliasBuddy", "in\0i\0param0\0", serv_alias_buddy_DBUS},
{"ServGotAlias", "in\0i\0gc\0in\0s\0who\0in\0s\0alias\0", serv_got_alias_DBUS},
{"PurpleServGotPrivateAlias", "in\0i\0gc\0in\0s\0who\0in\0s\0alias\0", purple_serv_got_private_alias_DBUS},
{"ServGotTyping", "in\0i\0gc\0in\0s\0name\0in\0i\0timeout\0in\0i\0state\0", serv_got_typing_DBUS},
{"ServGotTypingStopped", "in\0i\0gc\0in\0s\0name\0", serv_got_typing_stopped_DBUS},
{"ServGotIm", "in\0i\0gc\0in\0s\0who\0in\0s\0msg\0in\0i\0flags\0in\0i\0mtime\0", serv_got_im_DBUS},
{"ServJoinChat", "in\0i\0param0\0in\0a{ss}\0data\0", serv_join_chat_DBUS},
{"ServRejectChat", "in\0i\0param0\0in\0a{ss}\0data\0", serv_reject_chat_DBUS},
{"ServGotChatInvite", "in\0i\0gc\0in\0s\0name\0in\0s\0who\0in\0s\0message\0in\0a{ss}\0data\0", serv_got_chat_invite_DBUS},
{"ServGotJoinedChat", "in\0i\0gc\0in\0i\0id\0in\0s\0name\0out\0i\0RESULT\0", serv_got_joined_chat_DBUS},
{"PurpleServGotJoinChatFailed", "in\0i\0gc\0in\0a{ss}\0data\0", purple_serv_got_join_chat_failed_DBUS},
{"ServGotChatLeft", "in\0i\0g\0in\0i\0id\0", serv_got_chat_left_DBUS},
{"ServGotChatIn", "in\0i\0g\0in\0i\0id\0in\0s\0who\0in\0i\0flags\0in\0s\0message\0in\0i\0mtime\0", serv_got_chat_in_DBUS},
{"ServSendFile", "in\0i\0gc\0in\0s\0who\0in\0s\0file\0", serv_send_file_DBUS},
{"ServChatSendFile", "in\0i\0gc\0in\0i\0id\0in\0s\0file\0", serv_chat_send_file_DBUS},
{"PurpleMenuActionFree", "in\0i\0act\0", purple_menu_action_free_DBUS},
{"PurpleUtilSetCurrentSong", "in\0s\0title\0in\0s\0artist\0in\0s\0album\0", purple_util_set_current_song_DBUS},
{"PurpleUtilInit", "", purple_util_init_DBUS},
{"PurpleUtilUninit", "", purple_util_uninit_DBUS},
{"PurpleMimeDecodeField", "in\0s\0str\0out\0s\0RESULT\0", purple_mime_decode_field_DBUS},
{"PurpleTimeBuild", "in\0i\0year\0in\0i\0month\0in\0i\0day\0in\0i\0hour\0in\0i\0min\0in\0i\0sec\0out\0i\0RESULT\0", purple_time_build_DBUS},
{"PurpleMarkupEscapeText", "in\0s\0text\0in\0i\0length\0out\0s\0RESULT\0", purple_markup_escape_text_DBUS},
{"PurpleMarkupStripHtml", "in\0s\0str\0out\0s\0RESULT\0", purple_markup_strip_html_DBUS},
{"PurpleMarkupLinkify", "in\0s\0str\0out\0s\0RESULT\0", purple_markup_linkify_DBUS},
{"PurpleUnescapeText", "in\0s\0text\0out\0s\0RESULT\0", purple_unescape_text_DBUS},
{"PurpleUnescapeHtml", "in\0s\0html\0out\0s\0RESULT\0", purple_unescape_html_DBUS},
{"PurpleMarkupSlice", "in\0s\0str\0in\0i\0x\0in\0i\0y\0out\0s\0RESULT\0", purple_markup_slice_DBUS},
{"PurpleMarkupGetTagName", "in\0s\0tag\0out\0s\0RESULT\0", purple_markup_get_tag_name_DBUS},
{"PurpleMarkupUnescapeEntity", "in\0s\0text\0in\0i\0length\0out\0s\0RESULT\0", purple_markup_unescape_entity_DBUS},
{"PurpleMarkupGetCssProperty", "in\0s\0style\0in\0s\0opt\0out\0s\0RESULT\0", purple_markup_get_css_property_DBUS},
{"PurpleMarkupIsRtl", "in\0s\0html\0out\0i\0RESULT\0", purple_markup_is_rtl_DBUS},
{"PurpleHomeDir", "out\0s\0RESULT\0", purple_home_dir_DBUS},
{"PurpleUserDir", "out\0s\0RESULT\0", purple_user_dir_DBUS},
{"PurpleUtilSetUserDir", "in\0s\0dir\0", purple_util_set_user_dir_DBUS},
{"PurpleBuildDir", "in\0s\0path\0in\0i\0mode\0out\0i\0RESULT\0", purple_build_dir_DBUS},
{"PurpleUtilWriteDataToFile", "in\0s\0filename\0in\0s\0data\0in\0i\0size\0out\0i\0RESULT\0", purple_util_write_data_to_file_DBUS},
{"PurpleUtilWriteDataToFileAbsolute", "in\0s\0filename_full\0in\0s\0data\0in\0i\0size\0out\0i\0RESULT\0", purple_util_write_data_to_file_absolute_DBUS},
{"PurpleProgramIsValid", "in\0s\0program\0out\0i\0RESULT\0", purple_program_is_valid_DBUS},
{"PurpleRunningGnome", "out\0i\0RESULT\0", purple_running_gnome_DBUS},
{"PurpleRunningKde", "out\0i\0RESULT\0", purple_running_kde_DBUS},
{"PurpleRunningOsx", "out\0i\0RESULT\0", purple_running_osx_DBUS},
{"PurpleFdGetIp", "in\0i\0fd\0out\0s\0RESULT\0", purple_fd_get_ip_DBUS},
{"PurpleSocketGetFamily", "in\0i\0fd\0out\0i\0RESULT\0", purple_socket_get_family_DBUS},
{"PurpleSocketSpeaksIpv4", "in\0i\0fd\0out\0i\0RESULT\0", purple_socket_speaks_ipv4_DBUS},
{"PurpleStrequal", "in\0s\0left\0in\0s\0right\0out\0i\0RESULT\0", purple_strequal_DBUS},
{"PurpleNormalize", "in\0i\0account\0in\0s\0str\0out\0s\0RESULT\0", purple_normalize_DBUS},
{"PurpleNormalizeNocase", "in\0i\0account\0in\0s\0str\0out\0s\0RESULT\0", purple_normalize_nocase_DBUS},
{"PurpleStrHasPrefix", "in\0s\0s\0in\0s\0p\0out\0i\0RESULT\0", purple_str_has_prefix_DBUS},
{"PurpleStrHasSuffix", "in\0s\0s\0in\0s\0x\0out\0i\0RESULT\0", purple_str_has_suffix_DBUS},
{"PurpleStrdupWithhtml", "in\0s\0src\0out\0s\0RESULT\0", purple_strdup_withhtml_DBUS},
{"PurpleStrAddCr", "in\0s\0str\0out\0s\0RESULT\0", purple_str_add_cr_DBUS},
{"PurpleStrreplace", "in\0s\0string\0in\0s\0delimiter\0in\0s\0replacement\0out\0s\0RESULT\0", purple_strreplace_DBUS},
{"PurpleUtf8NcrEncode", "in\0s\0in\0out\0s\0RESULT\0", purple_utf8_ncr_encode_DBUS},
{"PurpleUtf8NcrDecode", "in\0s\0in\0out\0s\0RESULT\0", purple_utf8_ncr_decode_DBUS},
{"PurpleStrcasereplace", "in\0s\0string\0in\0s\0delimiter\0in\0s\0replacement\0out\0s\0RESULT\0", purple_strcasereplace_DBUS},
{"PurpleStrcasestr", "in\0s\0haystack\0in\0s\0needle\0out\0s\0RESULT\0", purple_strcasestr_DBUS},
{"PurpleStrSizeToUnits", "in\0i\0size\0out\0s\0RESULT\0", purple_str_size_to_units_DBUS},
{"PurpleStrSecondsToString", "in\0i\0sec\0out\0s\0RESULT\0", purple_str_seconds_to_string_DBUS},
{"PurpleStrBinaryToAscii", "in\0s\0binary\0in\0i\0len\0out\0s\0RESULT\0", purple_str_binary_to_ascii_DBUS},
{"PurpleGotProtocolHandlerUri", "in\0s\0uri\0", purple_got_protocol_handler_uri_DBUS},
{"PurpleUtilFetchUrlCancel", "in\0i\0url_data\0", purple_util_fetch_url_cancel_DBUS},
{"PurpleUrlDecode", "in\0s\0str\0out\0s\0RESULT\0", purple_url_decode_DBUS},
{"PurpleUrlEncode", "in\0s\0str\0out\0s\0RESULT\0", purple_url_encode_DBUS},
{"PurpleEmailIsValid", "in\0s\0address\0out\0i\0RESULT\0", purple_email_is_valid_DBUS},
{"PurpleIpAddressIsValid", "in\0s\0ip\0out\0i\0RESULT\0", purple_ip_address_is_valid_DBUS},
{"PurpleIpv4AddressIsValid", "in\0s\0ip\0out\0i\0RESULT\0", purple_ipv4_address_is_valid_DBUS},
{"PurpleIpv6AddressIsValid", "in\0s\0ip\0out\0i\0RESULT\0", purple_ipv6_address_is_valid_DBUS},
{"PurpleUriListExtractUris", "in\0s\0uri_list\0out\0as\0RESULT\0", purple_uri_list_extract_uris_DBUS},
{"PurpleUriListExtractFilenames", "in\0s\0uri_list\0out\0as\0RESULT\0", purple_uri_list_extract_filenames_DBUS},
{"PurpleUriEscapeForOpen", "in\0s\0unescaped\0out\0s\0RESULT\0", purple_uri_escape_for_open_DBUS},
{"PurpleUtf8TryConvert", "in\0s\0str\0out\0s\0RESULT\0", purple_utf8_try_convert_DBUS},
{"PurpleUtf8Salvage", "in\0s\0str\0out\0s\0RESULT\0", purple_utf8_salvage_DBUS},
{"PurpleUtf8StripUnprintables", "in\0s\0str\0out\0s\0RESULT\0", purple_utf8_strip_unprintables_DBUS},
{"PurpleGaiStrerror", "in\0i\0errnum\0out\0s\0RESULT\0", purple_gai_strerror_DBUS},
{"PurpleUtf8Strcasecmp", "in\0s\0a\0in\0s\0b\0out\0i\0RESULT\0", purple_utf8_strcasecmp_DBUS},
{"PurpleUtf8HasWord", "in\0s\0haystack\0in\0s\0needle\0out\0i\0RESULT\0", purple_utf8_has_word_DBUS},
{"PurpleTextStripMnemonic", "in\0s\0in\0out\0s\0RESULT\0", purple_text_strip_mnemonic_DBUS},
{"PurpleUnescapeFilename", "in\0s\0str\0out\0s\0RESULT\0", purple_unescape_filename_DBUS},
{"PurpleEscapeFilename", "in\0s\0str\0out\0s\0RESULT\0", purple_escape_filename_DBUS},
{"PurpleOscarConvert", "in\0s\0act\0in\0s\0protocol\0out\0s\0RESULT\0", _purple_oscar_convert_DBUS},
{"PurpleRestoreDefaultSignalHandlers", "", purple_restore_default_signal_handlers_DBUS},
{"PurpleGetHostName", "out\0s\0RESULT\0", purple_get_host_name_DBUS},
{"PurpleUuidRandom", "out\0s\0RESULT\0", purple_uuid_random_DBUS},
{"XmlnodeInsertChild", "in\0i\0parent\0in\0i\0child\0", xmlnode_insert_child_DBUS},
{"XmlnodeInsertData", "in\0i\0node\0in\0s\0data\0in\0i\0size\0", xmlnode_insert_data_DBUS},
{"XmlnodeGetData", "in\0i\0node\0out\0s\0RESULT\0", xmlnode_get_data_DBUS},
{"XmlnodeGetDataUnescaped", "in\0i\0node\0out\0s\0RESULT\0", xmlnode_get_data_unescaped_DBUS},
{"XmlnodeSetAttrib", "in\0i\0node\0in\0s\0attr\0in\0s\0value\0", xmlnode_set_attrib_DBUS},
{"XmlnodeSetAttribWithPrefix", "in\0i\0node\0in\0s\0attr\0in\0s\0prefix\0in\0s\0value\0", xmlnode_set_attrib_with_prefix_DBUS},
{"XmlnodeSetAttribWithNamespace", "in\0i\0node\0in\0s\0attr\0in\0s\0xmlns\0in\0s\0value\0", xmlnode_set_attrib_with_namespace_DBUS},
{"XmlnodeSetAttribFull", "in\0i\0node\0in\0s\0attr\0in\0s\0xmlns\0in\0s\0prefix\0in\0s\0value\0", xmlnode_set_attrib_full_DBUS},
{"XmlnodeGetAttrib", "in\0i\0node\0in\0s\0attr\0out\0s\0RESULT\0", xmlnode_get_attrib_DBUS},
{"XmlnodeGetAttribWithNamespace", "in\0i\0node\0in\0s\0attr\0in\0s\0xmlns\0out\0s\0RESULT\0", xmlnode_get_attrib_with_namespace_DBUS},
{"XmlnodeRemoveAttrib", "in\0i\0node\0in\0s\0attr\0", xmlnode_remove_attrib_DBUS},
{"XmlnodeRemoveAttribWithNamespace", "in\0i\0node\0in\0s\0attr\0in\0s\0xmlns\0", xmlnode_remove_attrib_with_namespace_DBUS},
{"XmlnodeSetNamespace", "in\0i\0node\0in\0s\0xmlns\0", xmlnode_set_namespace_DBUS},
{"XmlnodeGetNamespace", "in\0i\0node\0out\0s\0RESULT\0", xmlnode_get_namespace_DBUS},
{"XmlnodeSetPrefix", "in\0i\0node\0in\0s\0prefix\0", xmlnode_set_prefix_DBUS},
{"XmlnodeGetPrefix", "in\0i\0node\0out\0s\0RESULT\0", xmlnode_get_prefix_DBUS},
{"XmlnodeToStr", "in\0i\0node\0in\0i\0len\0out\0s\0RESULT\0", xmlnode_to_str_DBUS},
{"XmlnodeToFormattedStr", "in\0i\0node\0in\0i\0len\0out\0s\0RESULT\0", xmlnode_to_formatted_str_DBUS},
{"XmlnodeFree", "in\0i\0node\0", xmlnode_free_DBUS},
{"PurpleAttentionTypeNew", "in\0s\0ulname\0in\0s\0name\0in\0s\0inc_desc\0in\0s\0out_desc\0out\0i\0RESULT\0", purple_attention_type_new_DBUS},
{"PurpleAttentionTypeSetName", "in\0i\0type\0in\0s\0name\0", purple_attention_type_set_name_DBUS},
{"PurpleAttentionTypeSetIncomingDesc", "in\0i\0type\0in\0s\0desc\0", purple_attention_type_set_incoming_desc_DBUS},
{"PurpleAttentionTypeSetOutgoingDesc", "in\0i\0type\0in\0s\0desc\0", purple_attention_type_set_outgoing_desc_DBUS},
{"PurpleAttentionTypeSetIconName", "in\0i\0type\0in\0s\0name\0", purple_attention_type_set_icon_name_DBUS},
{"PurpleAttentionTypeSetUnlocalizedName", "in\0i\0type\0in\0s\0ulname\0", purple_attention_type_set_unlocalized_name_DBUS},
{"PurpleAttentionTypeGetName", "in\0i\0type\0out\0s\0RESULT\0", purple_attention_type_get_name_DBUS},
{"PurpleAttentionTypeGetIncomingDesc", "in\0i\0type\0out\0s\0RESULT\0", purple_attention_type_get_incoming_desc_DBUS},
{"PurpleAttentionTypeGetOutgoingDesc", "in\0i\0type\0out\0s\0RESULT\0", purple_attention_type_get_outgoing_desc_DBUS},
{"PurpleAttentionTypeGetIconName", "in\0i\0type\0out\0s\0RESULT\0", purple_attention_type_get_icon_name_DBUS},
{"PurpleAttentionTypeGetUnlocalizedName", "in\0i\0type\0out\0s\0RESULT\0", purple_attention_type_get_unlocalized_name_DBUS},
{"PurplePrplGotAccountIdle", "in\0i\0account\0in\0i\0idle\0in\0i\0idle_time\0", purple_prpl_got_account_idle_DBUS},
{"PurplePrplGotAccountLoginTime", "in\0i\0account\0in\0i\0login_time\0", purple_prpl_got_account_login_time_DBUS},
{"PurplePrplGotAccountActions", "in\0i\0account\0", purple_prpl_got_account_actions_DBUS},
{"PurplePrplGotUserIdle", "in\0i\0account\0in\0s\0name\0in\0i\0idle\0in\0i\0idle_time\0", purple_prpl_got_user_idle_DBUS},
{"PurplePrplGotUserLoginTime", "in\0i\0account\0in\0s\0name\0in\0i\0login_time\0", purple_prpl_got_user_login_time_DBUS},
{"PurplePrplGotUserStatusDeactive", "in\0i\0account\0in\0s\0name\0in\0s\0status_id\0", purple_prpl_got_user_status_deactive_DBUS},
{"PurplePrplChangeAccountStatus", "in\0i\0account\0in\0i\0old_status\0in\0i\0new_status\0", purple_prpl_change_account_status_DBUS},
{"PurplePrplGetStatuses", "in\0i\0account\0in\0i\0presence\0out\0ai\0RESULT\0", purple_prpl_get_statuses_DBUS},
{"PurplePrplSendAttention", "in\0i\0gc\0in\0s\0who\0in\0i\0type_code\0", purple_prpl_send_attention_DBUS},
{"PurplePrplGotAttention", "in\0i\0gc\0in\0s\0who\0in\0i\0type_code\0", purple_prpl_got_attention_DBUS},
{"PurplePrplGotAttentionInChat", "in\0i\0gc\0in\0i\0id\0in\0s\0who\0in\0i\0type_code\0", purple_prpl_got_attention_in_chat_DBUS},
{"PurplePrplGetMediaCaps", "in\0i\0account\0in\0s\0who\0out\0i\0RESULT\0", purple_prpl_get_media_caps_DBUS},
{"PurplePrplInitiateMedia", "in\0i\0account\0in\0s\0who\0in\0i\0type\0out\0i\0RESULT\0", purple_prpl_initiate_media_DBUS},
{"PurplePrplGotMediaCaps", "in\0i\0account\0in\0s\0who\0", purple_prpl_got_media_caps_DBUS},
{"PurpleFindPrpl", "in\0s\0id\0out\0i\0RESULT\0", purple_find_prpl_DBUS},
{NULL, NULL, NULL}
};
#define PURPLE_DBUS_REGISTER_BINDINGS(handle) purple_dbus_register_bindings(handle, bindings_DBUS)
